/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.adapter;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * The <i>sequence adapter</i> acts as a <i>generic type adapter</i> for 
  * sequence parameterised types, i.e. from {@code Sequence<E>} to
  * {@code Sequence<T>}, by using an adapter {@link AdapterDelegate delegate}
  * to perform the representation shift. <p>
  * 
  * An sequence adapter exhibit the same {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link Sequence#unique() unique}
  * properties as the adapted sequence. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * It is essentially an <i>object</i> adapter in that it forwards all
  * sequence requests to the adapted sequence, but a bit unusual in 
  * that it does not adapt the <i>raw type</i>, but the <i>instantiated 
  * generic type</i> of its own raw type. In that respect it is more related
  * to <i>class</i> adaptation than object adaptation. Of cause, a 
  * combination that adapts both the raw and parameter types can be
  * made as well. <p>
  *   
  * An adapted sequence based on the type parameters is more than just
  * a cast into a given sequence type. Casting with type parameters is 
  * unsafe, and would cause a compiler warning and perhaps runtime errors, 
  * but a sequence adapter constructs a new generic type based on the 
  * type parameters supplied to it with full type-safety. <p>
  * 
  * The {@code <? super E,? extends T>} bounds for the adapter 
  * delegate used ensures greater flexibility than just using 
  * {@code <E,T>}. For example, the {@link AdapterStrategy#NUMBER_TO_STRING}
  * strategy can be used to adapt any sequence delivering sub-classes
  * of {@link java.lang.Number java.lang.Number} into a sequence 
  * delivering {@link java.lang.String java.lang.String} values:
  * <pre>   
  *   {@link Sequence}&lt;Integer> integerSequence = ..;
  *   {@link Sequence}&lt;String> stringSequence = 
  *     new {@link #SequenceAdapter(Sequence, AdapterDelegate) SequenceAdapter}&lt;Integer,String>(
  *       integerSequence, 
  *       {@link AdapterStrategy#NUMBER_TO_STRING} // ok, Number super-class of Integer!
  *     );    
  * </pre> 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    AdapterDelegate 
  * @see    #create(Sequence)
  * 
  * @param  <E> The adaptee type adapted by this sequence adapter.
  * @param  <T> The adapted (target) type delivered by this sequence adapter.
  */
@ParticipantUsage(value = "Target", type = Sequence.class)
@Participant("Adapter")
public class SequenceAdapter<E,T> implements Sequence<T>  {

    /**
      * The original sequence handling type {@code E}. <p>
      * 
      * Never null. 
      */
    @Participant("Adaptee")
    final Sequence<E> sequence;
    /**
      * The adapter delegate to transform sequence values
      * of type {@code E} into type {@code T}. <p>
      * 
      * Never null.
      */
    private final AdapterDelegate<? super E,? extends T> adapterDelegate;
    
    /**
      * Constructor.
      * 
      * @param  sequence        The sequence adaptee; cannot be null.
      * @param  adapterDelegate The adapter delegate to adapt type {@code E} 
      *                         into type {@code T}; cannot be null.
      * @throws NullPointerException If either argument is null.                 
      */
    public SequenceAdapter(@Participant("Adaptee") Sequence<E> sequence, AdapterDelegate<? super E,? extends T> adapterDelegate) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        } else if ((this.adapterDelegate = adapterDelegate) == null) {
            throw new NullPointerException("Adapter delegate cannot be null");
        }
    }

    /**
      * Copy constructor. <p>
      * 
      * The adapted sequence is also copied.
      * 
      * @param  adapter  The sequence adapter to copy; cannot be null.
      * @throws NullPointerException If {@code adapter} is null.                 
      */
    public SequenceAdapter(SequenceAdapter<E,T> adapter) {
        this.sequence = adapter.sequence.copy();
        this.adapterDelegate = adapter.adapterDelegate;
    }
    
    /**
      * Factory method that creates a {@link Sequence sequence}
      * adapting the type of values supplied by {@code sequence}
      * to {@code Object}. <p>
      * 
      * This is useful if wild-cards is to be used as sequence
      * values, but where the type must still be known locally,
      * for example in case of decoration. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * The returned type is not just a cast into {@code Sequence<Object>},
      * which would cause an unchecked warning by the way, but a new type
      * with the specific generic type {@code Sequence<Object>}. The
      * generic type parameter {@code V} is required for the 
      * actual {@code SequenceAdapter} created, and hence a wild-card
      * cannot be used. 
      * 
      * @param  <V>      The type of values delivered by {@code sequence}.
      * @param  sequence The sequence to adapt; cannot be null.
      * @return An adapted sequence, where the returned values 
      *         are of type {@code Object}; never null.
      * @throws NullPointerException If {@code sequence} is null. 
      */
    public static <V> Sequence<Object> create(Sequence<V> sequence) {
        return new SequenceAdapter<V,Object>(sequence, AdapterStrategy.ANY_TO_OBJECT);
    }
    
    /**
      * 
      */
    public Sequence<T> copy() {
        return new SequenceAdapter<E,T>(this);
    }
    
    /**
      * Returns the adapter delegate used to perform the
      * representation shift from type {@code E} to type 
      * {@code T}. 
      *  
      * @return The delegate; never null.
      */
    public AdapterDelegate<? super E,? extends T> getAdapterDelegate() {
        return this.adapterDelegate;
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return this.sequence.bounded();
    }

    /**
      *
      */
    public boolean consistent() {
        return this.sequence.consistent();
    }

    /**
      *
      */
    public boolean unique() {
        return this.sequence.unique();
    }
    
    /**
      * 
      */
    public T current() {
        return this.adapterDelegate.adapt(this.sequence.current());
    }

    /**
      * 
      */
    public T next() {
        return this.adapterDelegate.adapt(this.sequence.next());
    }

    /**
      * 
      */
    public void reset() {
        this.sequence.reset();
    }

    /**
      * 
      */
    public State state() {
        return this.sequence.state();
    }

    /**
      * Returns the hash code of this sequence.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.sequence.hashCode();
    }
    
    /**
      * Returns true if {@code object} is a sequence
      * adapter adapting the same sequence as this
      * adapter based on {@code equals(Object)},
      * false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof SequenceAdapter)) {
            return false;
        }
        SequenceAdapter<?,?> sa = (SequenceAdapter<?,?>)object;
        return this.sequence.equals(sa.sequence);
    }
    
    /**
      * 
      */
    public String toString() {
        return this.toString(null).toString();
    }

    /**
      * 
      */
    public StringablePolicy<? super Sequence<T>> getStringablePolicy(StringablePolicy<? super Sequence<T>> policy) {
        if (policy == null) {
            return AdaptedPolicy.VISIBLE;
        }
        return policy;
    }
    
    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<T>> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }

}
