/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Const.EOF;
import static dk.rode.thesis.meta.util.Const.INFINITE;
import static java.lang.reflect.Proxy.isProxyClass;

import java.lang.reflect.InvocationHandler;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import dk.rode.thesis.memento.MemorizableException;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.memento.SequenceMemento;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.proxy.ProxyFactory;
import dk.rode.thesis.meta.reflect.proxy.ReferenceHandler;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Bridge</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Bridge</b> tests.
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the test fails unexpectedly.             
      */
    public static void main(String[] args) throws Exception {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }

    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) throws Exception {
        out.heading("Bridge tests...");

        boolean ok = true;
        
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(MemorizableSequenceAbstraction.class).setActive(log);
        LogFactory.getLog(ReferenceHandler.class).setActive(log);
        
        log = arguments.getArgumentValue("log.verbose", false, Boolean.class);
        LogFactory.getLog(InvocationHandler.class).setActive(log);
        
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        out.heading("Test values");
        
        // Test values...
        Integer[] values = {1, -3, 2, 5, 3, 4, -1, -4, 2, 3, 4, -5, 4, 3, 2, -4, 1, 0, -1, -2};
        List<Integer> list = Arrays.asList(values); 

        out.println("Values: ", list);
        out.println("Length: ", list.size());
        
        //
        // Implementations
        //
        
        SequenceValueArrayList<Integer> svali = 
            new SequenceValueArrayList<Integer>(Integer.class, new ArrayList<Integer>(list));

        // Sorted according to natural ordering, with duplicates...
        ArrayList<Integer> sorted = new ArrayList<Integer>(list);
        Collections.sort(sorted, null);
        SequenceValueArrayList<Integer> svalsi = 
            new SequenceValueArrayList<Integer>(Integer.class, sorted, true, true); 
        
        SequenceValueHashSet<Integer> svhsi = 
            new SequenceValueHashSet<Integer>(Integer.class, new HashSet<Integer>(list));

        SequenceValueLinkedHashSet<Integer> svlhsi = 
            new SequenceValueLinkedHashSet<Integer>(Integer.class, new LinkedHashSet<Integer>(list));

        SequenceValueTreeSet<Integer> svtsi = 
            new SequenceValueTreeSet<Integer>(Integer.class, new TreeSet<Integer>(list));

        // Same value range as the test values...
        SequenceValueRange svr = new SequenceValueRange(-5, 5);
        // Unbounded...
        SequenceValueRange svru = new SequenceValueRange(-5);

        //
        // Shared implementations...
        //
        
        ProxyFactory pf = new ProxyFactory();
        
        // Shared implementation that will only be copied when advanced or reset...
        SequenceValueGenerator<Integer> svg1 = pf.getSharedObject(svr.copy(), "copy", null, "first", "get", "reset");
        SequenceValueGenerator<Integer> svg2 = svg1.copy();
        
        // Shared implementation that will only be copied when reset. Not realistic
        // for a sequence, but it illustrates that the shared object can be used
        // "for some time" before being copied...
        SequenceValueGenerator<Integer> svg3 = pf.getSharedObject(svtsi.copy(), "copy", null, "reset");
        SequenceValueGenerator<Integer> svg4 = svg3.copy();
        
        
        List<SequenceValueGenerator<Integer>> implementations = 
            new ArrayList<SequenceValueGenerator<Integer>>(6);
        implementations.add(svali);
        implementations.add(svalsi);
        implementations.add(svhsi);
        implementations.add(svlhsi);
        implementations.add(svtsi);
        implementations.add(svr);
        implementations.add(svru);
        // Shared!
        implementations.add(svg1);
        implementations.add(svg2);
        implementations.add(svg3);
        implementations.add(svg4);

        out.heading("Created implementations");

        out.println("(last four are shared!)").println();
        
        for (int i = 0; i < implementations.size(); i++) {
            SequenceValueGenerator<Integer> svc = implementations.get(i);
            out.println("Implementation ", i, ": ", svc);
        }
        out.println();
        
        out.heading("Testing created implementations");
        
        for (int i = 0; i < implementations.size(); i++) {
            SequenceValueGenerator<Integer> svc = implementations.get(i);
            out.println("Implementation ", i, ": ", svc).println(" -> ", svc.toString(SequenceValuePolicy.VALUES)).println();
            
            // From minimum to maximum...
            for (int j = -5; j <= 5; j++) {
                int number = 1, n = 0;
                while ((n = svc.numberOf(j, number)) != EOF) {
                    out.println("Number of ", j, " from ", number, ": ", n);
                    number = n + 1;
                }
            }
            
            SequenceValueGenerator<Integer> copy = svc.copy();
            if (!equals(copy, svc)) {
                out.error("Implementation copy does not equal original: ", copy, " != ", svc, " (original)");
                ok = false;
            }
            out.println();
            int size = copy.size();
            if (size == INFINITE) {
                out.warn("Sequence is unbounded - fetching 25 values!");
                size = 25;
            }
            out.println("Getting values from number:");
            for (int j = 1; j <= size; j++) {
                // This can only be done in the forward direction...
                Integer v = (j % 2 == 0) ? copy.get(j) : copy.get();
                out.print(v).print(j < size, ", ");
            }
            out.println().println("Getting values from number in reverse:");
            for (int j = size; j > 0; j--) {
                Integer v = copy.get(j);
                out.print(v).print(j > 1, ", ");
            }
            out.println().println();
            
        }
        out.println();
        
        //
        // Abstractions
        //
        
        // Sequence abstractions...
        SequenceAbstraction<Integer> sasvali = new SequenceAbstraction<Integer>(svali);
        SequenceAbstraction<Integer> sasvalsi = new SequenceAbstraction<Integer>(svalsi);
        SequenceAbstraction<Integer> sasvhsi = new SequenceAbstraction<Integer>(svhsi);
        SequenceAbstraction<Integer> sasvlhsi = new SequenceAbstraction<Integer>(svlhsi);
        SequenceAbstraction<Integer> sasvtsi = new SequenceAbstraction<Integer>(svtsi);
        SequenceAbstraction<Integer> sasvri = new SequenceAbstraction<Integer>(svr);
        // Create, then attach...
        SequenceAbstraction<Integer> sasvriu = new SequenceAbstraction<Integer>();
        sasvriu.setGenerator(svru);
        
        // Memorizable sequence abstractions... We copy the implementations
        // do we do not have to share state...
        SequenceAbstraction<Integer> msasvali = new MemorizableSequenceAbstraction<Integer>(svali.copy());
        SequenceAbstraction<Integer> msasvalsi = new MemorizableSequenceAbstraction<Integer>(svalsi.copy());
        SequenceAbstraction<Integer> msasvhsi = new MemorizableSequenceAbstraction<Integer>(svhsi.copy());
        SequenceAbstraction<Integer> msasvlhsi = new MemorizableSequenceAbstraction<Integer>(svlhsi.copy());
        SequenceAbstraction<Integer> msasvtsi = new MemorizableSequenceAbstraction<Integer>(svtsi.copy());
        SequenceAbstraction<Integer> msasvri = new MemorizableSequenceAbstraction<Integer>(svr.copy());
        // Create, then attach...
        SequenceAbstraction<Integer> msasvriu = new MemorizableSequenceAbstraction<Integer>();
        msasvriu.setGenerator(svru.copy());

        // Synchronised sequence abstractions... We copy the implementations
        // do we do not have to share state...
        SequenceAbstraction<Integer> ssasvali = new SynchronisedSequenceAbstraction<Integer>(svali.copy());
        SequenceAbstraction<Integer> ssasvalsi = new SynchronisedSequenceAbstraction<Integer>(svalsi.copy());
        SequenceAbstraction<Integer> ssasvhsi = new SynchronisedSequenceAbstraction<Integer>(svhsi.copy());
        SequenceAbstraction<Integer> ssasvlhsi = new SynchronisedSequenceAbstraction<Integer>(svlhsi.copy());
        SequenceAbstraction<Integer> ssasvtsi = new SynchronisedSequenceAbstraction<Integer>(svtsi.copy());
        SequenceAbstraction<Integer> ssasvri = new SynchronisedSequenceAbstraction<Integer>(svr.copy());
        // Create, then attach...
        SequenceAbstraction<Integer> ssasvriu = new SynchronisedSequenceAbstraction<Integer>();
        ssasvriu.setGenerator(svru.copy());
        
        List<SequenceAbstraction<Integer>> sequences = 
            new ArrayList<SequenceAbstraction<Integer>>(10);
        sequences.add(sasvali);
        sequences.add(sasvalsi);
        sequences.add(sasvhsi);
        sequences.add(sasvlhsi);  
        sequences.add(sasvtsi); 
        sequences.add(sasvri);
        sequences.add(sasvriu);
        // Memorizable...
        sequences.add(msasvali);
        sequences.add(msasvalsi);
        sequences.add(msasvhsi);
        sequences.add(msasvlhsi);
        sequences.add(msasvtsi);
        sequences.add(msasvri);
        sequences.add(msasvriu);
        // Synchronised...
        sequences.add(ssasvali);
        sequences.add(ssasvalsi);
        sequences.add(ssasvhsi);
        sequences.add(ssasvlhsi);
        sequences.add(ssasvtsi);
        sequences.add(ssasvri);
        sequences.add(ssasvriu);
        
        out.heading("Created sequence abstractions");

        out.println("(implementations were copied for the memorizable sequences)").println();
        
        for (int i = 0; i < sequences.size(); i++) {
            Sequence<Integer> sequence = sequences.get(i);
            out.println("Sequence ", i, ": ", sequence, " -> ", sequence.current());
        }
        out.println();

        //
        // Testing
        //
        
        out.heading("Advancing sequences");
        
        // true = create memento, false = use memento
        boolean create = true;
        // Map storing the mementos
        Map<MemorizableSequence<Integer>, 
            SequenceMemento<Integer>> mementos = createMap(); 
        
        for (int i = 0, n = (int)(2.7 * values.length), l = 1; i < n; i++, l++) {
            // +1 because the initial value is the first sequence value...
            out.print("Original test value: ", l % values.length, " = ", values[l % values.length]);
            if (l % values.length == 0) {
                out.print(" - test value restart");
            }
            Boolean created = null;
            out.println().println();
            for (int j = 0; j < sequences.size(); j++) {
                Sequence<Integer> sequence = sequences.get(j);
                Object value = sequence.next();
                out.println("Sequence ", j, ": ", sequence, " -> ", value);
                if (i % (n / 3) == (n / 3) - 2) {
                    sequence.reset();
                    out.println("Reset: ", j, ": ", sequence, " -> ", sequence.current());
                    l = 0; // will be increased by the loop...
                }
                if ((i % (values.length * 0.4) == 6) && (sequence instanceof MemorizableSequence)) {
                    MemorizableSequence<Integer> ms = 
                        (MemorizableSequence<Integer>)sequence;
                    if (create) {
                        ok = create(out, mementos, ms) && ok;
                        created = true;
                    } else {
                        ok = use(out, mementos, ms) && ok;
                        created = false;
                    }
                }
            }
            if (created != null) {
                create = !created;
            }
            out.println();
        }
        out.println();
        
        out.heading("Bridge tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Creates a new map to store (memorizable sequence, memento) pairs.
      * 
      * @param  <E> The type of values delivered by applicable sequences.
      * @return A new map to use; never null.
      */
    private final static <E> Map<MemorizableSequence<E>,SequenceMemento<E>> createMap() {
        return new HashMap<MemorizableSequence<E>,SequenceMemento<E>>();
    }
    
    /**
      * Creates a new memento acquired from {@code sequence}.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  mementos The map to store the created memento; never null.
      * @param  sequence The sequence to create the memento; never null.
      * @return True upon success, false if not.
      */
    private final static <E> boolean create(Log out, Map<MemorizableSequence<E>,SequenceMemento<E>> mementos, MemorizableSequence<E> sequence) {
        SequenceMemento<E> sm = sequence.getMemorizableState();
        out.println("Created memento: ", sequence.toString(SequencePolicy.ID), " -> ", sm);
        mementos.put(sequence, sm);
        if (!sequence.current().equals(sm.getSequence().current())) {
            out.error("Current value different between sequence and created memento: ", sequence.current(), " != ", sm.getSequence().current());
            return false;
        }
        return true;
    }

    /**
      * Updates {@code sequence} using the relevant memento stored in
      * {@code mementos}.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  mementos The map to storing the created memento; never null.
      * @param  sequence The sequence to be updated; never null.
      * @return True upon success, false if not.
      */
    private final static <E> boolean use(Log out, Map<MemorizableSequence<E>,SequenceMemento<E>> mementos, MemorizableSequence<E> sequence) {
        try {
            SequenceMemento<E> sm = mementos.get(sequence);
            sequence.setMemorizableState(sm);
            out.println("Used memento: ", sm, " -> ", sequence.toString(SequencePolicy.ID));
            if (!sequence.current().equals(sm.getSequence().current())) {
                out.error("Current value different for between updated sequence and memento: ", sequence.current(), " != ", sm.getSequence().current());
                return false;
            }
            return true;
            
        } catch (MemorizableException me) {
            out.error("Error occurred while using memento").error(me);
            return false;
        }
    }

    /**
      * Test for equality including support for proxies. 
      * 
      * @param  a The first object; cannot be null.
      * @param  b The second object; cannot be null.
      * @return True if equal, false if not.
      */
    private final static boolean equals(Object a, Object b) {
        Object aa = isProxyClass(a.getClass()) ? a.toString() : a,
               bb = isProxyClass(b.getClass()) ? b.toString() : b;
        return aa.equals(bb);       
    }
    
} 
