/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Throwables.toThrowableType;
import dk.rode.thesis.memento.MemorizableException;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.memento.SequenceMemento;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>memorizable sequence abstraction</i> refines a sequence 
  * {@link SequenceAbstraction abstraction} so it can save and restore 
  * its internal implementation in form of {@link SequenceMemento mementos}. <p>
  * 
  * Mementos cannot be created or set until a generator has been 
  * set, either at {@link #MemorizableSequenceAbstraction(SequenceValueGenerator)
  * construction} time, any time before sequence use via the
  * {@link #setGenerator(SequenceValueGenerator)} method.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence.
  */
@Participant("RefinedAbstraction")
public class MemorizableSequenceAbstraction<E> extends SequenceAbstraction<E> 
  implements MemorizableSequence<E>, Sequence<E> {

    /**
      * Constructor. <p>
      * 
      * A value {@link SequenceValueGenerator generator} <b>must</b> explicitly 
      * be set using the {@link #setGenerator(SequenceValueGenerator)} 
      * method before this sequence is valid!  
      */
    public MemorizableSequenceAbstraction() {
        super();
    }
    
    /**
      * Constructor, which creates this memorizable sequence abstraction 
      * to use the value {@link SequenceValueGenerator generator}
      * supplied as {@code generator}. <p>
      *
      * The {@link #setGenerator(SequenceValueGenerator)} must <b>not</b>
      * be invoked for this sequence.
      * 
      * @param  generator The generator to use; cannot be null.
      * @throws NullPointerException If {@code generator} is null.
      */
    public MemorizableSequenceAbstraction(@Participant("Implementation") SequenceValueGenerator<E> generator) {
        super(generator);
    }
    
    /**
      * Copy constructor. <p>
      * 
      * If {@code sequence} does not have a value {@link SequenceValueGenerator 
      * generator} set, a generator <b>must</b> explicitly 
      * be set for this sequence using the 
      * {@link #setGenerator(SequenceValueGenerator)} method before 
      * this sequence is valid!  
      * 
      * @param  sequence The (memorizable) sequence abstraction to copy; 
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public MemorizableSequenceAbstraction(SequenceAbstraction<E> sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public MemorizableSequenceAbstraction<E> copy() {
        return new MemorizableSequenceAbstraction<E>(this);
    }
    
    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public SequenceMemento<E> getMemorizableState() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        // This will copy the implementation!...
        return new SequenceMemento<E>(this);
    }

    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public void setMemorizableState(SequenceMemento<E> memento) 
        throws IllegalStateException, MemorizableException {
        
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        try {
            // ISSUE: This does not generate a warning using NetBeans 5.5.1 and 
            //        the Java 1.6.0_01 compiler, but does so in Eclipse 3.3.
            @SuppressWarnings("unchecked") MemorizableSequenceAbstraction<E> sequence =
                (MemorizableSequenceAbstraction<E>)memento.getSequence();
            // No need to copy the implementation because it was already
            // copied when the memento was created...
            this.generator = sequence.generator;
            // Set the current value...
            int number = sequence.generator.number();
            this.current = (number < 1) 
                         ? (E)this.generator.first() // NetBeans complains otherwise
                         : this.generator.get(number); 
                
            LogFactory.getLog(this).println("Updated ", this.toString(SequencePolicy.IDENTITY), " using ", memento);
            
        } catch (Exception e) {
            throw toThrowableType(e, MemorizableException.class);
        }
    }

}
