/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import java.io.Serializable;
import java.util.Collection;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>sequence value collection</i> is a finite collection of non-null
  * sequence values that can be used as the implementation for a sequence 
  * {@link SequenceAbstraction abstraction}. <p>
  * 
  * Besides the inherited properties from {@link SequenceValueGenerator},
  * any sequence value collection will contain serializable values only.
  * This is a requirement in order to achieve <i>deep cloning</i> in a
  * generic fashion. <p>  
  * 
  * <b>Implementation notes</b>:<br>
  * By demanding that the type of contained values in this collection
  * are {@link java.io.Serializable serializable}, i.e. 
  * {@code <E extends Serializable>}, deep copying becomes easy for 
  * any value collection implementation. To copy a given 
  * collection, the {@link java.util.Collection java.util.Collection} 
  * type determined by the type parameter {@code C} is simply serialized, 
  * then deserialized to yield a new instance with new value instances as 
  * well. The structure of the values in the original collection are 
  * adhered to as serialization maintains relative object identity in 
  * serialized object graphs. Hence, singleton objects such as enumerations 
  * will not be copied, while actual values will be. <p>
  * 
  * It also ensures that standard value types such as {@code Integer},
  * {@code Long}, {@code Date}, etc, can be used as values, which
  * would <b>not</b> be the case if the the type parameter bound on
  * {@code E} had been {@link java.lang.Cloneable java.lang.Cloneable} 
  * instead. However, the work in copying a value collection is proportional 
  * with the number of values contained, and is thus more expensive the
  * larger the collection is! But, this is no different than if the
  * values had been cloned. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of serializable values stored in this value collection.
  * @param  <C> The internal collection type used by this value collection. 
  */
@Participant("Implementor")
public abstract class SequenceValueCollection<E extends Serializable,C extends Collection<E>>
           implements SequenceValueGenerator<E> {

    /**
      * The type of the sequence values stored in this collection. <p>
      * 
      * Never null.
      */
    protected final Class<E> type;
    /**
      * The actual collection used to store the sequence values. <p>
      * 
      * Never null.
      */
    protected final C values;
    /**
      * The number of the the last value delivered by either 
      * {@link #first()}, {@link #get()}, or {@link #get(int)}. <p>
      * 
      * Range: 1 (0) <= number <= {@link #size()} <p>
      */
    protected int number;
    
    /**
      * Creates this collection of values to use the actual sequence 
      * values supplied in {@code values}, each having the type supplied
      * as {@code type}. <p>
      *
      * The next value to return by {@link #get()} is the value that 
      * is the {@code number}'th value in this collection. It is the
      * responsibility of sub-classes to initialises this value collection
      * accordingly. <p> 
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * collection and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * {@code values} must not contain null entries. If it does, 
      * the sequence {@link SequenceAbstraction abstraction} using this 
      * collection will generate an internal error when that entry is 
      * returned by this collection!   
      *
      * @param  type   The type of values stored in this collection; 
      *                cannot be null.
      * @param  values The actual {@code Collection} instance storing the 
      *                sequence values; cannot be null, empty, or contain 
      *                null values.
      * @param  number The number pointing the the {@code number}'th value
      *                to be returned by the next call to {@link #get()}.                
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, or 
      *         contain a null value, or if {@code number} is illegal.
      */
    protected SequenceValueCollection(Class<E> type, C values, int number) {
        if ((this.type = type) == null) {
            throw new NullPointerException("Class cannot be null");
            
        // Must have values...
        } else if ((this.values = values) == null) {
            throw new NullPointerException("Collection cannot be null");
            
        // Values cannot be empty because we need a current non-null value
        // in any sequence...
        } else if (this.values.size() == 0) {
            throw new IllegalArgumentException("Collection is empty");
            
        // Check number (allowed to be zero on construction)...
        } else if (((this.number = number) < 0) || (this.number > this.values.size())) {
            throw new IllegalArgumentException("Illegal number: " + number);
        }
    }

    /**
      * Ensures the the number of a given value generated by this
      * values is legal. <p>
      * 
      * A valid range is: 1 <= number <= {@link #size()}
      * 
      * @param  number The number to check.
      * @param  fail   If true, an exception is thrown if {@code number}
      *                is illegal, otherwise false is returned. If legal,
      *                true is always returned.
      * @return If {@code fail} is false, true if {@code number} is 
      *         valid, false if not.                
      * @throws IllegalArgumentException If {@code fail} is true and if 
      *         {@code number} is illegal.
      */
    protected final boolean checkNumber(int number, boolean fail) {
        // One based...
        if ((number < 1) || (number > this.values.size())) {
            if (fail) {
                throw new IllegalArgumentException("Illegal number: " + number);
            }
            return false;
        }
        return true;
    }
    
    /**
      * 
      */
    public final Class<E> getValueType() {
        return this.type;
    }
    
    /**
      * 
      */
    public int size() {
        return this.values.size();
    }

    /**
      * 
      */
    public int number() {
        return this.number;
    }
    
    /**
      * {@inheritDoc}
      * 
      * The work in copying this collection is proportional with the
      * number of values contained, and is thus more expensive the
      * larger this collection is! <p>
      *
      * @return A copy of this value collection; never null.
      */
    public abstract SequenceValueCollection<E,C> copy();
    
    /**
      * Adds all values in this value collection to the collection supplied
      * as {@code collection}. <p>
      * 
      * The insertion order is determined by the {@code addAll} method
      * on {@code collection}.
      * 
      * @param  collection The collection; cannot be null.
      * @return {@code collection}; never null.
      * @throws NullPointerException If {@code collection} is null.
      */
    public Collection<? super E> toCollection(Collection<? super E> collection) {
        collection.addAll(this.values);
        return collection;
    }
    
    /**
      * Returns the hash code of this value collection. 
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.values.hashCode();
    }

    /**
      * Returns true if {@code object} is a {@code SequenceValueCollection}
      * instance that delivers the same {@link #getValueType() type} 
      * of values as this collection, has the same {@link #size() size},
      * has the same {@link #ordered() ordered}, {@link #sorted() sorted},
      * and {@link #duplicates() duplicate} properties, and contain the
      * same equivalent values adhering to the properties just listed. 
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof SequenceValueCollection)) {
            return false;
        }
        SequenceValueCollection<?,?> svc = (SequenceValueCollection<?,?>)object;
        return ((this.getValueType().equals(svc.getValueType())) &&
                (this.size()       == svc.size())                &&         
                (this.duplicates() == svc.duplicates())          &&
                (this.sorted()     == svc.sorted())              &&
                (this.ordered()    == svc.ordered())             &&
                (this.values.equals(svc.values)));
    }

    /**
      * Returns the string representation of this collection.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        return this.toString(null).toString();
    }

    /**
      * 
      */
    public StringablePolicy<? super SequenceValueGenerator<E>> getStringablePolicy(StringablePolicy<? super SequenceValueGenerator<E>> policy) {
        if (policy == null) {
            policy = SequenceValuePolicy.VERBOSE;
        }
        return policy;
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super SequenceValueGenerator<E>> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }
    
}
