/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Const.EOF;
import static dk.rode.thesis.meta.util.Const.INFINITE;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>sequence value range</i> generates unique {@code Integer}
  * values within a given range determined at 
  * {@link #SequenceValueRange(int, int) construction} time. <p>
  * 
  * A sequence value range is {@link #ordered() ordered},
  * {@link #sorted() sorted}, and do not contain {@link #duplicates() 
  * duplicates}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteImplementor")
public class SequenceValueRange implements SequenceValueGenerator<Integer> {

    /**
      * The current generated value corresponding to 
      * a number of {@code value} - {@link #minimum}. <p>
      * 
      * Range is: {@link #minimum} <= number [<= {@link #maximum}, if {@code maximum != INFINITE}]
      */
    private int value;
    /**
      * The minimum value in this range. 
      */
    private final int minimum;
    /**
      * The maximum value in this range. <p>
      * 
      * Will be {@link Const#INFINITE} if the range has no maximum.
      */
    private final int maximum;
    
    /**
      * Constructor, which creates this <i>sequence value range</i> 
      * to generate {@code Integer} values in the range 
      * {@code [minimum, ..[}, i.e. no fixed size.
      * 
      * @param  minimum The minimum value to generate.
      */
    public SequenceValueRange(int minimum) {
        this(minimum, INFINITE, minimum);
    }
    
    /**
      * Constructor, which creates this <i>sequence value range</i> 
      * to generate {@code Integer} values in the range 
      * {@code [minimum, maximum]}.
      * 
      * @param  minimum The minimum value to generate.
      * @param  maximum The maximum value to generate; must be >= {@code minimum}. 
      * @throws IllegalArgumentException If {@code minimum} > {@code maximum}. 
      */
    public SequenceValueRange(int minimum, int maximum) {
        this(minimum, maximum, minimum);
    }

    /**
      * Constructor, which creates this <i>sequence value range</i> 
      * to generate {@code Integer} values in the range 
      * {@code [minimum, maximum]}.
      * 
      * @param  minimum The minimum value to generate.
      * @param  maximum The maximum value to generate; must be >= {@code minimum}
      *                 if not {@link Const#INFINITE}.
      * @param  value   The current value, offset by {@code minimum}. 
      * @throws IllegalArgumentException If {@code minimum} > {@code maximum},
      *         or if {@code value} is illegal. 
      */
    private SequenceValueRange(int minimum, int maximum, int value) {
        if ((maximum != INFINITE) && (minimum > maximum)) {
            throw new IllegalArgumentException("Minimum cannot be larger than maximum");
        } else if ((value < minimum) || ((maximum != INFINITE) && (value > maximum))) {
            throw new IllegalArgumentException("Illegal value: " + value);
        }
        this.minimum = minimum;
        this.maximum = maximum;
        this.value = value;
    }

    /**
      * Ensures the the number of a given value generated by this
      * value range is legal. If not, and{@code fail} is true, an 
      * illegal argument exception is thrown. <p>
      * 
      * A valid range is {@link #minimum} <= number [<= {@link #maximum}, if {@code maximum != INFINITE}]
      * 
      * @param  number The number to check.
      * @param  fail   If true, an exception is thrown if {@code number}
      *                is illegal, otherwise {@link Const#EOF} is returned. 
      *                If legal, a number is always returned.
      * @return The corresponding value to {@code number}, or {@code EOF}.
      * @throws IllegalArgumentException If {@code fail} is true and if 
      *         {@code number} is illegal.
      */
    private final int checkNumber(int number, boolean fail) {
        int value = number + this.minimum - 1;
        // One based...
        if ((value < this.minimum) || ((this.size() != INFINITE) && (value > this.maximum))) {
            if (fail) {
                throw new IllegalArgumentException("Illegal number: " + number);
            }
            return EOF;
        }
        return value;
    }
    
    /**
      * 
      */
    public Class<Integer> getValueType() {
        return Integer.class;
    }

    /**
      * 
      */
    public int size() {
        return (this.maximum == INFINITE) ? INFINITE : this.maximum - this.minimum + 1; // inclusive
    }
    
    /**
      * 
      */
    public int number() {
        return this.value - this.minimum;
    }
    
    /**
      * 
      */
    public Integer first() {
        this.value = this.minimum;
        return this.value++;
    }
    
    /**
      * 
      */
    public Integer get() {
        Integer value = this.value++;
        if ((this.maximum != INFINITE) && (this.value > this.maximum)) { // inclusive
            this.value = this.minimum;
        } 
        return value;
    }

    /**
      * 
      */
    public Integer get(int number) {
        this.value = this.checkNumber(number, true);
        return this.get(); // this.value++;
    }

    /**
      * 
      */
    public int numberOf(Integer value, int fromNumber) {
        // Out of range?...
        if ((value < this.minimum) || ((this.maximum != INFINITE) && (value > this.maximum))) {
            return EOF;
        } else if ((this.checkNumber(fromNumber, false)) == EOF) {
            return EOF;
        }
        int number = value - this.minimum + 1;
        return number < fromNumber ? EOF : number;
    }
    
    /**
      * 
      */
    public boolean ordered() {
        return true;
    }
    
    /**
      *
      */
    public boolean sorted() {
        return true;
    }
    
    /**
      * 
      */
    public boolean duplicates() {
        return false;
    }

    
    /**
      * Returns the hash code of this value range. 
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.minimum;
    }

    /**
      * Returns true if {@code object} is a {@code SequenceValueRange}
      * instance that uses the same value range as this range, false
      * if not. 
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof SequenceValueRange)) {
            return false;
        }
        SequenceValueRange svr = (SequenceValueRange)object;
        return ((this.minimum == svr.minimum) && (this.maximum == svr.maximum));
    }
    
    /**
      * {@inheritDoc}
      */
    public SequenceValueRange copy() {
        return new SequenceValueRange(this.minimum, this.maximum, this.value);
    }

    /**
      * Returns the string representation of this value range.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return this.toString(null).toString();
    }

    /**
      *
      */
    public StringablePolicy<? super SequenceValueGenerator<Integer>> getStringablePolicy(StringablePolicy<? super SequenceValueGenerator<Integer>> policy) {
        if (policy == null) {
            policy = SequenceValuePolicy.IDENTITY;
        }
        return policy;
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super SequenceValueGenerator<Integer>> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }
    
}
