/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.InitialisableExpression;
import dk.rode.thesis.interpreter.TerminalExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * An <i>abstract expression builder</i> implements the basic
  * traits of the {@link ExpressionBuilder} interface, while
  * ensuring that the constructed {@link TerminalExpression terminal} 
  * expressions will all manipulate the same {@link Sequence sequence} 
  * instance. <p>
  * 
  * Any builder inheriting this class furthermore uses a {@link Context context} 
  * to register all unique instances of constructed {@link VariableExpression
  * variables} and store all created {@link ConstantExpression constants}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link #initialiseExpressions(Expression)} method is a 
  * <i>visitation method</i> used to visit all expressions in the
  * expression syntax-tree. See also the 
  * {@link dk.rode.thesis.visitor Visitor} pattern.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of constructed 
  *             expressions produces.
  */
@Participant("Builder")
public abstract class AbstractExpressionBuilder<E> 
           implements ExpressionBuilder<E> {

    /**
      * The {@link Sequence sequence} manipulated by terminal 
      * expressions. <p>
      * 
      * Never null.
      */
    protected final Sequence<? extends E> sequence;
    /**
      * The context used to register created 
      * {@link VariableExpression variables} and
      * store created {@link ConstantExpression constants}. <p>
      * 
      *  Never null.
      */
    protected final Context context;
    /**
      * The root expression returned by {@link #getRootExpression()}. <p>
      * 
      * Never null.
      */
    protected final FlowExpression<E> root;
    
    /**
      * Constructor. <p>
      * 
      * A local {@link Context context} is used to register created 
      * {@link VariableExpression variables} and store 
      * {@link ConstantExpression constants}.
      * 
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.              
      * @see    #getContext()    
      */
    public AbstractExpressionBuilder(Sequence<? extends E> sequence) {
        this(new Context(), sequence);
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link Context context} supplied as {@code context} is
      * used to register created {@link VariableExpression variables}
      * and store {@link ConstantExpression constants}.
      * 
      * @param  context  The context to use; cannot be null.
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #getContext()                  
      */
    public AbstractExpressionBuilder(Context context, Sequence<? extends E> sequence) {
        if ((this.context = context) == null) {
            throw new NullPointerException("Context cannot be null");
        } else if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        this.root = new FlowExpression<E>();
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Context context} and {@link Sequence sequence}
      * as used by {@code builder} will be used by this builder. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public AbstractExpressionBuilder(AbstractExpressionBuilder<E> builder) {
        this.context = builder.context;
        this.sequence = builder.sequence;
        this.root = new FlowExpression<E>();
    }
    
    /**
      * Return the {@link Context context} used by this
      * builder.
      *  
      * @return The context; never null.
      */
    public Context getContext() {
        return this.context;
    }
    
    /**
      * 
      */
    public Expression<E> getRootExpression() throws ExpressionException {
        return initialiseExpressions(this.root);
    }
 
    /**
      * Initialises all {@link InitialisableExpression initialisable} expressions
      * reachable from the expression supplied as {@code expression} that has
      * not already been initialised. 
      *  
      * @param  <V>        The type of value the evaluation of {@code expression} produces. 
      * @param  expression The expression (root); cannot be null.
      * @return {@code expression}; never null.
      * @throws NullPointerException If {@code expression} is null.
      * @throws ExpressionException If the initialisation fails.
      */
    protected static <V> Expression<V> initialiseExpressions(Expression<V> expression) throws ExpressionException {
        for (Expression<?> e : expression.operands()) {
            initialiseExpressions(e);
        }
        if (expression instanceof InitialisableExpression) {
            InitialisableExpression<?> ie = (InitialisableExpression<?>)expression;
            if (!ie.isInitialised()) {
                ie.initialise();
            }
        }
        return expression;
    }
    
    /**
      * 
      */
    public void buildExpression(Expression<? extends E> expression) throws ExpressionException {
        this.root.add(expression);
        LogFactory.getLog(this).println("Added expression to root expression: ", expression);
    }
    
    /**
      * 
      */
    public Sequence<? extends E> getSequence() {
        return this.sequence;
    }
    
    // Variables and constants

    /**
      * 
      */
    public <V> VariableExpression<V> buildVariableExpression(Class<V> type, String name) throws ExpressionException {
        VariableExpression<V> ve = null;
        if (this.context.exist(name)) {
            ve = this.context.getVariable(type, name);
            LogFactory.getLog(this).println("Reusing VARIABLE expression: ", ve);
        } else {
            this.context.register(ve = new VariableExpression<V>(type, name));
            LogFactory.getLog(this).println("Building VARIABLE expression: ", ve);
        }
        return ve;
    }

    /**
      * 
      */
    public <V> Expression<V> buildConstantExpression(Class<V> type, V value) throws ExpressionException {
        VariableExpression<V> ce = new ConstantExpression<V>(type, value);
        if (this.context.exist(ce)) {
            ce = this.context.getVariable(type, ce.name);
            LogFactory.getLog(this).println("Reusing CONSTANT expression: ", ce);
        } else {
            this.context.assign(ce, value);
            LogFactory.getLog(this).println("Building CONSTANT expression: ", ce);
        }
        return ce;
    }
    
    /**
      * Returns the hash code of this builder.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.sequence.hashCode();
    }
    
    /**
      * Returns true if {@code object} is an instance of this
      * class and uses the same {@link Sequence sequence} as
      * this builder, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
        } else if (this.getClass().equals(object.getClass())) {
            AbstractExpressionBuilder<?> asb = (AbstractExpressionBuilder<?>)object;
            return this.sequence.equals(asb.sequence);
        }
        return false;
    }
 
    /**
      * Returns the string representation of this builder.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY))
                 .append("; context: ").append(this.context)
                 .append(']').toString();
    }
    
}
