/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.InitialisableExpression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.prototype.StrictCopyable;

/**
  * An <i>expression builder</i> construct various {@link Expression}
  * types having no bounds on the values produced by the
  * constructed expressions. <p>
  * 
  * The constructed <i>root expression</i> can be fetched using
  * the {@link #getRootExpression()} method, and expressions
  * can be added to it using the {@link #buildExpression(Expression)}
  * method. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @param  <E> The main type of values the constructed expressions 
  *             operate on and/or produces.
  */
@ParticipantUsage(value = "Product", type = Expression.class)
@Participant("Builder")
public interface ExpressionBuilder<E> extends StrictCopyable<ExpressionBuilder<E>> {

    /**
      * Returns the <i>root expression</i> constructed by this builder. <p>
      * 
      * Expressions are added to the root expression using the
      * {@link #buildExpression(Expression)} method. Once the
      * root expression has been returned once, additional
      * expressions <b>cannot</b> be added to it! <p>
      * 
      * All {@link InitialisableExpression initialisable} expressions
      * contained in the root expression will have been 
      * {@link InitialisableExpression#initialise() initialised},
      * if not already.
      * 
      * @return The root expression, completely initialised; never null.
      * @throws ExpressionException If the initialisation fails.
      */
    public Expression<E> getRootExpression() throws ExpressionException;
    
    /**
      * Adds the constructed {@link Expression expression} supplied
      * as {@code expression} to the {@link #getRootExpression() root}
      * expression constructed by this builder. <p>
      * 
      * If {@code expression} is {@link InitialisableExpression initialisable}
      * but not {@link InitialisableExpression#isInitialised() initialised},
      * it will be initialised when the root expression is returned.  
      * 
      * @param  expression The expression to add; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      * @throws ExpressionException If the adding fails in case the
      *         root expression has already been fetched.
      */
    public void buildExpression(Expression<? extends E> expression) throws ExpressionException;
    
    /**
      * Returns the {@link Sequence sequence} manipulated by the
      * terminal expressions constructed by this builder.
      * 
      * @return The sequence; never null.
      */
    public Sequence<? extends E> getSequence();
    
    //
    // Non Terminal Expressions
    //

    // Boolean expressions
    
    /**
      * Builds a new <i>short-circuit</i> AND expression.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<Boolean> buildAndExpression(Expression<Boolean> first, Expression<Boolean> second);
    
    /**
      * Builds a new non <i>short-circuit</i> AND expression.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<Boolean> buildNonShortCircuitAndExpression(Expression<Boolean> first, Expression<Boolean> second);
    
    /**
      * Builds a new <i>short-circuit</i> OR expression.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<Boolean> buildOrExpression(Expression<Boolean> first, Expression<Boolean> second);
    
    /**
      * Builds a new non <i>short-circuit</i> OR expression.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<Boolean> buildNonShortCircuitOrExpression(Expression<Boolean> first, Expression<Boolean> second);
    
    /**
      * Builds a new NOT expression.
      * 
      * @param  expression The expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public Expression<Boolean> buildNotExpression(Expression<Boolean> expression);
    
    /**
      * Builds a new EQUAL expression.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<Boolean> buildEqualExpression(Expression<?> first, Expression<?> second);
        
    // Flow, break, conditional, and assignment
    
    /**
      * Builds a new {@link FlowExpression#isInitialised() initialised}
      * FLOW expression.
      * 
      * @param  expressions The expressions, in order, to be 
      *                     associated with the returned expression;
      *                     cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If {@code expressions} is null, or
      *         contain a null entry.
      * @throws ExpressionException If the building fails.         
      */
    public Expression<E> buildInitialisedFlowExpression(Expression<? extends E>... expressions) throws ExpressionException;
    
    /**
      * Builds a new {@link FlowExpression#isInitialised() uninitialised}
      * FLOW expression.
      * 
      * @return The constructed expression; never null.
      */
    public FlowExpression<E> buildFlowExpression();
    
    /**
      * Builds a new BREAK expression.
      * 
      * @param  expression The target expression, if any; can be null.
      * @return The constructed expression; never null.
      */
    public Expression<E> buildBreakExpression(TypedExpression<E> expression);
    
    /**
      * Builds a new CONDITIONAL expression.
      * 
      * @param  condition The condition expression; cannot be null.
      * @param  first     The first expression operand; cannot be null.
      * @param  second    The second expression operand; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      */
    public Expression<E> buildConditionalExpression(Expression<Boolean> condition, Expression<? extends E> first, Expression<? extends E> second);

    /**
      * Builds a new ASSIGNMENT expression.
      * 
      * @param  variable   The variable to be assigned the result of the
      *                    evaluation of {@code expression}; cannot be null,
      *                    or be a {@link #buildConstantExpression(Class, Object) constant}.      
      * @param  expression The expression to deliver the variable value when
      *                    evaluated; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code variable} is a constant.                   
      */
    public Expression<E> buildAssignmentExpression(VariableExpression<E> variable, Expression<? extends E> expression);
 
    // Variables and constants

    /**
      * Builds a new VARIABLE expression.
      * 
      * @param  <V>  The type of value the constructed expression 
      *              produces.
      * @param  type The type of the value represented by 
      *              the constructed expression; cannot be null.
      * @param  name The name of the constructed expression; cannot be 
      *              null or empty.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code name} is illegal.
      * @throws ExpressionException If the building fails.
      */
    public <V> VariableExpression<V> buildVariableExpression(Class<V> type, String name) throws ExpressionException;
    
    /**
      * Builds a new CONSTANT expression.
      * 
      * @param  <V>   The type of value the constructed expression 
      *               produces.
      * @param  type  The type of the constant represented by the
      *               constructed expression; cannot be null.
      * @param  value The value of the constant; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either expression is null.
      * @throws ExpressionException If the building fails.
      */
    public <V> Expression<V> buildConstantExpression(Class<V> type, V value) throws ExpressionException;
    
    //
    // Terminal expressions
    //
    
    /**
      * Builds a new CURRENT expression.
      * 
      * @return The constructed expression; never null.
      */
    public Expression<E> buildCurrentExpression();

    /**
      * Builds a new RESET expression.
      * 
      * @return The constructed expression; never null.
      */
    public Expression<E> buildResetExpression();

    /**
      * Builds a new REVERSE expression.
      * 
      * @param  reverse The expression that will determine if the
      *                 {@link #getSequence() sequence} is <i>tried</i> 
      *                 reversed when the constructed expression is 
      *                 evaluated.
      * @return The constructed expression; never null.                  
      * @throws NullPointerException If {@code reverse} is null.
      */
    public Expression<Boolean> buildReverseExpression(Expression<Boolean> reverse);

    /**
      * Builds a new NEXT expression.
      * 
      * @param  count The expression that will determine the number 
      *               of times {@code next()} will be invoked on 
      *               the {@link #getSequence() sequence} when the 
      *               constructed expression is evaluated.
      * @return The constructed expression; never null.                  
      * @throws NullPointerException If {@code count} is null.
      */
    public Expression<E> buildNextExpression(Expression<? extends Number> count);
    
}
