/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.builder;

import java.util.Map;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.interpreter.AndExpression;
import dk.rode.thesis.interpreter.AssignmentExpression;
import dk.rode.thesis.interpreter.BreakExpression;
import dk.rode.thesis.interpreter.CompareExpression;
import dk.rode.thesis.interpreter.ConditionalExpression;
import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.CurrentExpression;
import dk.rode.thesis.interpreter.EqualExpression;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.Interpreter;
import dk.rode.thesis.interpreter.NextExpression;
import dk.rode.thesis.interpreter.NotExpression;
import dk.rode.thesis.interpreter.OrExpression;
import dk.rode.thesis.interpreter.ResetExpression;
import dk.rode.thesis.interpreter.ReverseExpression;
import dk.rode.thesis.interpreter.SetExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.meta.util.Util;

/**
  * <b>Builder</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Director")
public class Main implements Testable {

    /**
      * Executes the <b>Builder</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Builder tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        boolean ok = true;
       
        // True for verbose class logging...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        this.log(log);
        
        //
        // Create sequences and builders
        //
        
        // Contexts
        Context context1 = new Context();
        Context context2 = new Context();
        Context context3 = null; // fetched from builder
        
        // Integer interpreter..
        Interpreter<Integer> interpreter = new Interpreter<Integer>(Integer.class);
        // Sequence...
        int range = MathFacade.getInstance().getRandomNumber(111);
        int size  = (int)(range * 2.3);
        RangeSequence sequence = new RangeSequence(0, size);
        // Builders using a separate context each...
        StandardComparableExpressionBuilder<Integer> builder1 = 
            new StandardComparableExpressionBuilder<Integer>(context1, sequence);
        CountingComparableExpressionBuilder<Integer> builder2 = 
            new CountingComparableExpressionBuilder<Integer>(context2, sequence);
        TypedComparableExpressionBuilder<Integer> builder3 = 
            new TypedComparableExpressionBuilder<Integer>(Integer.class, builder2);
        context3 = builder3.getContext();
        
        // Test value. Cannot be larger than the sequence size, 
        // because we use the "set" expression!...
        Integer value = MathFacade.getInstance().getRandomNumber(range),
                result1 = null, result2 = null, result3 = null;

        out.heading("Created objects");
        out.println("Interpreter:    ", interpreter);
        out.println("Sequence:       ", sequence);
        out.println("Builder 1:      ", builder1);
        out.println("Builder 2:      ", builder2);
        out.println("Builder 3:      ", builder3).println();
        
        out.println("Sequence range:   0 - ", size);
        out.println("Expression value: ", value);
 
        Expression<Integer> expression = null;
        
        //
        // Builder 1
        //

        String exp1 = "", exp2 = "";
        try {
            out.heading("Builder 1");
            out.println(builder1).println();

            sequence.reset();
            expression = this.create(builder1, Integer.class, value);
            
            out.println().println("Expression (value: ", value, "): ", exp1 = expression.asSymbol(context1.reset())).println(); 
            result1 = interpreter.interpret(context1, expression);
            out.println().println("Result of evaluation: ", result1).println();

            // Reset sequence for reuse... We know we have reversed it as well...
            sequence.reverse();
            sequence.reset();
           
        } catch (Exception e) {
            out.error("Builder 1 error").error(e);
            ok = false;
        }
        
        //
        // Builder 2
        //
 
        try {
            out.heading("Builder 2");
            out.println(builder2).println();
            
            expression = this.create(builder2, Integer.class, value);
            
            out.println().println("Expression (value: ", value, "): ", exp2 = expression.asSymbol(context2.reset())).println(); 
            result2 = interpreter.interpret(context2, expression);
            out.println().println("Result of evaluation: ", result2).println();
            
            // Print count...
            out.println("Expressions created:")
               .println("--------------------")
               .println();
            Map<String,Integer> counts = builder2.getExpressionCount();
            for (String name : counts.keySet()) {
                out.println(name, " = ", counts.get(name), " time(s)");
            }
            
            // Reset sequence for reuse... We know we have reversed it as well...
            sequence.reverse();
            sequence.reset();
            
        } catch (Exception e) {
            out.error("Builder 2 error").error(e);
            ok = false;
        }

        //
        // Builder 3
        //
 
        try {
            out.heading("Builder 3");
            out.println(builder3).println();
            
            expression = this.create(builder3, Integer.class, value);
            
            out.println().println("Expression (value: ", value, "): ", expression.asSymbol(context3.reset())).println(); 
            result3 = interpreter.interpret(context3, expression);
            out.println().println("Result of evaluation: ", result3).println();
            
        } catch (Exception e) {
            out.error("Builder 3 error").error(e);
            ok = false;
        }
        
        // This only works because we know builders 1 and 2 generate the 
        // same symbolic representations (builder 3 includes type information)!
        if (exp1 == null) {
            if (exp2 != null) {
                out.error("Expressions are different:");
                out.println("1: null");
                out.println("2: ", exp2);
                ok = false;
            }
        } else if (!exp1.equals(exp2)) {
            out.error("Expressions are different:");
            out.println("1: ", exp1);
            out.println("2: ", exp2);
            ok = false;
        }
        
        // Test results...
        if (!Util.equals(result1, result2, result3)) {
            out.error("Different evaluation results: ")
               .error("Result 1: ", result1)
               .error("Result 2: ", result2)
               .error("Result 3: ", result3);
            ok = false;
        }

        out.heading("Builder tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * True to enable logging per expression type.
      * 
      * @param log True to enable verbose logging per expression type,
      *            false to disable.
      */
    private void log(boolean log) {
        LogFactory.getLog(AndExpression.class).setActive(log);
        LogFactory.getLog(AssignmentExpression.class).setActive(log);
        LogFactory.getLog(CompareExpression.class).setActive(log);
        LogFactory.getLog(ConditionalExpression.class).setActive(log);
        LogFactory.getLog(ConstantExpression.class).setActive(log);
        LogFactory.getLog(CurrentExpression.class).setActive(log);
        LogFactory.getLog(EqualExpression.class).setActive(log);
        LogFactory.getLog(BreakExpression.class).setActive(log);
        LogFactory.getLog(FlowExpression.class).setActive(log);
        LogFactory.getLog(NextExpression.class).setActive(log);
        LogFactory.getLog(NotExpression.class).setActive(log);
        LogFactory.getLog(OrExpression.class).setActive(log);
        LogFactory.getLog(ResetExpression.class).setActive(log);
        LogFactory.getLog(ReverseExpression.class).setActive(log);
        LogFactory.getLog(SetExpression.class).setActive(log);
        LogFactory.getLog(VariableExpression.class).setActive(log);
        
        LogFactory.getLog(Context.class).setActive(log);
        LogFactory.getLog(Interpreter.class).setActive(log);
        
        LogFactory.getLog(StandardExpressionBuilder.class).setActive(log);
        LogFactory.getLog(StandardComparableExpressionBuilder.class).setActive(log);
        LogFactory.getLog(CountingExpressionBuilder.class).setActive(log);
        LogFactory.getLog(CountingComparableExpressionBuilder.class).setActive(log);
        LogFactory.getLog(TypedExpressionBuilder.class).setActive(log);
        LogFactory.getLog(TypedComparableExpressionBuilder.class).setActive(log);
    }
    
    /**
      * Builds:
      * <pre>
      *   next[20], y = current, (reverse[true] ? next[10], x = current, (x < value ? x : break[y]) : current)
      * </pre>
      * 
      * Depending on {@code builder}, type information may be included.
      * 
      * @param  <E>      The type of value the evaluation of the
      *                  returned expression produces.
      * @param  builder  The builder to use; cannot be null.
      * @param  type     The actual type; cannot be null.
      * @param  value    The value to use in the expression; cannot be null.
      * @return The constructed expression; never null.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the construction fails.
      */
    private final <E extends Comparable<? super E>> Expression<E> create(ComparableExpressionBuilder<E> builder, Class<E> type, E value) throws ExpressionException {

        // Outer flow: next[20], y = current, (reverse[true] ? inner-flow : current)
  
        // next[20]
        builder.buildExpression(
            builder.buildNextExpression(
                builder.buildConstantExpression(
                    Integer.class, 20
                )
            )
        );
        // y = current
        VariableExpression<E> y = builder.buildVariableExpression(type, "y");
        builder.buildExpression(
            builder.buildAssignmentExpression(
                y, 
                builder.buildCurrentExpression()
            )
        );
        
        // Inner flow: next[10], x = current, (x < value ? x : break[y])
        FlowExpression<E> inner = builder.buildFlowExpression();
        // next[10]
        inner.add( 
            builder.buildNextExpression(
                builder.buildConstantExpression(
                    Long.class, 10L
                )
            )
        );
        // x = current
        VariableExpression<E> x = builder.buildVariableExpression(type, "x");
        inner.add( 
            builder.buildAssignmentExpression(
                x, 
                builder.buildCurrentExpression()
            )
        );
        // (x < value ? x : break[y])
        inner.add( 
            builder.buildConditionalExpression(
                builder.buildSmallerThanExpression(
                    x, 
                    builder.buildConstantExpression(type, value)
                ), 
                // This will reuse the actual "x" variable instance already
                // created...
                builder.buildVariableExpression(type, "x"), 
                builder.buildBreakExpression(y)
            )
        );
        
        // Outer flow
        
        // (reverse[true] ? inner-flow : current)
        builder.buildExpression(
            builder.buildConditionalExpression(
                // reverse[true]
                builder.buildReverseExpression(
                    builder.buildConstantExpression(Boolean.class, true)
                ), 
                // inner-flow
                inner, // automatically initialised...
                // current
                builder.buildCurrentExpression()
            )
        );

        return builder.getRootExpression();
    }
    
}
