/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.CompareExpression;
import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.SetExpression;
import dk.rode.thesis.interpreter.TerminalExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.interpreter.CompareExpression.Comparison;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>comparable expression builder</i> construct various {@link Expression}
  * types having a {@code Comparable} bound on the value produced by 
  * the constructed expressions and on the expression operands used. <p>
  *
  * The constructed {@link TerminalExpression terminal} expressions
  * will all manipulate the same {@link Sequence sequence} instance. <p>
  * 
  * A comparable expression builder uses a {@link Context context} to
  * register all unique instances of constructed {@link VariableExpression
  * variables} and store all created {@link ConstantExpression constants}.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The {@code Comparable} type of values the constructed expressions 
  *             operate on and/or produces.
  */
@Participant("ConcreteBuilder")
public class StandardComparableExpressionBuilder<E extends Comparable<? super E>> 
     extends StandardExpressionBuilder<E> 
  implements ComparableExpressionBuilder<E> {
    
    /**
      * Constructor. <p>
      * 
      * A local {@link Context context} is used to register created 
      * {@link VariableExpression variables} and store 
      * {@link ConstantExpression constants}.
      * 
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.              
      * @see    #getContext()    
      */
    public StandardComparableExpressionBuilder(Sequence<? extends E> sequence) {
        super(sequence);
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link Context context} supplied as {@code context} is
      * used to register created {@link VariableExpression variables}
      * and store {@link ConstantExpression constants}.
      * 
      * @param  context  The context to use; cannot be null.
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #getContext()                  
      */
    public StandardComparableExpressionBuilder(Context context, Sequence<? extends E> sequence) {
        super(context, sequence);
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Context context} and {@link Sequence sequence}
      * as used by {@code builder} will be used by this builder. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public StandardComparableExpressionBuilder(StandardComparableExpressionBuilder<E> builder) {
        super(builder);
    }
    
    /**
      * 
      */
    public StandardComparableExpressionBuilder<E> copy() {
        return new StandardComparableExpressionBuilder<E>(this);   
    }

    /**
      * 
      */
    public Expression<Boolean> buildSmallerThanExpression(Expression<E> first, Expression<E> second) {
        Expression<Boolean> expression = new CompareExpression<E>(Comparison.SMALLER, first, second);
        LogFactory.getLog(this).println("Building SMALLER THAN expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildGreaterThanExpression(Expression<E> first, Expression<E> second) {
        Expression<Boolean> expression = new CompareExpression<E>(Comparison.GREATER, first, second);
        LogFactory.getLog(this).println("Building GREATER THAN expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildSetExpression(Expression<E> value) {
        Expression<E> expression = new SetExpression<E>(this.sequence, value, true);
        LogFactory.getLog(this).println("Building SET expression: ", expression);
        return expression;
    }
           
}
