/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.AndExpression;
import dk.rode.thesis.interpreter.AssignmentExpression;
import dk.rode.thesis.interpreter.BreakExpression;
import dk.rode.thesis.interpreter.ConditionalExpression;
import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.CurrentExpression;
import dk.rode.thesis.interpreter.EqualExpression;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.NextExpression;
import dk.rode.thesis.interpreter.NotExpression;
import dk.rode.thesis.interpreter.OrExpression;
import dk.rode.thesis.interpreter.ResetExpression;
import dk.rode.thesis.interpreter.ReverseExpression;
import dk.rode.thesis.interpreter.TerminalExpression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>standard expression builder</i> construct various {@link Expression}
  * types having no bounds on the values produced by constructed expressions. <p>
  * 
  * The constructed {@link TerminalExpression terminal} expressions
  * will all manipulate the same {@link Sequence sequence} instance. <p>
  * 
  * A standard expression builder uses a {@link Context context} to
  * register all unique instances of constructed {@link VariableExpression
  * variables} and store all created {@link ConstantExpression constants}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This builder does not use covariant return types to comply with the 
  * canonical "Gang of Four" examples, but the {@link TypedExpressionBuilder}
  * class does use covariant return types. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ComparableExpressionBuilder
  * 
  * @param  <E> The type of value the evaluation of constructed 
  *             expressions produces.
  */
@Participant("ConcreteBuilder")
public class StandardExpressionBuilder<E> 
     extends AbstractExpressionBuilder<E> 
  implements ExpressionBuilder<E> {
    
    /**
      * Constructor. <p>
      * 
      * A local {@link Context context} is used to register created 
      * {@link VariableExpression variables} and store 
      * {@link ConstantExpression constants}.
      * 
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.              
      * @see    #getContext()    
      */
    public StandardExpressionBuilder(Sequence<? extends E> sequence) {
        super(sequence);
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link Context context} supplied as {@code context} is
      * used to register created {@link VariableExpression variables}
      * and store {@link ConstantExpression constants}.
      * 
      * @param  context  The context to use; cannot be null.
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #getContext()                  
      */
    public StandardExpressionBuilder(Context context, Sequence<? extends E> sequence) {
        super(context, sequence);
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Context context} and {@link Sequence sequence}
      * as used by {@code builder} will be used by this builder. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public StandardExpressionBuilder(StandardExpressionBuilder<E> builder) {
        super(builder);
    }
    
    /**
      * 
      */
    public StandardExpressionBuilder<E> copy() {
        return new StandardExpressionBuilder<E>(this);   
    }
    
    //
    // Non Terminal Expressions
    //

    // Boolean expressions
    
    /**
      * 
      */
    public Expression<Boolean> buildAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        Expression<Boolean> expression = new AndExpression(true, first, second);
        LogFactory.getLog(this).println("Building AND (&&) expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildNonShortCircuitAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        Expression<Boolean> expression = new AndExpression(false, first, second);
        LogFactory.getLog(this).println("Building AND (&) expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        Expression<Boolean> expression = new OrExpression(true, first, second);
        LogFactory.getLog(this).println("Building OR (||) expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildNonShortCircuitOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        Expression<Boolean> expression = new OrExpression(false, first, second);
        LogFactory.getLog(this).println("Building OR (|) expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildNotExpression(Expression<Boolean> expression) {
        expression = new NotExpression(expression);
        LogFactory.getLog(this).println("Building NOT expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildEqualExpression(Expression<?> first, Expression<?> second) {
        Expression<Boolean> expression = new EqualExpression(first, second);
        LogFactory.getLog(this).println("Building EQUAL expression: ", expression);
        return expression;
    }
       
    // Flow, break, conditional, and assignment
    
    /**
      * 
      */
    public Expression<E> buildInitialisedFlowExpression(Expression<? extends E>... expressions) throws ExpressionException {
        FlowExpression<E> fe = new FlowExpression<E>();
        for (Expression<? extends E> e : expressions) {
            fe.add(e);
        }
        fe = fe.initialise(); 
        LogFactory.getLog(this).println("Building FLOW (initialised) expression: ", fe);
        return fe;
    }

    /**
      * 
      */
    public FlowExpression<E> buildFlowExpression() {
        FlowExpression<E> expression = new FlowExpression<E>();
        LogFactory.getLog(this).println("Building FLOW expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<E> buildBreakExpression(TypedExpression<E> expression) {
        if (expression == null){
            expression = new BreakExpression<E>();
            LogFactory.getLog(this).println("Building EXIT expression: ", expression);
        } else {
            expression = new BreakExpression<E>(expression);
            LogFactory.getLog(this).println("Building BREAK expression: ", expression);
        }
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildConditionalExpression(Expression<Boolean> condition, Expression<? extends E> first, Expression<? extends E> second) {
        Expression<E> expression = new ConditionalExpression<E>(condition, first, second); 
        LogFactory.getLog(this).println("Building CONDITIONAL expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildAssignmentExpression(VariableExpression<E> variable, Expression<? extends E> expression) {
        Expression<E> exp = new AssignmentExpression<E>(variable, expression);
        LogFactory.getLog(this).println("Building ASSIGNMENT expression: ", exp);
        return exp;
    }
 
    //
    // Terminal expressions
    //

    /**
      * 
      */
    public Expression<E> buildCurrentExpression() {
        Expression<E> expression = new CurrentExpression<E>(this.sequence, true);
        LogFactory.getLog(this).println("Building CURRENT expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildResetExpression() {
        Expression<E> expression = new ResetExpression<E>(this.sequence, true);
        LogFactory.getLog(this).println("Building RESET expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<Boolean> buildReverseExpression(Expression<Boolean> reverse) {
        Expression<Boolean> expression = new ReverseExpression<E>(this.sequence, reverse, true);
        LogFactory.getLog(this).println("Building REVERSE expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildNextExpression(Expression<? extends Number> count) {
        Expression<E> expression = new NextExpression<E>(this.sequence, count, true);
        LogFactory.getLog(this).println("Building NEXT expression: ", expression);
        return expression;
    }
    
}
