/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An <i>abstract handler chain</i> implements the basic
  * traits of the {@link HandlerChain} interface.  
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @param  <R> The type of requests handled by this chain.
  */
public abstract class AbstractHandlerChain<R> implements HandlerChain<R> {

    /**
      * No-arg constructor.
      */
    protected AbstractHandlerChain() {
        super();
    }

    /**
      * Handles the request supplied as {@code request} by forwarding
      * it to each {@link Handler handler} in this chain until a 
      * handler can handle it, starting from the first registered handler. <p>
      * 
      * The handlers used are those delivered by {@link #getHandlers(Handler)}
      * with a null argument.
      * 
      * @param  request The request to handle. This chain allows it to 
      *                 be null, but individual handlers may not.
      * @return The handler that handled {@code request}, if any; may be null. 
      */
    public Handler<R> handle(R request) {
        return this.forward(request, (Handler<R>)null); // from the start...
    }

    /**
      * 
      */
    public Handler<R> forward(R request, Handler<R> current) {
        Log out = LogFactory.getLog(this);
        for (Handler<R> handler : this.getHandlers(current)) {
            // We pass along an null link to ensure that actual handlers
            // to not recurse into this chain when forwarding requests...
            Handler<R> h = handler.handle(request, null); 
            if (h != null) {
                out.println("Handled request ", ObjectPolicy.ID.toString(this), ": ", request, " -> ", handler);
                return h;
            }
        }
        out.println("Could not handle request ", ObjectPolicy.ID.toString(this), ": ", request);
        return null;
    }
    
    /**
      * 
      */
    public Handler<R> handle(R request, HandlerLink<R> link) {
        Handler<R> handler = this.handle(request);
        if (handler != null) {
            return handler;
        } 
        return link == null ? null : link.forward(request, this);
    }

    /**
      * Returns the string representation of this handler chain.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; size: ").append(this.size()).append(']').toString(); 
    }
    
}
