/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import dk.rode.thesis.flyweight.Character;
import dk.rode.thesis.flyweight.Letter;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;

/**
  * A <i>letter case handler</i> is a handler that can handle
  * {@link Letter letter} characters in either upper or lower
  * case only.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Letter#isUpperCase()
  * @see    Letter#isLowerCase()
  */
@Participant("ConcreteHandler")
public class LetterCaseHandler extends CharacterHandler {

    /**
      * True for upper case only, false for lower case only. 
      */
    private final boolean upperCase;
    
    /**
      * Constructor. <p>
      * 
      * This handler is not registered to any 
      * handler {@link HandlerChain chain} until specifically 
      * registered using the {@link #register(HandlerChain)} 
      * method.
      * 
      * @param upperCase True for upper case only, false for lower case only.
      */
    public LetterCaseHandler(boolean upperCase) {
        super();
        this.upperCase = upperCase;
    }
    
    /**
      * Constructor, which registers this letter handler to the
      * handler chain supplied as {@code chain}.
      * 
      * @param  chain     The handler chain; cannot be null.
      * @param  upperCase True for upper case only, false for lower case only.
      * @throws NullPointerException If {@code chain} is null.
      */
    public LetterCaseHandler(HandlerChain<Character> chain, boolean upperCase) {
        super(chain);
        this.upperCase = upperCase;
    }
    
    /**
      * 
      */
    protected boolean handle(Character character) {
        if (character instanceof Letter) {
            Log out = LogFactory.getLog(this);
            Letter letter = (Letter)character;
            if (this.upperCase) {
                if (letter.isUpperCase()) {
                    out.println("Upper case letter handled: " + character);
                    return true;
                }
            } else if (letter.isLowerCase()) {
                out.println("Lower case letter handled: " + character);
                return true;
            }
        }
        return false;
    }
    
}
