/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.command;

import static dk.rode.thesis.meta.util.Throwables.unwind;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.command.EvilCommand.EvilCommandException;
import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.facade.RandomSequence;
import dk.rode.thesis.memento.MemorizableEnglishAlphabetSequence;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.state.ReversiblePrimeSequence;

/**
  * <b>Command</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Command</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(final Log out, Arguments arguments) {
        out.heading("Command tests...");
        
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        boolean ok = true;
        
        // True to enable class related log output (verbose)...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        
        LogFactory.getLog(NextCommand.class).setActive(log);
        LogFactory.getLog(ResetCommand.class).setActive(log);
        LogFactory.getLog(ReverseCommand.class).setActive(log);
        LogFactory.getLog(CompositeCommand.class).setActive(log);
        LogFactory.getLog(CommandProcessor.class).setActive(true);
                
        //
        // Create sequences
        //
        
        // If copy is true, we use sequence prototypes. Each sequence will 
        // be copied before each test to ensure that the same initial state
        // is used if "copy" is true...
        Boolean copy = arguments.getArgumentValue("copy", Boolean.class);
        if (copy == null) {
            copy = MathFacade.getInstance().getRandomBoolean();
        }
        
        Sequence<?>[] sequences = {
             // unbounded, unique, consistent                
             new LongSequence(),            
             // bounded, not unique, not consistent
             new RandomSequence(10),
             // bounded, consistent, unique
             new EnglishAlphabetSequence(),  
             // unbounded, consistent, unique, reversible, memorizable
             new RangeSequence(),        
             // bounded, consistent, unique, reversible, memorizable             
             new RangeSequence(10, 20),  
             // bounded, consistent, unique, memorizable
             new MemorizableEnglishAlphabetSequence(),
             // bounded, unique, reversible
             new ReversiblePrimeSequence(100),
        };

        // Print instances...
        out.heading("Created sequences");
        for (int i = 0; i < sequences.length; i++) {
            out.println("Sequence ", i, ": ", sequences[i]);
        }

        out.println();
        out.println("Sequences ", copy ? "WILL" : "will NOT", " be copied before use!");
        out.println();
        out.println();
        
        //
        // Tests
        //
        
        // Test 1...
        ok = test(out, "non-composite", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test1(out, sequence, fail);
            }
        }, null, copy); // no forced error

        // Test 2 - same as 1, but with forced undo...
        ok = test(out, "non-composite with forced undo (no fail on undo)", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test1(out, sequence, fail);
            }
        }, false, copy) && ok; // evil command that do not fail on undo

        // Test 3 - same as 1, but with forced undo...
        ok = test(out, "non-composite with forced undo (fail on undo)", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test1(out, sequence, fail);
            }
        }, true, copy) && ok; // evil command that do fail on undo
        
        // Test 4
        ok = test(out, "composite", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test2(out, sequence, fail);
            }
        }, null, copy) && ok; // no forced error

        // Test 5 - same as 4, but with forced undo...
        ok = test(out, "composite with forced undo (no fail on undo)", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test2(out, sequence, fail);
            }
        }, false, copy) && ok; // evil command that do not fail on undo

        // Test 6 - same as 4, but with forced undo...
        ok = test(out, "composite with forced undo (fail on undo)", sequences, new Creator() {
            <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail) {
                return test2(out, sequence, fail);
            }
        }, true, copy) && ok; // evil command that do fail on undo
        
        out.heading("Command tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Executes a given command test for all sequences.
      * 
      * @param  out       The log to use; never null.
      * @param  heading   Title of the test; never null.
      * @param  sequences The sequences to use; never null.
      * @param  creator   The creator used to create and supply the commands to
      *                   test; never null.
      * @param  fail      Flag to determine if the test is supposed to fail or not.
      * @param  copy      True if the sequences should be copied before use, false if not.
      * @return True on success.
      */
    private final static boolean test(Log out, String heading, Sequence<?>[] sequences, Creator creator, Boolean fail, boolean copy) {
        boolean ok = true;
        out.heading("Command test - ", heading);
        for (int i = 0; i < sequences.length; i++) {
            ok = doTest(out, i, sequences[i], creator, fail, copy) && ok;
        }
        return ok;
    }

    /**
      * Executes a given command test for a given sequence.
      *
      * @param  <E>       The type of values delivered by {@code s}.
      * @param  out       The log to use; never null.
      * @param  index     The index of {@code sequence}.
      * @param  s         The sequence to use; never null.
      * @param  creator   The creator used to create and supply the commands to
      *                   test; never null.
      * @param  fail      Flag to determine if the test is supposed to fail or not.
      * @param  copy      True if the sequences should be copied before use, false if not.
      * @return True on success.
      */
    private final static <E> boolean doTest(Log out, int index, Sequence<E> s, Creator creator, Boolean fail, boolean copy) {
        boolean ok = true;
        // If "copy" is true, we always use a copy of the initial 
        // sequence to ensure that we have the same state. If false,
        // we keep working on the same instances...
        Sequence<E> sequence = copy ? s.copy() : s;
        out.println("Sequence ", index, " before execution: ", sequence);
        out.println();
        // For comparison...
        Sequence<?> original = sequence.copy();
        // Create processor to execute the commands...
        CommandProcessor processor = new CommandProcessor();
        // True if executed, false if all was undone because of error...
        boolean executed = true, failed = false;
        CommandProcessingResult<E> result = null;
        try {
            result = processor.execute(creator.create(sequence, fail));
            executed = result.executed;
            
        } catch (CommandException ce) {
            // Execution and undo failed, but see if it was on purpose...
            EvilCommandException evc = unwind(ce, EvilCommandException.class);
            if (evc == null) {
                ok = false;
            }
            failed = true;
            executed = false;
        }
        out.println();
        out.println("Sequence ", index, " after execution: ", sequence);
        if (fail != null) {
            if (!original.current().equals(sequence.current())) {
                // if we undid all, the should be identical!
                if ((!failed) && (!executed)) {
                    out.error("Does not equal original sequence value: ", original.current());
                    ok = false;
                }
            }
            if (!original.state().equals(sequence.state())) {
                // if we undid all, the should be identical!
                if ((!failed) && (!executed)) {
                    if (original.state() == State.START) {
                        out.warn("Does not equal original sequence state, but original state was START: ", sequence.state());
                    } else {
                        out.error("Does not equal original sequence state: ", original.state(), " != ", sequence.state());
                        ok = false;
                    }
                }
            }
        }
        out.println();
        if (ok) {
            out.print("Command execution is as expected:");
            if (failed) {
                out.println(": execution and undo failed!");   
            } else if (executed) {
                out.println(": execution succeeded: ", result.result);   
            } else {
                out.println(": execution failed, but undo succeeded");   
            }
        } else {
            out.error("Command execution failed!");
        }
        out.println();
        out.println();
        return ok;
    }
    
    /**
      * Creates command for the first composite test.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  sequence The sequence to use; never null.
      * @param  fail     Flag to determine if the test is supposed to fail or not.
      * @return The created commands; never null.
      */
    private final static <E> List<Command<E>> test1(Log out, Sequence<E> sequence, Boolean fail) {
        List<Command<E>> commands = new ArrayList<Command<E>>();
        
        // Prototype commands...
        NullCommand<E> nuc = new NullCommand<E>();
        NextCommand<E> nc = new NextCommand<E>(sequence);
        ResetCommand<E> rc = new ResetCommand<E>(sequence);
        ReverseCommand<E> rvc = new ReverseCommand<E>(sequence);
        // Will spawn commands if not reversible...
        @SuppressWarnings("unchecked") // Generic array construction...
        ReverseCommand<E> rvc1 = new ReverseCommand<E>(sequence, nc.copy(), nuc.copy());

        // Assuming integer sequence starting from zero...
        
        // next:    1 1 (column = not reversible)
        commands.add(nc);
        // next     2 2
        commands.add(nc.copy());
        // reset    0 0 
        commands.add(rc);
        // next     1 1
        commands.add(nc.copy());
        // next     2 2
        commands.add(nc.copy());
        // null 
        commands.add(nuc);
        // reverse  2 2
        commands.add(rvc);
        // next     1 3
        commands.add(nc.copy());
        // next     0 4 
        commands.add(nc.copy());
        // next    -1 5
        commands.add(nc.copy());
        // next    -2 6
        commands.add(nc.copy());
        // reverse -2 6
        commands.add(rvc.copy());
        // next    -1 7
        commands.add(nc.copy());
        // next     0 8
        commands.add(nc.copy());
        // next     1 9
        commands.add(nc.copy());
        // next     2 10
        commands.add(nc.copy());
        // reverse  2 1 // spawned next and null commands!
        commands.add(rvc1.copy());
        // reset    0 2
        commands.add(rc.copy());
        // next    -1 3 
        commands.add(nc.copy());
        // next    -2 4
        commands.add(nc.copy());
        // next    -3 5
        commands.add(nc.copy());
        
        if (fail != null) {
            int index = MathFacade.getInstance().getRandomNumber(commands.size() - 1);
            Command<E> c = new EvilCommand<E>(fail); 
            commands.add(index, c);
            out.warn("Added random ", c, " at: ", index + 1, "/", commands.size());
        }
        
        return commands;
    }

    /**
      * Creates command for the second composite test.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  sequence The sequence to use; never null.
      * @param  fail     Flag to determine if the test is supposed to fail or not.
      * @return The created commands; never null.
      */
    @SuppressWarnings("unchecked") // Generic array constructions...
    private final static <E> List<Command<E>> test2(Log out, Sequence<E> sequence, Boolean fail) {
        List<Command<E>> commands = new ArrayList<Command<E>>();
        
        // Prototype commands...
        NextCommand<E> nc = new NextCommand<E>(sequence);
        ResetCommand<E> rc = new ResetCommand<E>(sequence);
        ReverseCommand<E> rvc = new ReverseCommand<E>(sequence);
        NullCommand<E> nuc = new NullCommand<E>();

        // "stay if reversible"...
        CompositeCommand<E> cs1 = new CompositeCommand<E>(nc, rvc, nuc, nc, rvc); // copied by the constructor
        // "minus 1 from start"
        CompositeCommand<E> cs2 = new CompositeCommand<E>(rc, rvc, nc);
        // "plus 5 from here"
        CompositeCommand<E> cs3 = new CompositeCommand<E>(nc, nc, nuc, nc, nc, nc);
        // composite of composites...
        CompositeCommand<E> cs4 = new CompositeCommand<E>(cs1, cs2, cs3, cs2, cs1);
        // Add a copy...
        cs4.addCommand(cs4.copy());
        // Spawn composite if not reversible...
        ReverseCommand<E> rvc1 = new ReverseCommand<E>(sequence, nuc, cs3, nc.copy(), nuc);
        
        // Mix composite and normal commands...
        commands.add(nc);
        commands.add(cs1);
        commands.add(cs2);
        commands.add(nc);
        commands.add(cs3);
        commands.add(cs4);
        commands.add(rvc);
        commands.add(cs3.copy());
        commands.add(cs2.copy());
        commands.add(rvc1);
        commands.add(cs1.copy());

        if (fail != null) {
            int index = MathFacade.getInstance().getRandomNumber(commands.size() - 1);
            Command<E> c = new EvilCommand<E>(fail);
            commands.add(index, c);
            out.warn("Added random ", c, " at: ", index + 1, "/", commands.size());
        }
        
        return commands;
    }
    
}

/**
  * Simple helper class to create a list of commands based on
  * a sequence for the {@link Main#test(Log, Arguments)} method. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
abstract class Creator {
    
    /**
      * No-arg constructor.
      */
    Creator() {
        super();
    }
    
    /**
      * Returns a list of test commands primed with {@code sequence}
      * as the receiver.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  sequence The sequence to use; never null.
      * @param  fail     Flag to determine if the test is supposed to fail or not.
      * @return The created commands; never null.
      */
    abstract <E> List<Command<E>> create(Sequence<E> sequence, Boolean fail);
}

