/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.composite;

import java.util.Date;
import java.util.List;

import dk.rode.thesis.adapter.AdapterStrategy;
import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.facade.RandomSequence;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.DigitSequence;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.observer.DateSequence;
import dk.rode.thesis.prototype.CountdownSequence;
import dk.rode.thesis.prototype.SymbolSequence;
import dk.rode.thesis.singleton.DanishAlphabetSequence;
import dk.rode.thesis.singleton.NorwegianAlphabetSequence;
import dk.rode.thesis.singleton.SmileySequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Composite</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Composite</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Composite tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        //
        // Create sequences and composites
        //
        
        CompositeSequence<CharSequence> cs = new CharSequenceCompositeSequence(CompositeStrategy.BREATH_FIRST);

        cs.addSequence(new EnglishAlphabetSequence());
        cs.addSequence(new DigitSequence());
        cs.addSequence(new SequenceAdapter<Long,String>(new LongSequence(), AdapterStrategy.NUMBER_TO_STRING));

        // Create anonymous composite that adds numbers retrieved from
        // contained sequences...
        CompositeSequence<Number> cs2 = 
            new AbstractCompositeSequence<Number>(CompositeStrategy.DEPTH_FIRST) {
                // Problematic with anonymous classes! Could implement clone, and
                // call super.clone to get a new object, though...
                public CompositeSequence<Number> copy() {
                    return this; 
                }
                public Number current() {
                    long value = 0;
                    for (Sequence<? extends Number> s : this.sequences) {
                        value += s.current().longValue();
                    }
                    return value;
                }
                public Number next() {
                    long value = 0;
                    for (Sequence<? extends Number> s : this.sequences) {
                        value += s.next().longValue();
                    }
                    // Calculate state...
                    this.calculateNextState();
                    
                    return value;
                }
            };
 
        cs2.addSequence(new PrimeSequence(100));   
        cs2.addSequence(new RandomSequence(5));
        
        // Note: by wrapping a composite in an adapter, we loose the ability to
        // test for instanceof! This also influences the strategies used for
        // traversal, printing, etc. A solution could be to let all "decorating
        // sequences" implement a "getSequence" method, but that will expose the
        // internals...
        cs.addSequence(new SequenceAdapter<Number,String>(cs2, AdapterStrategy.NUMBER_TO_STRING));

        // cs3 and cs4 are used for traversal tests because of different 
        // strategies...
        CompositeSequence<CharSequence> cs3 = new CharSequenceCompositeSequence(CompositeStrategy.BREATH_FIRST);
        CompositeSequence<CharSequence> cs4 = new CharSequenceCompositeSequence(CompositeStrategy.DEFAULT); // depth first
        cs4.addSequence(new SequenceAdapter<Date,CharSequence>(new DateSequence(new Date(), 5), AdapterStrategy.ANY_TO_STRING)); 
        cs4.addSequence(NorwegianAlphabetSequence.getInstance());

        cs3.addSequence(cs4);
        cs3.addSequence(new SequenceAdapter<Integer,CharSequence>(new PrimeSequence(97), AdapterStrategy.NUMBER_TO_STRING)); 
        cs3.addSequence(new SymbolSequence());
        cs3.addSequence(new EnglishAlphabetSequence());
        cs3.addSequence(SmileySequence.getSequence());

        // Deep structure...
        cs.addSequence(cs3);
        
        // Notice adapter usage: ok to use super type as "from type", i.e. Number for Long,
        // as well as using sub type in "to type", i.e. String for CharSequence...
        cs3.addSequence(new SequenceAdapter<Long,CharSequence>(new LongSequence(), AdapterStrategy.NUMBER_TO_STRING));
        cs.addSequence(cs3);

        cs.addSequence(new EnglishAlphabetSequence());
        cs.addSequence(DanishAlphabetSequence.Instance);

        // Create composite sequences where each contained sequence restarts
        // at the same time...
        CompositeSequence<CharSequence> cs5 = new CharSequenceCompositeSequence(CompositeStrategy.DEPTH_FIRST);
        // Range with 5 numbers...
        cs5.addSequence(new SequenceAdapter<Integer,CharSequence>(new RangeSequence(0,5), AdapterStrategy.NUMBER_TO_STRING));
        // Five day sequence from "now"...
        cs5.addSequence(new SequenceAdapter<Date,CharSequence>(new DateSequence(new Date(), 5), AdapterStrategy.ANY_TO_STRING)); 
        
        // Define array of sequences to be manipulated anonymously...
        Sequence<?>[] sequences = new Sequence[8];
        sequences[0] = cs;
        sequences[1] = CountdownSequence.getInstance();
        sequences[2] = cs2;
        sequences[3] = new DateSequence();
        sequences[4] = cs3;
        sequences[5] = new CharSequenceCompositeSequence(); // empty
        sequences[6] = cs4;
        sequences[7] = cs5;

        out.heading("Created sequences");

        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> s = sequences[i];
            out.println("Sequence ", i, ":");
            out.println(s);
            out.println("-> ", s.toString(CompositePolicy.VISIBLE));
            out.println();
        }
        
        //
        // Usage
        //
        
        out.heading("Advancing sequences");
        
        for (int i = 0; i < 20; i++) {
            out.println();
            out.println("Iteration ", i, ":");
            for (int j = 0; j < sequences.length; j++) {
                Sequence<?> sequence = sequences[j];
                sequence.next();
                out.println("Advancing sequence ", j, ": ", sequence);
                
                if (i % 8 == 7) {
                    sequence.reset();
                    out.println("Resetting sequence ", j, ": ", sequence);
                    out.println();
                }
            }
        }

        //
        // Print structures
        //
        
        out.heading("Component structure");

        out.println();
        out.println("Note: general problems with adapted, decorated, or proxied");
        out.println("      sequences, because the actual type information is lost.");
        out.println("      Because of this, such decorated composites are not");
        out.println("      discovered via \"instanceof CompositeSequence\"!");
        out.println();
        
        for (int j = 0; j < sequences.length; j++) {
            Sequence<?> sequence = sequences[j];
            out.println();
            print(out, j, "", sequence);
        }

        //
        // Composite traversal
        //
        
        out.heading("Component traversal");

        out.println();
        out.println("Sequences contained in composites, depth-first, breath-first, and combined (instanceof)...");
        for (int j = 0; j < sequences.length; j++) {
            Sequence<?> sequence = sequences[j];
            if (sequence instanceof CompositeSequence) {
                CompositeSequence<?> composite = (CompositeSequence<?>)sequence;
                out.println();
                out.println("Composite ", j, ": ", composite);
                out.println("-> Depth-first:");
                for (Sequence<?> s : composite.getSequences(CompositeStrategy.DEPTH_FIRST, false, null/*any*/)) {
                    out.println("   -> ", s.toString(SequencePolicy.ID));
                }
                out.println("-> Breath-first:");
                for (Sequence<?> s : composite.getSequences(CompositeStrategy.BREATH_FIRST, false, null/*any*/)) {
                    out.println("   -> ", s.toString(SequencePolicy.ID));
                }
                out.println("-> Combined (defaults):");
                for (Sequence<?> s : composite.getSequences(CompositeStrategy.COMBINED, false, null/*any*/)) {
                    out.println("   -> ", s.toString(SequencePolicy.ID));
                }
            }
        }

        out.println();
        out.println();
        out.println("Composites contained in other composites, depth-first (instanceof)...");
        for (int j = 0; j < sequences.length; j++) {
            Sequence<?> sequence = sequences[j];
            if (sequence instanceof CompositeSequence) {
                CompositeSequence<?> composite = (CompositeSequence<?>)sequence;
                out.println();
                out.println("Composite ", j, ": ", composite);
                // This will not find decorated, proxied, or adapted sequences!
                for (Sequence<?> s : composite.getSequences(CompositeStrategy.DEPTH_FIRST, false, Boolean.TRUE/*nodes*/)) {
                    out.println("-> ", s);
                }
            }
        }
        
        out.println();
        out.println();
        out.println("Non-composites contained in other composites, combined (default) strategies (instanceof)...");
        for (int j = 0; j < sequences.length; j++) {
            Sequence<?> sequence = sequences[j];
            if (sequence instanceof CompositeSequence) {
                CompositeSequence<?> composite = (CompositeSequence<?>)sequence;
                out.println();
                out.println("Composite ", j, ": ", composite);
                for (Sequence<?> s : composite.getSequences(CompositeStrategy.COMBINED, false, Boolean.FALSE/*leafs*/)) {
                    out.println("-> ", s);
                }
            }
        }
        
        out.println();
        out.println();
        out.println("Adapted sequences contained in other composites, depth-first (instanceof)...");
        for (int j = 0; j < sequences.length; j++) {
            Sequence<?> sequence = sequences[j];
            if (sequence instanceof CompositeSequence) {
                CompositeSequence<?> composite = (CompositeSequence<?>)sequence;
                out.println();
                out.println("Composite ", j, ": ", composite);
                // Proper type!
                for (SequenceAdapter<?,?> s : composite.getSequences(CompositeStrategy.DEPTH_FIRST, false, null/*all*/, SequenceAdapter.class)) {
                    out.println("-> ", s);
                }
            }
        }
                
        out.heading("Composite tests succeeded!");
        
        return true;
    }

    /**
      * Recursive helper method to print the structure of the (composite)
      * sequence supplied as {@code sequence}.
      * 
      * @param  out      The log to use; never null.
      * @param  i        The index of {@code sequence} among its 
      *                  <i>sibling</i> sequences.
      * @param  indent   The current indent to use; never null.
      * @param  sequence The sequence to print; never null. 
      */
    private final static void print(Log out, int i, String indent, Sequence<?> sequence) {
        out.print(indent, "-> ", i, ": ");
        out.println(sequence);
        if (sequence instanceof CompositeSequence) {
            CompositeSequence<?> cs = (CompositeSequence<?>)sequence;
            List<?> sequences = cs.getSequences();
            for (int j = 0; j < sequences.size(); j++) {
                print(out, j, indent + "      ", (Sequence<?>)sequences.get(j));
            }
        }
    }
    
}
