/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.facade;

import static dk.rode.thesis.meta.util.Const.INFINITE;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Const;

/**
  * An <i>Ackermann sequence</i> calculates the next value of the Ackermann function, 
  * <i>v = A(m, n)</i>, each time {@link #next()} is invoked, using a fixed value for 
  * <i>m</i> and thus increasing <i>n</i>, either indefinitely, or until a maximum 
  * value has been reached. <p> 
  * 
  * The running time of {@code next()} is <b>not</b> consistent each time 
  * invoked! <p>
  * 
  * An Ackermann sequence is {@link Sequence#bounded() bounded} if a maximum
  * value of <i>n</i> is supplied at {@link #AckermannSequence(int, int) construction}
  * time, otherwise unbounded, but always {@link #unique() unique} and 
  * {@link Sequence#consistent() consistent}. <p>
  * 
  * <b>Implementation note:</b><br>
  * This implementation is naive. The stack depth <b>very</b> quickly becomes 
  * an issue! 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("SubsystemClass")
class AckermannSequence extends AbstractSequence<Integer> implements Sequence<Integer> {

    /**
      * The constant value of <i>m</i> for all Ackermann
      * values <i>v = A(m,n)</i> calculated by this sequence. <p>
      */
    private final int m;
    /**
      * The value of <i>n</i> for the Ackermann value <i>v = A(m,n)</i> 
      * {@link #current() currently} set for this sequence. <p>
      * 
      * <i>n</i> will be increased with each call to {@link #next()}
      * until {@link #maximumN} value has been reached, if any, which
      * will cause this sequence to restart. 
      */
    private int n;
    /**
      * The maximum <i>n</i> value to consider. <p>
      * 
      * A value of {@link Const#INFINITE} means no maximum <i>n</i>
      * value.
      */
    private final int maximumN;
    
    /**
      * The Ackermann values calculated so far for {@link #m} 
      * and all <i>n</i> smaller or equal to {@link #n}. <p>
      * 
      * Never null.
      */
    private List<Integer> calculated;
    
    /**
      * Constructor, which uses zero as the value of <i>m</i>. <p>
      * 
      * There is no maximum value of <i>n</i>, i.e. from zero to
      * infinity.  
      */
    public AckermannSequence() {
        this(0);
    }

    /**
      * Constructor, which uses the fixed value of <i>m</i> supplied
      * as {@code m}. <p>
      * 
      * There is no maximum value of <i>n</i>, i.e. from zero to
      * infinity.  
      * 
      * @param  m The fixed <i>m</i> value used for all <i>n</i>
      *           in <i>A(m,n)</i> when calculating the next
      *           value by this sequence; cannot be negative.
      * @throws IllegalArgumentException If {@code m} is negative.           
      */
    public AckermannSequence(int m) {
        this(m, INFINITE);
    }

    /**
      * Constructor, which uses the fixed value of <i>m</i> supplied
      * as {@code m}. <p>
      * 
      * <i>n</i> will run from zero to {@code maximumN}, if 
      * {@code maximumN} is not {@link Const#INFINITE}.
      *  
      * @param  m        The fixed <i>m</i> value used for all <i>n</i>
      *                  in <i>A(m,n)</i> when calculating the next
      *                  value by this sequence; cannot be negative.
      * @param  maximumN The maximum value for <i>n</i> to consider. A value
      *                  of {@link Const#INFINITE} means no maximum. Cannot be
      *                  negative.           
      * @throws IllegalArgumentException If {@code m} or {@code maximumN} are negative.           
      */
    public AckermannSequence(int m, int maximumN) {
        super(false);
        if ((this.m = m) < 0) {
            throw new IllegalArgumentException("'m' cannot be negative");
        } else if ((this.maximumN = maximumN) < INFINITE) {
            throw new IllegalArgumentException("Maximum 'n' cannot be negative");
        }
        this.calculated = new ArrayList<Integer>();
        this.reset();
        this.state = State.START;
    }
    
    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public AckermannSequence(AckermannSequence sequence) {
        super(sequence);
        this.m = sequence.m;
        this.n = sequence.n;
        this.maximumN = sequence.maximumN;
        this.calculated = new ArrayList<Integer>(sequence.calculated);
    }
    
    /**
      * 
      */
    public AckermannSequence copy() {
        return new AckermannSequence(this); 
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return (this.maximumN == INFINITE);
    }

    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * 
      */
    public Integer current() {
        return this.calculated.get(this.n - 1);
    }

    /**
      * 
      */
    public Integer next() {
        if (this.state != State.NORMAL) {
            this.state = State.NORMAL;
        }
        int n = this.n + 1;
        if ((this.maximumN != INFINITE) && (n == this.maximumN)) {
            n = 0;
            this.state = State.RESTART;
        }
        if (this.calculated.size() <= n) {
            this.calculated.add(calculate(this.m, this.n));
        }
        this.n = n;
        return this.current();
    }

    /**
      * 
      */
    public void reset() {
        super.reset();
        this.n = 0;
        this.next();
    }

    /**
      * Returns the constant value of <i>m</i> for all Ackermann
      * values <i>v = A(m,n)</i> calculated by this sequence.
      *  
      * @return The constant value of <i>m</i>.
      */
    public int m() {
        return this.m;
    }
    
    /**
      * The value of <i>n</i> for the Ackermann value <i>v = A(m,n)</i> 
      * {@link #current() currently} set for this sequence.
      *  
      * @return The value of <i>n</i>.
      */
    public int n() {
        return this.n; 
    }
    
    /**
      * Calculates <i>A(m,n)</i> and returns the value.
      * 
      * @param  m The <i>m</i> argument.
      * @param  n The <i>n</i> argument.
      * @return The Ackermann value <i>A(m,n)</i>.
      */
    private final int calculate(int m, int n) {
        if (m == 0) {
            return n + 1; 
        } else if ((m > 0) && (n == 0)) {
            return calculate(m - 1, 1);
        } else {
            return calculate(m - 1, calculate(m, n - 1));
        }
    }

    /**
      * Calculates all <i>A(m,n)</i>.
      * 
      * @param  m  The m value.
      * @param  n  The n value.
      * @param  ms A map storing already calculated values with <i>m</i> and <i>n</i> as keys.
      * @return The Ackermann value <i>A(m,n)</i>.
      */
    final int calculate(int m, int n, Map<Integer,Map<Integer,Integer>> ms) {
        if (m == 0) {
            return n + 1; 
        } 
        int mm = m - 1;
        if ((m > 0) && (n == 0)) {
            // Values for m - 1... 
            Map<Integer,Integer> ns = ms.get(mm);
            if (ns == null) {
                ms.put(mm, ns = new HashMap<Integer,Integer>());
            }
            // (m - 1, 1) value calculated?...
            Integer v = ns.get(1);
            if (v != null) {
                return v;
            }
            v = calculate(mm, 1, ms);
            ns.put(1, v);
            return v;
            
        } else {
            // Values for m..
            Map<Integer,Integer> ns = ms.get(m);
            if (ns == null) {
                ms.put(m, ns = new HashMap<Integer,Integer>());
            }
            // (m, n - 1) value calculated?...  
            Integer v = ns.get(--n); 
            if (v == null) {
                v = calculate(m, n, ms); // = n - 1
                ns.put(n, v);
            }
            // Values for m - 1...
            ns = ms.get(mm);
            if (ns == null) {
                ms.put(mm, ns = new HashMap<Integer,Integer>());
            }
            // (m - 1, A(m, n - 1)) value calculated?...
            Integer w = ns.get(v);
            if (w != null) {
                return w;
            }
            w = calculate(mm, v, ms);
            ns.put(v, w);
            return w;
        }
    }
        
    /**
      * Returns the hash code of this sequence.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.getClass().hashCode() ^ this.m;
    }
    
    /**
      * Returns true if {@code object} is an instance of
      * this sequence type using the same <i>m</i> as 
      * this sequence and currently at the same <i>n</i>,
      * false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof AckermannSequence)) {
            return false;
        }
        AckermannSequence as = (AckermannSequence)object;
        return ((this.m == as.m) && (this.n == as.n)); // ignore calculated... 
    }
    
}
