/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.facade;

import java.util.Random;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>random sequence</i> returns a pseudo integer random number   
  * with each call to {@link #next()} between {@code [0, maximum]}, where
  * {@code maximum} is supplied at construction time. <p>
  * 
  * A random sequence is {@link #bounded() bounded}, but
  * {@link #consistent() inconsistent} and <b>not</b> 
  * {@link #unique() unique}.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    UnboundedRandomSequence
  */
@Participant("SubsystemClass")
public class RandomSequence extends AbstractSequence<Integer> implements Sequence<Integer> {

    /**
      * The random generator used. <p>
      * 
      * Never null.
      */
    private final Random random;
    /**
      * The maximum random number delivered by this sequence.  
      */
    private final int maximum;
    /**
      * The current random sequence number. <p>
      * 
      * Range: [0, {@link #maximum}].
      */
    private int sequence;
    
    /**
      * Constructor.
      * 
      * @param  maximum The maximum random value. Must be larger than 1.
      * @throws IllegalArgumentException If {@code max} is smaller than 1.
      */
    public RandomSequence(int maximum) {
        super();
        if (maximum < 1) {
            throw new IllegalArgumentException("Maximum value must be larger than one");
        }
        // To include max...
        this.maximum = maximum + 1;
        this.random = new Random();
        // Cannot call next because we might be sub-classed...
        this.sequence = this.random.nextInt(this.maximum);
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public RandomSequence(RandomSequence sequence) {
        super(sequence);
        this.maximum  = sequence.maximum;
        this.sequence = sequence.sequence;
        this.random   = new Random();
    }
    
    /**
      * 
      */
    public RandomSequence copy() {
        return new RandomSequence(this); 
    }
    
    /**
      * Returns true.
      *  
      * @return True. 
      */
    public boolean bounded() {
        return true;
    }

    /**
      * Returns false.
      *  
      * @return False. 
      */
    public boolean consistent() {
        return false;
    }

    /**
      * Returns false.
      *  
      * @return False.
      */ 
    public boolean unique() {
        return false;
    }
    
    /**
      * 
      */
    public Integer current() {
        return this.sequence;
    }

    /**
      * 
      */
    public Integer next() {
        if (this.state != State.NORMAL) {
            this.state = State.NORMAL;
        }
        return this.sequence = this.random.nextInt(this.maximum);
    }

    /**
      * 
      */
    public void reset() {
        // Only changes the internal state for show...
        super.reset();
    }

}
