/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.facade;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>unbounded random sequence</i> returns a pseudo integer random number  
  * with each call to {@link #next()} with no upper bound. <p>
  * 
  * It uses an internal sequence to deliver an an integer value that
  * will be multiplied with a random value to generate a sequence value. <p>
  * 
  * An unbounded random sequence is {@link #bounded() unbounded}, 
  * {@link #consistent() inconsistent}, and <b>not</b> 
  * {@link #unique() unique}.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    RandomSequence
  */
@Participant("SubsystemClass")
class UnboundedRandomSequence extends RandomSequence implements Sequence<Integer> {

    /**
      * A <i>sequence</i> to deliver unbounded long values. <p>
      * 
      * Never null.
      */
    private final Sequence<? extends Number> sequence;

    /**
      * No-arg constructor. <p>
      *
      * A {@link LongSequence} is used to deliver the unbounded
      * long values used internally. 
      */
    public UnboundedRandomSequence() {
        this(new LongSequence(-100));
    }
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence to deliver a long value
      *                  that will be multiplied with a random value;
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.                   
      */
    public UnboundedRandomSequence(Sequence<? extends Number> sequence) {
        super(Math.max(100, Math.min(10, sequence.current().intValue())));
        this.sequence = sequence;
        
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public UnboundedRandomSequence(UnboundedRandomSequence sequence) {
        super(sequence);
        this.sequence = sequence.sequence.copy();
    }
    
    /**
      * 
      */
    public UnboundedRandomSequence copy() {
        return new UnboundedRandomSequence(this); 
    }
    
    /**
      * Returns false.
      *  
      * @return False. 
      */
    public boolean bounded() {
        return false;
    }
    
    /**
      * 
      */
    public Integer current() {
        return value(super.current(), this.sequence.current());
    }

    /**
      * 
      */
    public Integer next() {
        return value(super.next(), this.sequence.next());
    }

    /**
      * Calculates the current value of this sequence.
      * 
      * @param  random The random number.
      * @param  number The unbounded number.
      * @return The sequence value.
      */
    private final static int value(int random, Number number) {
        int value = number.intValue();
        return (random * value) * ((random % (value == 0 ? 1 : value) % 2 == 0) ? -1 : 1);
    }
    
    /**
      * 
      */
    public void reset() {
        // Only sets state...
        super.reset();
        // Fine, because this sequence is still inconsistent...        
        if (this.sequence != null) {
            this.sequence.reset();  
        }
    }

}
