/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.EvilCommand;
import dk.rode.thesis.command.NextCommand;
import dk.rode.thesis.command.NullCommand;
import dk.rode.thesis.command.ResetCommand;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;

/**
  * An <i>evil sequence command creator</i> is a <b>test creator</b>
  * that flips the functionality offered by the sequence command 
  * {@link SequenceCommandCreator creator}, and <i>may decide</i> to 
  * return a new {@link EvilCommand evil} command as a default command. <p>
  *
  * This creator can create the following commands based on the 
  * supplied internal sequence state: <ul>
  * 
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
  *      creates a new {@link ResetCommand} instance. <p>
  *       
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
  *      creates a new {@link NextCommand} instance. <p>
  *      
  * <li> Otherwise a {@link NullCommand} <b>or</b> or
  *      an {@link EvilCommand} is returned. <p> 
  * 
  * </ul>
  * 
  * Notice that the regular commands are flipped based on the supplied state! <p>
  * 
  * An evil sequence command creator is stateless and thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * By throwing exceptions in the factory methods, we test the full functionality 
  * of the {@link #getCommand(Sequence, Object)} method. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by created commands (and thus sequences).
  */
@Participant("ConcreteCreator")
public class EvilSequenceCommandCreator<E> extends SequenceCommandCreator<E> {

    /**
      * No-arg constructor.
      */
    public EvilSequenceCommandCreator() {
        super();
    }
    
    /**
      * 
      */
    @Override
    protected Command<E> create(Sequence<E> sequence, State state) throws Exception {
        if (state == null) {
            state = State.RESTART;
        }
        // Flip state, and then use super to do the actual creation...
        switch (state) {
            case NORMAL:
                state = State.RESET;
                break;
            case RESET:
                state = State.NORMAL;
                break;
            // To test the creation procedure...
            default: {
                if (sequence.bounded()) {
                    throw new IllegalArgumentException("Cannot handle: " + state);
                }
            }
        }
        return super.create(sequence, state);
    }
    
    /**
      * 
      */
    @Override
    protected EvilCommand<E> createDefault(Sequence<E> sequence) throws Exception {
        // To test the creation procedure...
        if (sequence.bounded()) {
            throw new IllegalStateException("Cannot create default command for bounded sequences");
        }
        return new EvilCommand<E>(sequence.unique());
    }
    
}
