/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.NextCommand;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.meta.reflect.TypeLiteral;

/**
  * A <i>reflective command creator</i> creates new
  * {@link Command commands} based on {@link TypeLiteral 
  * type literals} as tokens. <p>
  * 
  * A type literal represents a fully parameterised command type, 
  * for example {@link TypeLiteral}{@code <}{@link NextCommand}{@code <Integer>>}. 
  * The creator can create any command, parameterised or not, declaring 
  * a no-arg constructor (no receiver) and/or constructor with a 
  * single parameter of type {@link Sequence} (a receiver in form
  * of a sequence). <p> 
  * 
  * <b>Implementation notes</b>: <br>
  * The creation is completely type safe with no casting required even though
  * this creator creates commands with type parameters. This is because
  * the {@link TypeLiteral} class ensures that the type parameters are
  * accessible at runtime. This would <b>not</b> have been possible by merely
  * using {@code Class} literals as tokens, as the type parameter information 
  * cannot be expressed using class literals. However, this is only 
  * feasible because we operate on concrete type parameters, e.g. 
  * {@code NextCommand<Integer>}, and not on erased generic types 
  * like {@code NextCommand<E>}. <p>
  * 
  * Furthermore, the factory method in this creator <i>upgrades</i> the type 
  * literal supplied as the token to an {@link InstantiableTypeLiteral 
  * instantiable} type literal, if not already, which may throw a runtime
  * exception in case the generic type represented by the type literal is
  * an interface or abstract class, which naturally cannot be instantiated.
  * This is no different than if a class literal was used, though. As the 
  * bounds on the type literals used by this creator is 
  * {@code <? extends Command<E>>} this will only be the case here 
  * if type literals are created as 
  * {@link TypeLiteral}{@code <}{@link Command}{@code <Integer>>}, i.e. using the 
  * {@link Command} interface instead of a concrete command implementation. 
  * An instantiable type literal can create a generic type based on a given 
  * constructor, here the no-arg or the constructor accepting a single 
  * {@code Sequence} argument as described above. This is utilised in the 
  * factory method in this creator. <p>
  * 
  * Using the {@linkplain dk.rode.thesis.prototype Prototype} pattern as a means
  * to create different types of commands is an, perhaps safer, alternative 
  * to creating commands using reflection. The
  * {@linkplain dk.rode.thesis.abstractfactory Abstract Factory} illustrates the use
  * of prototypes in the creation process.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    TypeLiteral
  * @see    InstantiableTypeLiteral
  * 
  * @param  <E> The type of values delivered by created commands (and thus sequences).
  */
@Participant("ConcreteCreator")
public class ReflectiveCommandCreator<E> 
     extends CommandCreator<E,TypeLiteral<? extends Command<E>>> { // dependency between E and T!

    /**
      * No-arg constructor.
      */
    public ReflectiveCommandCreator() {
        super();
    }

    /**
      * @throws IllegalArgumentException If {@code type} does not represent a concrete,
      *         parameterised {@code Command} type.
      */
    @Override
    protected Command<E> create(Sequence<E> sequence, TypeLiteral<? extends Command<E>> type) throws Exception {
        // May already be instantiable...
        if (type instanceof InstantiableTypeLiteral) {
            // ISSUE: NetBeans 5.5.1 with Java 1.6.0_01 generates an 
            //        unchecked cast warning...
            InstantiableTypeLiteral<? extends Command<E>> itl =
                (InstantiableTypeLiteral<? extends Command<E>>)type;
            // No-arg constructor?
            if (itl.getConstructor().getParameterTypes().length == 0) {
                return itl.newInstance();
            } else {
                return itl.newInstance(sequence);
            }
            
        // If not, we assume the constructor accepting a sequence type must be used...
        } else {
            return InstantiableTypeLiteral.create(type, Sequence.class).newInstance(sequence);
        }
    }
     
}
