/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>binary expression</i> represents any expression
  * that involves (at least) two sub-expressions as operands. <p>
  * 
  * The evaluation of a binary expression is not guaranteed 
  * to be commutative.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The type of value the evaluation of the expression 
  *             operands produces. Normally, but not required to
  *             be, identical with {@code E}.
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.             
  */
public abstract class BinaryExpression<T,E> 
              extends AbstractExpression<E>
           implements NonTerminalExpression<E> {

    /**
      * The first operand. <p>
      * 
      * Never null.
      */
    protected final Expression<? extends T> first;
    /**
      * The second operand. <p>
      * 
      * Never null.
      */
    protected final Expression<? extends T> second;
    
    /**
      * Constructor.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @throws NullPointerException If either expression is null.
      */
    protected BinaryExpression(Expression<? extends T> first, Expression<? extends T> second) {
        if ((this.first = first) == null) {
            throw new NullPointerException("First expression cannot be null");
        } else if ((this.second = second) == null) {
            throw new NullPointerException("Second expression cannot be null");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    protected BinaryExpression(BinaryExpression<? extends T,E> expression) {
        this.first = expression.first.copy();
        this.second = expression.second.copy();
    }

    /**
      *
      */
    public List<Expression<?>> operands() {
        return add(new ArrayList<Expression<?>>(3), this.first, this.second);
    }
    
    /**
      * Returns {@code x separator y}, where {@code x} is the symbolic 
      * representation of the first expression operand and {@code y} 
      * of the the second. 
      * 
      * @param  context The context to use; cannot be null.
      * @param  separator The operand separator; cannot be null.
      * @return A formatted symbol string; never null.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the symbol cannot be generated.
      */
    protected String asSymbol(Context context, String separator) throws ExpressionException {
        return this.asSymbol(context, separator, new StringBuilder()).toString();
    }
    
    /**
      * Returns {@code x separator y}, where {@code x} is the symbolic 
      * representation of the first expression operand and {@code y} 
      * of the the second.
      *   
      * @param  context   The context to use; cannot be null.
      * @param  separator The operand separator; cannot be null.
      * @param  sb        The string builder to use and return; cannot be null.
      * @return {@code sb}; never null.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the symbol cannot be generated.
      */
    protected StringBuilder asSymbol(Context context, String separator, StringBuilder sb) throws ExpressionException {
        // First...
        SymbolIdiom.asSymbol(context, this, this.first, sb);
        // Separator...
        sb.append(' ').append(separator).append(' ');
        // Second...
        return SymbolIdiom.asSymbol(context, this, this.second, sb);
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; first: ").append(this.first)
                 .append("; second: ").append(this.second)
                 .append(']').toString();
    }
    
}
