/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>compare expression</i> can compare two expressions 
  * evaluating to the <b>same</b> {@code Comparable} type 
  * as <i>smaller than</i>, <i>smaller than or equal</i>,
  *  <i>equal</i>, <i>not equal</i>, <i>greater than</i>, or
  *  <i>equal or greater than</i>.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    EqualExpression
  * @see    AndExpression
  * @see    OrExpression
  * @see    NotExpression
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
public class CompareExpression<E extends Comparable<? super E>> // has to be comparable to it self, i.e. same type!
     extends BinaryExpression<E,Boolean>
  implements NonTerminalExpression<Boolean> {

    /**
      * The comparison type to be performed by a 
      * {@link CompareExpression compare} expression. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public static enum Comparison {
        
        /**
          * Compares expression {@code a} and {@code b} to see 
          * if {@code a} is smaller. 
          */
        SMALLER {
            protected boolean matches(int value) {
                return (value < 0);
            }
            protected String asSymbol() {
                return "<";
            }
        },

        /**
          * Compares expression {@code a} and {@code b} to see 
          * if {@code a} is smaller or equal. 
          */
        SMALLER_EQUAL {
            protected boolean matches(int value) {
                return (value <= 0);
            }
            protected String asSymbol() {
                return "<=";
            }
        },
        
        /**
          * Compares expression {@code a} and {@code b} to see 
          * if they are equal. 
          */
        EQUAL {
            protected boolean matches(int value) {
                return (value == 0);
            }
            protected String asSymbol() {
                return "==";
            }
        },

        /**
          * Compares expression {@code a} and {@code b} to see 
          * if they are not equal. 
          */
        NOT_EQUAL {
            protected boolean matches(int value) {
                return (value != 0);
            }
            protected String asSymbol() {
                return "!=";
            }
        },
        
        /**
          * Compares expression {@code a} and {@code b} to see 
          * if {@code a} is greater. 
          */
        GREATER {
            protected boolean matches(int value) {
                return (value > 0);
            }
            protected String asSymbol() {
                return ">";
            }
        },
        
        /**
          * Compares expression {@code a} and {@code b} to see 
          * if {@code a} is equal or greater. 
          */
        GREATER_EQUAL {
            protected boolean matches(int value) {
                return (value >= 0);
            }
            protected String asSymbol() {
                return ">=";
            }
        };
        
        /**
          * Return true if this comparison type matches the
          * value supplied as {@code value}, false if not.
          *   
          * @param  value The value to test.
          * @return True if this comparison type matches {@code value},
          *         false if it does not.
          */
        protected abstract boolean matches(int value);
        
        /**
          * Return the symbol of this comparison type.
          *  
          * @return The symbol; never null;
          */
        protected abstract String asSymbol();
        
    }
    
    /**
      * The type of comparison performed by this expression. <p>
      * 
      * Never null.
      */
    public final Comparison comparison;
    
    /**
      * Constructor. <p>
      * 
      * The second expression used is a {@link ConstantExpression constant}
      * expression created based on {@code type} and {@code value}.
      * 
      * @param  comparison The comparison type; cannot be null.
      * @param  first      The first expression operand; cannot be null.
      * @param  type       The type of the constant expression to use
      *                    as the second expression operand; cannot be null.
      * @param  value      The value of the constant expression to use
      *                    as the second expression operand; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public CompareExpression(Comparison comparison, Expression<? extends E> first, Class<E> type, E value) {
        this(comparison, first, new ConstantExpression<E>(type, value));
    }

    /**
      * Constructor. <p>
      * 
      * The second expression used is a {@link VariableExpression variable}
      * expression created based on {@code type} and {@code name}.
      * 
      * @param  comparison The comparison type; cannot be null.
      * @param  first      The first expression operand; cannot be null.
      * @param  type       The type of the variable expression to use
      *                    as the second expression operand; cannot be null.
      * @param  name       The name of the variable expression to use
      *                    as the second expression operand; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public CompareExpression(Comparison comparison, Expression<? extends E> first, Class<E> type, String name) {
        this(comparison, first, new VariableExpression<E>(type, name));
    }

    /**
      * Constructor.
      * 
      * @param  comparison The comparison type; cannot be null.
      * @param  first      The first expression operand; cannot be null.
      * @param  second     The second expression operand; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public CompareExpression(Comparison comparison, Expression<? extends E> first, Expression<? extends E> second) {
        super(first, second);
        if ((this.comparison = comparison) == null) {
            throw new NullPointerException("Comparison cannot be null");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public CompareExpression(CompareExpression<E> expression) {
        super(expression);
        this.comparison = expression.comparison;
    }

    /**
      * 
      */
    public CompareExpression<E> copy() {
        return new CompareExpression<E>(this);
    }

    /**
      * @return {@code compare}
      */
    public final String name() {
        return "compare";
    }
    
    /**
      *
      */
    public Boolean evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        return this.comparison.matches(this.first.evaluate(context).compareTo(this.second.evaluate(context)));
    }

    /**
      * Returns {@code x operator y}, where {@code x} is the symbolic 
      * representation of the first expression operand, {@code y} 
      * of the the second, and {@code operator} is the operator used
      * for the comparison.
      */
    public String asSymbol(Context context) throws ExpressionException {
        return this.asSymbol(context, this.comparison.asSymbol()); 
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; comparison: ").append(this.comparison)
                 .append("; first: ").append(this.first)
                 .append("; second: ").append(this.second)
                 .append(']').toString();
    }
    
}
