/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>conditional expression</i> represents an <i>if-then-else</i>
  * expression.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("NonTerminalExpression")
public class ConditionalExpression<E> 
     extends BinaryExpression<E,E> 
  implements NonTerminalExpression<E> {

    /**
      * The condition in form of an expression that returns
      * a {@code Boolean} value when evaluated. <p>
      * 
      * Never null.
      */
    private final Expression<Boolean> condition;

    /**
      * Constructor.
      *
      * @param  condition The condition expression; cannot be null.
      * @param  first     The first expression operand; cannot be null.
      * @param  second    The second expression operand; cannot be null.
      * @throws NullPointerException If either expression is null.
      */
    public ConditionalExpression(Expression<Boolean> condition, Expression<? extends E> first, Expression<? extends E> second) {
        super(first, second);
        if ((this.condition = condition) == null) {
            throw new NullPointerException("Condition cannot be null");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public ConditionalExpression(ConditionalExpression<E> expression) {
        super(expression);
        this.condition = expression.condition.copy();
    }
    
    /**
      *
      */
    public ConditionalExpression<E> copy() {
        return new ConditionalExpression<E>(this);
    }

    /**
      * @return {@code conditional}
      */
    public final String name() {
        return "conditional";
    }
    
    /**
      *
      */
    @Override
    public List<Expression<?>> operands() {
        return add(new ArrayList<Expression<?>>(), this.condition, this.first, this.second);
    }
 
    /**
      *
      */
    public E evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        return this.condition.evaluate(context) 
             ? (E)this.first.evaluate(context) // NetBeans complains otherwise 
             : this.second.evaluate(context);
    }

    /**
      * 
      */
    @Override
    public boolean contains(Expression<?> expression) {
        return super.contains(expression) || this.condition.contains(expression);
    }
    
    /**
      * Returns {@code (c ? x : y)}, where {@code c} is the symbolic 
      * representation of the condition, {@code x} of the first 
      * expression operand, and {@code y} of the the second.  
      */
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        sb.append('(');
        // Condition...
        SymbolIdiom.asSymbol(context, this, this.condition, sb);
        sb.append(" ? ");
        // Operands...
        this.asSymbol(context, ":", sb);
        return sb.append(')').toString(); 
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; condition: ").append(this.condition)
                 .append("; first: ").append(this.first)
                 .append("; second: ").append(this.second)
                 .append(']').toString();
    }

}
