/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.interpreter;

import java.util.LinkedHashMap;
import java.util.Map;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.interpreter.CompareExpression.Comparison;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.singleton.DanishAlphabetSequence;

/**
  * <b>Interpreter</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Interpreter</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * Accepted argument: {@code -value=n}, where {@code n} is an integer value. 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Interpreter tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        boolean ok = true;
        
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        this.log(log);
        
        //
        // Create context, interpreter, and sequences
        //
        
        // Context
        Context context = new Context();
        // Number interpreter..
        Interpreter<Number> interpreter = new Interpreter<Number>(Number.class);
        // Comparable interpreter (can also handle any Number)...
        TypeLiteral<Comparable<?>> type = new TypeLiteral<Comparable<?>>(){/*magic*/};
        Interpreter<Comparable<?>> comparableInterpreter = new Interpreter<Comparable<?>>(type);
        
        // Sequence...
        int range = MathFacade.getInstance().getRandomNumber(111); 
        int size  = (int)(range * 2.3);
        RangeSequence sequence = new RangeSequence(0, size);
        // Test values. Cannot be larger than the sequence size, 
        // because we use the "set" expression!...
        Integer value1 = MathFacade.getInstance().getRandomNumber(range);
        Integer value2 = Math.min(31, arguments.getArgumentValue("value", value1 / 3, Integer.class));

        out.heading("Created objects");
        out.println("Context:        ", context);
        out.println("Interpreter:    ", interpreter);
        out.println("Interpreter:    ", comparableInterpreter);
        out.println("Sequence:       ", sequence).println();
        out.println("Sequence range: 0 - ", size).println();
        
        out.println("Value 1:        ", value1);
        out.println("Value 2:        ", value2).println();

        Expression<Integer> expression = null;
        Integer valueAsInteger = null;
        Number valueAsNumber = null;
        
        //
        // Expression 1
        //
        
        try {
            out.heading("Expression 1");

            sequence.reset();
            expression = this.create1(sequence, Integer.class, value1, context);
            
            out.println("Expression (value: ", value1, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsNumber = interpreter.interpret(context, expression);
            out.println().println("Result of evaluation: ", valueAsNumber).println();

            sequence.reset();
            expression = this.create1(sequence, Integer.class, value2, context);
            
            out.println("Expression (value: ", value2, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsInteger = interpreter.interpret(context, expression, Integer.class);
            out.println().println("Result of evaluation: ", valueAsInteger);
            
        } catch (Exception e) {
            out.error("Expression 1 error").error(e);
            ok = false;
        }
            
        //
        // Expression 2
        //

        try {
            out.heading("Expression 2");
            
            sequence.reset();
            expression = this.create2(sequence, Integer.class, value1, context);
            
            out.println("Expression (value: ", value1, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsNumber = interpreter.interpret(context, expression);
            out.println().println("Result of evaluation: ", valueAsNumber).println();

            sequence.reset();
            expression = this.create2(sequence, Integer.class, value2, context);
            
            out.println("Expression (value: ", value2, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsInteger = interpreter.interpret(context, expression, Integer.class);
            out.println().println("Result of evaluation: ", valueAsInteger);
        
        } catch (Exception e) {
            out.error("Expression 2 error").error(e);
            ok = false;
        }

        //
        // Expression 3
        //

        try {
            out.heading("Expression 3");

            sequence.reset();
            expression = this.create3(sequence, Integer.class, value1, context);
            
            out.println("Expression (value: ", value1, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsNumber = interpreter.interpret(context, expression);
            out.println().println("Result of evaluation: ", valueAsNumber).println();

            sequence.reset();
            expression = this.create3(sequence, Integer.class, value2, context);
            
            out.println("Expression (value: ", value2, "): ", expression.asSymbol(context.reset())).println(); 
            valueAsInteger = interpreter.interpret(context, expression, Integer.class);
            out.println().println("Result of evaluation: ", valueAsInteger);
        
        } catch (Exception e) {
            out.error("Expression 3 error").error(e);
            ok = false;
        }
        
        //
        // Expression 4
        //

        try {
            out.heading("Expression 4");

            sequence.reset();
            expression = this.create4(sequence, Integer.class, value1, value2, context);
            
            out.println("Expression (values: [", value1, ",", value2, "]): ", expression.asSymbol(context.reset())).println(); 
            valueAsNumber = interpreter.interpret(context, expression);
            out.println().println("Result of evaluation: ", valueAsNumber).println();

            sequence.reset();
            expression = this.create4(sequence, Integer.class, value2, value1, context);
            
            out.println("Expression (values: [", value2, ",", value1, "]: ", expression.asSymbol(context.reset())).println(); 
            valueAsInteger = interpreter.interpret(context, expression, Integer.class);
            out.println().println("Result of evaluation: ", valueAsInteger);
                    
        } catch (Exception e) {
            out.error("Expression 4 error").error(e);
            ok = false;
        }
        
        //
        // Expression 5
        //

        // Skip verbose logging...
        this.log(false);
        
        out.heading("Expression 5 (15 expression loops with mixed types and sequences)");

        String[] strings = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l"};
       
        // Sequences...
        
        // String...
        Sequence<String> stringSequence = DanishAlphabetSequence.Instance;
        // Integer... ++ to ensure never zero!
        Sequence<Integer> integerSequence = new RangeSequence(1, ++value2);
        
        out.println("Using range sequence [1-", value2, "] and Danish alphabet sequence:");
        out.println("- Range sequence:           ", integerSequence);
        out.println("- Danish alphabet sequence: ", stringSequence);
        
        out.println("String test values:  ", strings);
        out.println("Integer test values: 0-19");

        Map<Comparable<?>,Integer> results = new LinkedHashMap<Comparable<?>,Integer>();

        boolean first = true;
        
        // Integer value...
        for (int i = 0; i < 20; i++) {
            // String value...
            for (String s : strings) {
                Expression<Comparable<?>> exp = null;
                String string = null; 
                try {
                    Comparable<?> c = null;
                    exp = this.create5(context, stringSequence, integerSequence, s, i);
                    string = exp.asSymbol(context.reset());
                    if (first) {
                        first = false;
                        out.println().println("Sample expression [i, s] = [", i, ", ", s ,"]: ", string).println();
                    }
                    c = comparableInterpreter.interpret(context, exp);
                    
                    Integer count = results.get(c);
                    if (count == null) {
                        count = 1;
                    }
                    results.put(c, count + 1);
                        
                } catch (Exception e) {
                    out.error("Expression 5 error: [i, s] = [", i, ", ", s, "]: ", string).error(e);
                    ok = false;
                }
            }
        }
        
        out.println("Results:");
        for (Comparable<?> c : results.keySet()) {
            if (c == null) {
                out.println("Interpreted exited with no value -> ", results.get(c), " time(s)");
            } else {
                out.println(c.getClass().getName(), " = ", c, " -> ", results.get(c), " time(s)");
            }
        }
        
        out.heading("Interpreter tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * True to enable logging per expression type.
      * 
      * @param log True to enable verbose logging per expression type,
      *            false to disable.
      */
    private void log(boolean log) {
        LogFactory.getLog(AndExpression.class).setActive(log);
        LogFactory.getLog(AssignmentExpression.class).setActive(log);
        LogFactory.getLog(CompareExpression.class).setActive(log);
        LogFactory.getLog(ConditionalExpression.class).setActive(log);
        LogFactory.getLog(ConstantExpression.class).setActive(log);
        LogFactory.getLog(CurrentExpression.class).setActive(log);
        LogFactory.getLog(EqualExpression.class).setActive(log);
        LogFactory.getLog(BreakExpression.class).setActive(log);
        LogFactory.getLog(FlowExpression.class).setActive(log);
        LogFactory.getLog(NextExpression.class).setActive(log);
        LogFactory.getLog(NotExpression.class).setActive(log);
        LogFactory.getLog(OrExpression.class).setActive(log);
        LogFactory.getLog(ResetExpression.class).setActive(log);
        LogFactory.getLog(ReverseExpression.class).setActive(log);
        LogFactory.getLog(SetExpression.class).setActive(log);
        LogFactory.getLog(VariableExpression.class).setActive(log);
        
        LogFactory.getLog(Context.class).setActive(log);
        LogFactory.getLog(Interpreter.class).setActive(log);
    }
    
    /**
      * Creates: 
      * <pre>
      *   next[20], y = current, (reverse[true] ? next[10], x = current, (x < value ? x : break[y]) : current)
      * </pre>
      * Where {@code value} is the value supplied as the {@code value} argument.
      *
      * @param  <E>      The type of comparable value the evaluation of
      *                  the returned expression produces.
      * @param  sequence The sequence to manipulate; never null.
      * @param  type     The expression and value type; never null.
      * @param  value    The value used in the expression; never null.
      * @param  context  The context to use; never null.
      * @return The created expression; never null.
      * @throws ExpressionException If the creation fails.
      */
    private <E extends Comparable<? super E>> Expression<E> create1(Sequence<E> sequence, Class<E> type, E value, Context context) throws ExpressionException {
        VariableExpression<E> x = new VariableExpression<E>(type, "x");
        VariableExpression<E> y = new VariableExpression<E>(type, "y");
        ConstantExpression<E> z = new ConstantExpression<E>(type, value);
        context.assign(z, value);
        
        return new FlowExpression<E>()
        .add(new NextExpression<E>(
                 sequence, 
                 20
             ))
        .add(new AssignmentExpression<E>(
                 y,
                 new CurrentExpression<E>(sequence)
             ))
        .add(new ConditionalExpression<E>(
                 new ReverseExpression<E>(sequence), 
                 new FlowExpression<E>()
                     .add(new NextExpression<E>(
                              sequence, 
                              10
                          ))
                     .add(new AssignmentExpression<E>(
                              x,
                              new CurrentExpression<E>(sequence)
                          ))
                     .add(new ConditionalExpression<E>(
                              new CompareExpression<E>(
                                  Comparison.SMALLER, 
                                  x, 
                                  context.getVariable(type, z.name) // or just z :)
                              ),
                              x,
                              new BreakExpression<E>(y)
                          )).initialise(),
                 new CurrentExpression<E>(sequence)
             )).initialise();
    }

    /**
      * Creates: 
      * <pre>
      *   next[value], foo = Integer{next[(boo ? current : value)]}, (reverse[foo < value] ? next[set[value], next[bar = foo], (foo > value ? bar : break[foo])] : foo)      
      * </pre>
      * Where {@code value} is the value supplied as the {@code value} argument.
      * 
      * @param  <E>      The type of comparable numeric value the evaluation of
      *                  the returned expression produces.
      * @param  sequence The sequence to manipulate; never null.
      * @param  type     The expression and value type; never null.
      * @param  value    The value used in the expression; never null.
      * @param  context  The context to use; never null.
      * @return The created expression; never null.
      * @throws ExpressionException If the creation fails.
      */
    private <E extends Number & Comparable<? super E>> Expression<E> create2(Sequence<E> sequence, Class<E> type, E value, Context context) throws ExpressionException {
        VariableExpression<E> foo = new VariableExpression<E>(type, "foo");
        VariableExpression<E> bar = new VariableExpression<E>(type, "bar");
        ConstantExpression<E> goo = new ConstantExpression<E>(type, value);
        VariableExpression<Boolean> boo = new VariableExpression<Boolean>(Boolean.class, "boo");
        
        context.assign(boo, value.intValue() % 2 == 0);
        context.assign(goo, value);
        
        return new FlowExpression<E>()
        .add(new NextExpression<E>(
                 sequence, 
                 goo // goo must evaluate to a Number
             )) 
        .add(new AssignmentExpression<E>(
                 foo,
                 new TypedExpressionDecorator<E>(
                     new NextExpression<E>(
                         sequence, 
                         new ConditionalExpression<E>(
                             boo,
                             new CurrentExpression<E>(sequence), // Number
                             new ConstantExpression<E>(type, value) // Number
                         )
                     ),
                     type
                 )    
             ))
        .add(new ConditionalExpression<E>(
                 new ReverseExpression<E>(
                     sequence,
                     new CompareExpression<E>(
                         Comparison.SMALLER, 
                         foo, 
                         goo 
                     )
                 ), 
                 new NextExpression<E>(
                     sequence,
                     new FlowExpression<E>() // Number
                         .add(new SetExpression<E>(
                                  sequence, 
                                  goo
                              ))
                         .add(new NextExpression<E>(
                                  sequence, 
                                  new AssignmentExpression<E>( // Number
                                      bar,
                                      foo
                                  )
                              ))
                         .add(new ConditionalExpression<E>(
                                  new CompareExpression<E>(
                                      Comparison.GREATER, 
                                      foo, 
                                      goo 
                                  ),
                                  bar,
                                  new BreakExpression<E>(foo)
                              )).initialise()
                      ),
                 foo
             )).initialise();
    }

    /**
      * Creates a <i>loop-structure</i> by using cyclic expression
      * references without intermediate variables: 
      * <pre>
      *    (current < value ? next[1], {{flow -> conditional}} : current)
      * </pre>
      * Where {@code value} is supplied as the argument {@code value}.
      * 
      * @param  <E>      The type of comparable value the evaluation of
      *                  the returned expression produces.
      * @param  sequence The sequence to manipulate; never null.
      * @param  type     The expression and value type; never null.
      * @param  value    The value used in the expression; never null.
      * @param  context  The context to use; never null.
      * @return The created expression; never null.
      * @throws ExpressionException If the creation fails.
      * 
      */
    private <E extends Comparable<? super E>> Expression<E> create3(Sequence<E> sequence, Class<E> type, E value, Context context) throws ExpressionException {
        FlowExpression<E> loop = new FlowExpression<E>();

        loop.add(new ConditionalExpression<E>(
                new CompareExpression<E>(
                    Comparison.SMALLER, 
                    new CurrentExpression<E>(sequence), type, value
                ),
                new FlowExpression<E>()
                    .add(new NextExpression<E>(sequence))
                    .add(loop)   // cyclic reference...
                    .initialise(), 
                new CurrentExpression<E>(sequence)
            ));
        return loop.initialise();
    }
    
    /**
      * Generates the mother of all sequence expressions: 
      * <pre>
      *    (current < value1 ? next[20], y = current, (reverse[true] ? next[10], x = current, (x < value2 ? x : break[y]) : current), x = current, (current < value1 ? next[1], {{flow -> flow}} : current), y = current, (reverse[(next[y] > x | (y == x ? break[boo] : false))] ? break[x] : (((!current == next[value1], foo = Integer{next[(boo ? current : value1)]}, (reverse[foo < value1] ? next[set[value1], next[bar = foo], (foo > value1 ? bar : break[foo])] : foo)) && x = bar == (current < value2 ? next[1], {{flow -> flow}} : current)) ? x : foo)) : exit)
      * </pre>
      * Where {@code value1} and {@code value2} are the values supplied 
      * as arguments to this method. <p>
      * 
      * @param  <E>      The type of comparable value the evaluation of
      *                  the returned expression produces.
      * @param  sequence The sequence to manipulate; never null.
      * @param  type     The expression and value type; never null.
      * @param  value1   The first value used in the expression; never null.
      * @param  value2   The second value used in the expression; never null.
      * @param  context  The context to use; never null.
      * @return The created expression; never null.
      * @throws ExpressionException If the creation fails.
      */
    private <E extends Number & Comparable<? super E>> Expression<E> create4(Sequence<E> sequence, Class<E> type, E value1, E value2, Context context) throws ExpressionException {
        VariableExpression<E> foo = new VariableExpression<E>(type, "foo");
        VariableExpression<E> bar = new VariableExpression<E>(type, "bar");
        VariableExpression<E> x = new VariableExpression<E>(type, "x");
        VariableExpression<E> y = new VariableExpression<E>(type, "y");
        VariableExpression<Boolean> boo = new VariableExpression<Boolean>(Boolean.class, "boo");
        
        // Make sure the context is initialised with proper variables
        // before use, since we include expressions from the other
        // test methods!!!
        
        if (!context.hasValue(boo)) {
            context.assign(boo, value1.intValue() % 2 != 0);
        }
        if (!context.hasValue(foo)) {
            context.assign(foo, value1);
        }
        if (!context.hasValue(bar)) {
            context.assign(bar, value2);
        }
        
        CurrentExpression<E> current = new CurrentExpression<E>(sequence);
        
        ConditionalExpression<E> condition = 
            new ConditionalExpression<E>(
                 new CompareExpression<E>(
                     Comparison.SMALLER, 
                     current, type, value1
                 ),
                 new FlowExpression<E>()
                     .add(create1(sequence, type, value2, context))
                     .add(new AssignmentExpression<E>(x, current))
                     .add(create3(sequence, type, value1, context))
                     .add(new AssignmentExpression<E>(y, current))
                     .add(new ConditionalExpression<E>(
                              new ReverseExpression<E>(
                                  sequence,
                                  new OrExpression( 
                                      false, // no short-circuit
                                      new CompareExpression<E>(
                                          Comparison.GREATER, 
                                          new NextExpression<E>( // side-effect
                                              sequence, 
                                              y),  
                                          x  
                                      ),
                                      new ConditionalExpression<Boolean>(
                                          new EqualExpression( // any type works
                                              y, x
                                          ),
                                          // This cannot be handled by any interpreter
                                          // interpreting Number values. It will cause
                                          // the interpreter to exit with a null value...
                                          new BreakExpression<Boolean>(boo), 
                                          new ConstantExpression<Boolean>(Boolean.class, false)
                                      )
                                  )
                              ),
                              new BreakExpression<E>(x),
                              new ConditionalExpression<E>(
                                  new AndExpression( // short-circuit
                                      new NotExpression(
                                          new EqualExpression(
                                              current, create2(sequence, type, value1, context)        
                                          )
                                      ),
                                      new EqualExpression(
                                          new AssignmentExpression<E>(x, bar), 
                                          create3(sequence, type, value2, context)
                                      )
                                  ),
                                  x,
                                  foo) // set by create2
                          )).initialise(),
                 new BreakExpression<E>()
             );
        return condition;
    }
    
    /**
      * Creates an expression with a loop having 15 iterations as well as 
      * different sequences and types: 
      * <pre>
      *    ((!next[1] == 15) ? Comparable{c} = (Boolean{boo} = (Integer{i} > Integer{current} || (String{s} == String{next[Integer{next[1]}]} ? break[Boolean{boo}] : booboo = (!booboo))) ? Integer{next[1]} : String{current}), {{flow -> conditional}} : Comparable{c})
      * </pre>
      * Where {@code stringValue} is the initial value of {@code s}, and 
      * {@code intergerValue} of {@code i}. <p>
      * 
      * The identity hash codes of the created sequences are listed as well
      * for each terminal expression. 
      * 
      * @param  context      The context to use; never null.
      * @param  string       A sequence delivering strings; never null.
      * @param  integer      A sequence delivering integers; never null.
      * @param  stringValue  A string value used in the expression; never null.
      * @param  integerValue An integer value used in the expression; never null.
      * @return The created expression; never null.
      * @throws ExpressionException If the creation fails.
      */
    @SuppressWarnings("unchecked") // Suppress warnings for use of raw Comparable type...
    private Expression<Comparable<?>> create5(Context context, Sequence<String> string, Sequence<Integer> integer, String stringValue, Integer integerValue) throws ExpressionException {
        
        // true argument: include type information in symbolic representation
        VariableExpression<Comparable> c = new VariableExpression<Comparable>(Comparable.class, "c", true);
        VariableExpression<String> s = new VariableExpression<String>(String.class, "s", true);
        VariableExpression<Integer> i = new VariableExpression<Integer>(Integer.class, "i", true);
        VariableExpression<Boolean> boo = new VariableExpression<Boolean>(Boolean.class, "boo", true);
        VariableExpression<Boolean> booboo = new VariableExpression<Boolean>(Boolean.class, "booboo", false);
        
        context.assign(c, "default-value");
        context.assign(s, stringValue);
        context.assign(i, integerValue);

        context.assign(boo, integerValue % 2 == 0);
        context.assign(booboo, integerValue % 3 == 1);
        
        // Loop count sequence...
        Sequence<Long> loopCount = new LongSequence(0);
        
        // Short cuts to expressions...
        Expression<String> currentString = 
            new TypedExpressionDecorator<String>(
                new CurrentExpression<String>(string, true), // true = include hash code in symbolic representation
                String.class
            );
        Expression<Integer> currentInteger = 
            new TypedExpressionDecorator<Integer>(
                new CurrentExpression<Integer>(integer, true),
                Integer.class
            );
        
        Expression<Integer> nextInteger = 
            new TypedExpressionDecorator<Integer>(
                new NextExpression<Integer>(integer, true),
                Integer.class
            );
        Expression<String> nextString  = 
            new TypedExpressionDecorator<String>(
                new NextExpression<String>(string, nextInteger, true), 
                String.class
            );

        // Expression to evaluate inside a loop...
        ConditionalExpression<Comparable<?>> expression = 
            new ConditionalExpression<Comparable<?>>(
                new AssignmentExpression<Boolean>(
                    boo,
                    new OrExpression( 
                        true, // short-circuit
                        new CompareExpression<Integer>( // can only compare same type with it self
                            Comparison.GREATER, 
                            i,
                            currentInteger 
                        ),
                        new ConditionalExpression<Boolean>(
                            new EqualExpression( // equal works on any type
                                s, nextString
                            ),
                            new BreakExpression<Boolean>(boo), // also comparable!  
                            new AssignmentExpression<Boolean>(
                                booboo,
                                new NotExpression(booboo)
                            )
                        )
                    )
                ),
                nextInteger,
                currentString
            );

        // Loop...
        FlowExpression<Comparable> loop = new FlowExpression<Comparable>();
        loop.add(new ConditionalExpression<Comparable>(
                     new NotExpression(
                         new EqualExpression (
                             new NextExpression<Long>(loopCount, true), 
                             new ConstantExpression<Long>(Long.class, 15L) // number of iterations
                         )
                     ),
                     new FlowExpression<Comparable>()
                         .add(new AssignmentExpression<Comparable>(c, expression))
                         .add(loop)   // cyclic reference...
                         .initialise(), 
                     c
                 ));
        Expression<?> e = loop.initialise();
        return (Expression<Comparable<?>>)e;
    }
    
}
