/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.iterator;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.adapter.AdapterStrategy;
import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.DigitSequence;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.prototype.SymbolSequence;
import dk.rode.thesis.proxy.NonResettableSequence;
import dk.rode.thesis.singleton.DanishAlphabetSequence;
import dk.rode.thesis.singleton.NorwegianAlphabetSequence;
import dk.rode.thesis.singleton.SmileySequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Iterator</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Iterator</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Iterator tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(LoggingValueProcessor.class).setActive(log);
        
        final int valuesPerLine = arguments.getArgumentValue("line", 10, Integer.class);
        
        //
        // Create sequences
        //
        
        List<Sequence<String>> sequences = new ArrayList<Sequence<String>>();

        sequences.add(new IterableSequence<String>(new NonResettableSequence<String>(new EnglishAlphabetSequence())));
        sequences.add(new IterableSequence<String>(DanishAlphabetSequence.Instance));
        sequences.add(new SequenceAdapter<CharSequence,String>(new DigitSequence(), AdapterStrategy.CHAR_TO_STRING));
        sequences.add(new AppenderDecorator(new EnglishAlphabetSequence(), "<<", ">>"));
        sequences.add(new SequenceAdapter<Integer,String>(new PrimeSequence(97), AdapterStrategy.NUMBER_TO_STRING));
        sequences.add(NorwegianAlphabetSequence.getInstance()); 
        sequences.add(new SequenceAdapter<CharSequence,String>(new SymbolSequence('$', '%', '!', '?'), AdapterStrategy.CHAR_TO_STRING)); 
        sequences.add(SmileySequence.getSequence()); 
        
        // Print instances...
        out.heading("Sequence instances");
        for (int i = 0; i < sequences.size(); i++) {
            Sequence<String> sequence = sequences.get(i);
            out.println("Sequence ", i, ": ", sequence);
        }
        out.println()
           .println("NOTE: sequence 0 is a non-resettable sequence and")
           .println("      sequence 5 is an anonymous singleton, both of")
           .println("      which will influence their (re-)use!")
           .println();
        
        //
        // Iterator usage
        //
        
        out.heading("Sequence iterator usage (external iterators)");
        for (int i = 0; i < sequences.size(); i++) {
            iterate(out, i, sequences.get(i), valuesPerLine, true);
        }

        out.heading("Sequence processing (internal iterator)");
        
        // Only print the first 10 values per sequence...
        int count = 10;
        LoggingValueProcessor lvp = new LoggingValueProcessor(count);
        out.println("Using processor: ", lvp);
        
        for (int i = 0; i < sequences.size(); i++) {
            Sequence<String> sequence = sequences.get(i);
            ProcessableSequence<String> ps = new ProcessableSequence<String>(sequence);
            out.println();
            out.println("Sequence ", i, ": ", sequence, " -> ", ps);
            if (ps.process(lvp)) {
                out.println("All values processed");
            } else {
                out.println("Processor existed after ", count, " values");
            }
        }
        
        out.heading("Iterator tests succeeded!");
        
        return true;
    }

    /**
      * Iterates over the sequence values in {@code sequence}.
      *  
      * @param <V>           The type of values delivered by {@code sequence}. 
      * @param out           The log to use; never null.
      * @param index         The sequence index.
      * @param sequence      The sequence; never null.
      * @param valuesPerLine Number of sequence values to log per line.
      * @param recurse       True to recurse.
      */
    private final static <V> void iterate(Log out, int index, Sequence<V> sequence, int valuesPerLine, boolean recurse) {
        out.println().println("Sequence ", index, ": ", sequence);
        // Use built in for-each loop...
        if (sequence instanceof IterableSequence) {
            iterate(out, (IterableSequence<V>)sequence, valuesPerLine, true);
        } else {
            iterate(out, new SequenceIterator<V>(sequence), valuesPerLine, true);
        }
        out.println().println("Done...");
    }

    /**
      * Iterates over the sequence values delivered indirectly by {@code iterator}.
      *  
      * @param <V>           The type of values delivered by {@code sequence}. 
      * @param out           The log to use; never null.
      * @param iterator      The iterator; never null.
      * @param valuesPerLine Number of sequence values to log per line.
      * @param recurse       True to recurse.
      */
    private final static <V> void iterate(Log out, SequenceIterator<V> iterator, int valuesPerLine, boolean recurse) {
        out.println("Iterator: ", iterator.toString(SequencePolicy.VERBOSE));
        int i = 0;
        boolean first = true;
        while (iterator.hasNext()) {
            if (!first) {
                out.print(", ");
            } else {
                first = false;
            }
            out.print(iterator.next());
            if (++i % valuesPerLine == 0) {
                out.println();
                first = true;
                if ((recurse) && (i == valuesPerLine)) {
                    out.println()
                       .println("Spawning...")
                       .println("Iterable sequence based on: ", iterator.sequence);
                    iterate(out, new IterableSequence<V>(iterator.sequence), valuesPerLine, false);
                    out.println()
                       .println("Done...")
                       .println()
                       .println("Continuing: ", iterator, "...");
                }
            }
        }
    }
    
    /**
      * Iterates over the sequence values delivered by {@code sequence}.
      *  
      * @param <V>           The type of values delivered by {@code sequence}. 
      * @param out           The log to use; never null.
      * @param sequence      The sequence; never null.
      * @param valuesPerLine Number of sequence values to log per line.
      * @param recurse       True to recurse.
      */
    private final static <V> void iterate(Log out, IterableSequence<V> sequence, int valuesPerLine, boolean recurse) {
        out.println("Iterable: ", sequence);
        // The returned iterator *does* advance "sequence"...
        int i = 0;
        boolean first = true;
        for (V value : sequence) {
            if (!first) {
                out.print(", ");
            } else {
                first = false;
            }
            out.print(value);
            if (++i % valuesPerLine == 0) {
                out.println();
                first = true;
            }
            if ((recurse) && (i == valuesPerLine)) {
                out.println()
                   .println("Spawning...")
                   .println("Iterator based on copy of: ", sequence);
                iterate(out, new SequenceIterator<V>(sequence.copy()), valuesPerLine, false);
                out.println()
                   .println("Done...")
                   .println()
                   .println("Continuing: ", sequence, "...");
            }
        }
    }
    
}
