/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Iterator</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.257]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Provide a way to access the elements of an aggregate object 
  *    sequentially without exposing its underlying representation  
  *    
  * </i></ul>
  * 
  * This package provides both <b>external</b> and <b>internal iterators</b>
  * that can be used to iterate over 
  * {@link dk.rode.thesis.meta.model.Sequence#bounded() bounded}
  * {@link dk.rode.thesis.meta.model.Sequence} values instead 
  * of using {@link dk.rode.thesis.meta.model.Sequence#next()}. <p>
  * 
  * For <b>external</b> iterators, the <b>Iterator</b> participant is represented 
  * by Java's built-in {@link java.util.Iterator java.util.Iterator} interface. A single class 
  * representing the <b>ConcreteIterator</b> participant is supplied here, 
  * namely {@link dk.rode.thesis.iterator.SequenceIterator}. <p>
  * 
  * The <b>Aggregate</b> participant is represented by Java's built in
  * {@link java.lang.Iterable java.lang.Iterable} interface and the <b>ConcreteAggregate</b>
  * participant is represented by the 
  * {@link dk.rode.thesis.iterator.IterableSequence} class. <p>
  * 
  * Regarding <b>internal</b> iterators, the <b>Iterator</b> participant
  * is represented by the {@link dk.rode.thesis.iterator.ProcessableSequence}
  * class. A processable sequence decorates any bounded sequence to allow
  * processing of sequence values using a 
  * {@link dk.rode.thesis.iterator.ValueProcessor} instance supplied
  * at iteration time. Hence, <i>decoration</i> and <i>composition</i> is used 
  * in favour of inheritance as used by Gamma et al. for their internal
  * iterators [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.267-269]. Any {@code ValueProcessor} implementation 
  * is considered a representation of the <b>ConcreteIterator</b> participant
  * for internal iterators, for example the 
  * {@link dk.rode.thesis.iterator.LoggingValueProcessor} class. <p>
  * 
  * No classes have been defined to represent the <b>Aggregate</b> and
  * <b>ConcreteAggregate</b> participants for internal iterators. Java defines
  * no standard interface for internal iterators. Here, the construction of 
  * {@code ProcessableSequence} objects is left to the discretion of the
  * context using them. <p> 
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<br>
  * <img src="doc-files/Iterator.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br/>
  * The {@code IterableSequence} and {@code ProcessableSequence} classes 
  * utilises the {@linkplain dk.rode.thesis.decorator Decorator} pattern 
  * to allow any bounded {@code Sequence} to become iterable. The delegation
  * of work to a {@code ValueProcessor} from a {@code ProcessableSequence} 
  * instance is an application of the {@linkplain dk.rode.thesis.strategy 
  * Strategy} pattern. <p>  
  * 
  * Unless care is taken to copy a sequence before it is used by the 
  * iterators defined here, the iterators cannot be considered robust. 
  * They are not fail-fast as Java's iterators normally are. They cannot 
  * mimic a <i>null iterator</i> [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.262-263] because of the 
  * {@code Sequence} requirement to always deliver a value, i.e. 
  * sequences are never empty. As the sequences to iterate over are decorated
  * by the iterators, the iterators only have access to public {@code Sequence} 
  * behaviour.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Iterator", 
         scope        = Scope.Object,
         purpose      = Purpose.Behavioural,
         participants = {"Iterator", "ConcreteIterator", "Aggregate", "ConcreteAggregate"})
package dk.rode.thesis.iterator;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;

