/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import dk.rode.thesis.meta.reflect.Reflection;

/**
  * A <i>simple log manager</i> to create and deliver logs to use, 
  *	either globally or per unique class. <p>
  *
  * Per default, the global default log is {@link Log#setActive(boolean) enabled}, 
  * while class specific are disabled. <p>
  *  
  * Its usage and concrete sub-class is hidden as it is accessed
  * through the {@link LogFactory}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This manager utilises the {@linkplain dk.rode.thesis.templatemethod Template Method} and 
  * {@linkplain dk.rode.thesis.factorymethod Factory Method}
  * patterns: the {@link #create(Class, Class, AtomicInteger)} method is the 
  * factory method, which is utilised by the {@link #getLog(Class)} template method. <p>
  * 
  * In a real-life scenario, there should be such a thing as log-timeout
  * to ensure the file logs are closed properly.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class LogManager {

    /**
      * Created logs per class. <p>
      *
      * Null key represents the global log. <p>
      *
      * Never null.
      */
    private final Map<Class<?>,Log> logs;
    /**
      * Class of log to use. <p>
      */
    private final Class<?> type;
    /**
      * Line number values, if not null. 
      */
    private final AtomicInteger lineNumber;
    
    /**
      * No-arg constructor.
      */
    protected LogManager() {
        this(null, true);
    }

    /**
      * Constructor.
      * 
      * @param type        The type of log to use; can be null, in which
      *                    case the default type of log is used.
      * @param lineNumbers True to log line numbers.
      * 
      * @param <T> The type of log created by this manager.
      */
    protected <T extends Log> LogManager(Class<T> type, boolean lineNumbers) {
        this.type = type == null ? AppendableLog.class : type;
        this.logs = new HashMap<Class<?>,Log>();
        this.lineNumber = lineNumbers ? new AtomicInteger(1) : null;
    }
    
    /**
      * Return the default log.
      *
      * @return The default log; never null.
      */
    public final Log getLog() {
        return this.getLog(null); // default
    }

    /**
      * Returns, and creates if necessary, the log associated
      * with the class supplied as {@code class}, if any. <p>
      * 
      * If {@code clazz} is null, the default log is used. If 
      * {@code clazz} is an anonymous class, the first non-anonymous
      * super class is used <b>or</b> the first and only 
      * interface implemented (recursively as well). 
      *
      * @param  clazz The class; can be null, i.e. default log.
      * @return The log to use; never null.
      */
    public final synchronized Log getLog(Class<?> clazz) {
        clazz = Reflection.getClass(clazz);
        Log log = this.logs.get(clazz);
        if (log == null) {
            // Create...
            try {
                if ((log = this.create(clazz, this.type, this.lineNumber)) == null) {
                    throw new NullPointerException("No log created");
                }
                
            // If we could not create a log, we create a default one in 
            // order to make the application will not fail because of logger
            // failure...
            } catch (Throwable t) {
                log = new AppendableLog(System.out, clazz);
                log.error("Could not create log for ", clazz == null ? "main logger" : clazz.getName(), " - using: ", log);
                log.error("Cause: ", t);
            }
            this.logs.put(clazz, log);
        }
        return log;
    }

    /**
      * Factory method creating the types of log to use. <p>
      *
      * Override this for different behaviour.
      *
      * @param  clazz      The class to associate the created log with, if any;
      *                    can be null, i.e. default log.
      * @param  type       The type of log to use; never null.                     
      * @param  lineNumber Line number values; can be null.                
      * @return A new {@code Log} instance; never null.
      * @throws Exception In case the creation fails (a default log 
      *         will be created in its place). 
      */
    protected Log create(Class<?> clazz, Class<?> type, AtomicInteger lineNumber) throws Exception {
        // Same number values for the global log, unique for specific logs...
        if (FileLog.class.isAssignableFrom(type)) {
            return new FileLog(true, clazz, this.lineNumber); 
            
        // Same number values since we write to System.out, and per default
        // disable non-global logs...
        } else if (AppendableLog.class.isAssignableFrom(type)) {
            return new AppendableLog(System.out, false, clazz, clazz == null, lineNumber);
            
        } else {
            return new NullLog(clazz);
        }
    }
    
}
