/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.model;

import dk.rode.thesis.state.ReversiblePrimeSequence;

/**
  * A <i>prime sequence</i> returns with each call to {@link #next()}
  * the next prime number in the sequence of prime numbers smaller
  * than {@code maximum}, where {@code maximum} is supplied at
  * construction time. Hence, returned values are in the range of 
  * {@code [2, maximum]}. <p>
  * 
  * A prime sequence is {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link #unique() unique}. <p>
  *  
  * <b>Implementation notes</b>:<br>
  * Though this class is supposed to simulate complex behaviour that
  * has to be calculated on the fly to illustrate the use of 
  * sequences, the implementation uses a <b>very</b> simple version 
  * of the <b>Sieve of Eratosthenes</b>. The amount of memory used by
  * an a prime sequence is linear with {@code maximum}. Java has 
  * built-in support for generating primes in form of 
  * {@link java.math.BigInteger java.math.BigInteger}, but that is
  * beside the point. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ReversiblePrimeSequence
  */
public class PrimeSequence extends AbstractSequence<Integer> implements Sequence<Integer> {

    // Implementation note: the "Sieve of Eratosthenes" algorithm is 
    // identical with the ReversablePrimeSequence implementation, but the 
    // internal sequence design differs considerable between the two.
    
    /**
      * Maximum number to consider.  
      */
    private final int maximum;
    /**
      * The next index to start processing from on calls
      * to {@link #next()}. After each call to next has
      * completed, the value is {@link #prime} + 1. <p>
      * 
      * Range: [3, {@link #maximum}].
      */
    private int index;
    /**
      * The current prime number. <p>
      *   
      * Range: [2, {@link #maximum}].
      */
    private int prime;
    /**
      * The actual sieve. <p>
      * 
      * If {@link #calculate} is true, an index smaller than 
      * {@link #index} having a false value represents a prime 
      * number (normally a true value represent
      * a prime number, but by flipping it, we do not have to 
      * initialise the array with true). <p>
      * 
      * If {@code calculate} is false, any index having a false
      * value represents a prime number. <p>
      * 
      * The length is {@link #maximum} + 1. Index zero and one 
      * are unused.
      */
    private boolean[] sieve;
    /**
      * Returns false if all primes have been calculated, false if
      * not.
      */
    private boolean calculate;
    
    /**
      * Constructor.
      * 
      * @param  maximum The maximum number to consider. Must be larger than 1.
      * @throws IllegalArgumentException If {@code maximum} is smaller than 2.
      */
    public PrimeSequence(int maximum) {
        super(true); // reset
        if ((this.maximum = maximum) < 2) {
            throw new IllegalArgumentException("Maximum number must be > 2");
        }
        this.sieve = new boolean[this.maximum + 1];
        this.calculate = (this.maximum > 3);
        // Identify multiples of 2 as not prime, because 2 is the
        // initial value...
        for (int j = 2; j * 2 <= this.maximum; j++) {
            this.sieve[j * 2] = true; // not prime
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public PrimeSequence(PrimeSequence sequence) {
        super(sequence);
        this.calculate = sequence.calculate;
        this.maximum = sequence.maximum;
        this.prime = sequence.prime;
        this.index = sequence.index;
        this.sieve = new boolean[sequence.sieve.length];
        System.arraycopy(sequence.sieve, 2, this.sieve, 2, sequence.sieve.length - 2);
    }
    
    /**
      * 
      */
    public PrimeSequence copy() {
        return new PrimeSequence(this); 
    }
    
    /**
      * Returns true.
      *  
      * @return True. 
      */
    public boolean bounded() {
        return true;
    }
    
    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean consistent() {
        return true;
    }
    
    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * 
      */
    public Integer current() {
        return this.prime;
    }

    /**
      * 
      */
    public Integer next() {
        this.state = State.NORMAL;
        // Calculate...
        if (this.calculate) {
            for (; this.calculate = this.index * this.index <= this.maximum; this.index++) {
                // Identify multiples of "index" as not prime...
                if (!this.sieve[this.index]) {
                    for (int j = this.index; j * this.index <= this.maximum; j++) {
                        this.sieve[j * this.index] = true; // not prime
                    }
                    return this.prime = this.index++;
                }
            }
        }

        // Advance until next prime is found, i.e. a false value...
        while (this.index <= this.maximum) {
            // Not prime...
            if (this.sieve[this.index]) {
                this.index++;
            // Prime...
            } else {
                return this.prime = this.index++;
            }
        }
        
        // Restart...
        this.state = State.RESTART;
        this.index = 3;
        return this.prime = 2;
    }

    /**
      * 
      */
    public void reset() {
        super.reset();
        this.prime = 2;
        this.index = 3;
    }
   
}
