/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.model;

import dk.rode.thesis.prototype.StrictCopyable;
import dk.rode.thesis.strategy.Stringable;

/**
  * Represents a <i>sequence</i> that will deliver the 
  * {@link #next() next}, or {@link #current() current}, 
  * value in the sequence on demand. <p>
  * 
  * Sequences are more than mere iterators; they are <i>value
  * centric</i>. Sequences always have a lower bound, i.e. the 
  * initial sequence value, and may be {@link #bounded bounded}, 
  * limiting the number of possible values it can deliver. If a 
  * sequence is bounded and deliver {@link #consistent() consistent} 
  * values, it will <i>restart</i> when the upper bound is reached 
  * on an invocation of {@link #next()}, i.e. the same values will
  * be delivered again, in order. The sequence values of bounded, 
  * consistent sequences are thus deterministic: two instances of the
  * same sequence type initialised identically will return the same
  * sequence values if utilised in the same manner. Sequences 
  * may also deliver {@link #unique() unique} sequence values until 
  * reset or restarted. <p> 
  *  
  * Sequences can be {@link #reset() reset} explicitly, which will cause 
  * the sequence to restart if it is consistent. <p>
  * 
  * <b>Examples</b>:
  * <pre>
  *    Consistent, bounded, and unique number sequence:               [0, 1, 2, 3, 4, 5]
  *    
  *    Consistent, unbounded, and unique odd sequence:                [1, 3, 5, 7, 9, ..[
  *    
  *    Inconsistent, bounded, and not unique random sequence ([0,2]): [0, 2, 2, 1, 0, ..]
  *    
  *    Inconsistent, unbounded, and not unique odd*random sequence:   [1*0, 3*2, 5*2, 7*1, 9*0, 11*.., ..[
  *
  *               consistent       consistent       inconsistent     inconsistent
  *               bounded          unbounded        bounded          unbounded
  *               unique           unique           not unique       not unique
  *    current(): 0 <- low bound    1 <- low bound  0 <- low bound   0*1  =  0 <- inconsistent low bound
  *    next():    1                 3               2                2*3  =  0 <- not unique
  *    next():    2                 5               2 <- not unique  2*5  = 10
  *    next():    3                 7               1                1*7  =  7
  *    current(): 3                 7               1                1*7  =  7
  *    next():    4                 9               0                0*9  =  0
  *    next():    5 <- high bound  11               2                2*11 = 22
  *    next():    1 <- restart     13               0                0*13 =  0
  *    next():    2                15               1                1*15 = 15
  *    next():    3                17               2                2*17 = 34
  *    reset():   1 <- restart      1 <- restart    1 <- no restart  1*1  =  1 <- restart of odd only!
  *    next():    2                 3               2                2*3  =  6
  *    ..
  * </pre>
  * Notice that the initial current value is the lower sequence bound,
  * and that the first invocation of {@code next()} will return the 
  * <i>second</i> sequence value because of the requirement that a 
  * sequence <b>always</b> have a current value. <p>
  * 
  * Values from sequences are delivered on demand by calling 
  * {@link #next()}, which may calculate the value. Hence, different 
  * invocations to {@code next()} may not perform the same 
  * amount of work. Once a call to {@code next()} completes, 
  * the value returned is considered the <i>current</i> 
  * value of the sequence and can be retrieved with 
  * {@link #current()} until {@code next()} or {@link #reset()} 
  * is invoked. <p>
  * 
  * A sequence will never deliver a null element from the
  * {@link #current()} and {@link #next()} methods; it <b>always</b> 
  * has a given non-null value, even if {@code next()} has not 
  * been invoked yet (the initial value which is the lower bound). <p>
  * 
  * The current internal state of a sequence can be acquired
  * via the {@link #state()} method. <p>
  * 
  * Unless otherwise stated, sequences are considered equal based 
  * on identity (==). 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
public interface Sequence<E> extends StrictCopyable<Sequence<E>>, Stringable<Sequence<E>> { 
   
    /**
      * The possible <i>internal</i> states a {@link Sequence} can have. <p>
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      * 
      * @see    Sequence#state()
      */
    public static enum State {
        /**
          * Initial internal state: neither {@link Sequence#next()}, nor 
          * {@link Sequence#reset()} has been invoked yet. <p>
          * 
          * However, a {@link Sequence#current() current} value 
          * can still be obtained as sequences never return 
          * null values. 
          */
        START, 
        /**
          * A {@link #bounded() bounded} and {@link #consistent() consistent}
          * sequence will have this internal state if it has restarted its 
          * sequence of values.   
          */
        RESTART, 
        /**
          * Internal state explicitly set using {@link Sequence#reset()}. 
          */
        RESET, 
        /**
          * Internal state of any sequence unless {@link #START}, {@link #RESET},
          * or {@link #RESTART} applies to it.   
          */
        NORMAL;
    }
    
    /**
      * Resets this sequence to start over if it is consistent. <p>
      * 
      * If this sequence is {@link #consistent() consistent}, the 
      * sequence will restart.
      */
    public void reset();
    
    /**
      * Returns the <i>next</i> element from this sequence. <p>
      *   
      * @return The next element; never null.
      * @see    #current()
      * @see    #state()
      */
    public E next();
    
    /**
      * Returns the <i>current</i> element from this sequence. <p>
      *
      * This method can be invoked even if {@link #next()}
      * has not been invoked yet, thus delivering the initial
      * value of this sequence.
      * 
      * @return The current element; never null.
      */
    public E current();

    /**
      * Returns true if this sequence is <i>bounded</i>, i.e. 
      * deliver values between the initial sequence value and
      * some upper bound. <p>
      * 
      * The same type of sequence may represent both bounded and 
      * unbounded sequences and the behaviour is determined and fixed 
      * at construction time.
      * 
      * Bounded sequences will <i>restart</i> if they deliver 
      * {@link #consistent() consistent} values. 
      *  
      * @return True if bounded, false if not.
      * @see    #unique()
      */
    public boolean bounded();
    
    /**
      * Returns true if this sequence is <i>consistent</i>, i.e. deliver
      * the same values, in order, after restart or {@link #reset() reset}. <p>
      * 
      * Only {@link #bounded() bounded} consistent sequences will restart.
      * Consistent sequences need not deliver {@link #unique() unique}
      * sequence values. <p> 
      * 
      * Instances of the same type of sequences are always consistent or
      * inconsistent; it is determined at design time, not construction time. 
      * 
      * @return True if consistent, false if not.
      */
    public boolean consistent();

    /**
      * Returns true if this sequence deliver a given sequence value at most
      * one time until {@link #reset() reset} or restarted. <p>
      * 
      * {@link #bounded() Unbounded} sequences that are unique will never 
      * deliver the same sequence value more than once. <p>
      *  
      * The same type of sequence may represent both unique and 
      * not unique sequences and the behaviour is determined and fixed 
      * at construction time.
      *  
      * @return True if unique, false if not.
      * @see    #consistent()
      */
    public boolean unique();
    
    /**
      * Returns the <i>internal</i> state of this sequence.
      * 
      * @return The internal state; never null.
      */
    public State state();  
    
    /**
      * Returns a copy of this sequence that will start at the
      * same sequence index as this sequence. 
      *  
      * @return A copy of this sequence; never null.
      */
    public Sequence<E> copy();
    
}
