/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
  * A <i>primitive constant</i> associates a primitive type with its wrapper class. <p>
  * 
  * For example, the primitive type {@code int} is represented by the 
  * {@code java.lang.Integer.TYPE} class, and the associated wrapper class 
  * is the {@code java.lang.Integer} class. <p>
  *  
  * The order of declaration is as follows: <ul><p>
  * <li> {@link #Integer}
  * <li> {@link #Long}
  * <li> {@link #Double}
  * <li> {@link #Float}
  * <li> {@link #Boolean}
  * <li> {@link #Character}
  * <li> {@link #Byte}
  * <li> {@link #Short}
  * </ul><p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public enum Primitive {

    /**
      * Constant corresponding to the primitive type {@code int}, 
      * represented by the {@code java.lang.Integer.TYPE} class, 
      * and the non-primitive wrapper class {@code java.lang.Integer}. <p>
      * 
      * This type is compatible with the following types: <ul><p>
      * 
      * <li> {@link #Integer}
      * <li> {@link #Float}
      * <li> {@link #Long}
      * <li> {@link #Double}
      * </ul>
      */
    Integer(java.lang.Integer.TYPE, java.lang.Integer.class) {
        public Integer getDefaultValue() {
            return 0;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Integer:
                case Float:
                case Long:
                case Double:
                    return true;
                default:
                    return false;
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Integer i = (Integer)value;
            switch (primitive) {
                case Integer:
                    return i;
                case Long:
                    return new java.lang.Long(i);
                case Float:
                    return new java.lang.Float(i);
                default: // Double
                    return new java.lang.Double(i); 
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code long} 
      * ({@code java.lang.Long.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Long}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Long}
      * <li> {@link #Double}
      * </ul>
      */
    Long(java.lang.Long.TYPE, java.lang.Long.class) {
        public Long getDefaultValue() {
            return 0L;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Long:
                case Double:
                    return true;
                default:
                    return false;
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Long l = (Long)value;
            switch (primitive) {
                case Long:
                    return l;
                default: // Double
                    return new java.lang.Double(l);
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code double} 
      * ({@code java.lang.Double.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Double}. <p>
      * 
      * This type is only compatible with itself.
      */
    Double(java.lang.Double.TYPE, java.lang.Double.class) {
        public Double getDefaultValue() {
            return 0.0D;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            return (this == primitive);
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            return value;
        }
    },
    
    /**
      * Corresponding to the primitive type {@code float} 
      * ({@code java.lang.Float.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Float}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Float}
      * <li> {@link #Double}
      * </ul><p>
      * 
      * Note that neither {@link #Integer} or {@link #Long} are not compatible,
      * since the decimal points would then be truncated. 
      */
    Float(java.lang.Float.TYPE, java.lang.Float.class) {
        public Float getDefaultValue() {
            return 0.0F;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Float:
                case Double:
                    return true;
                default:
                    return false;
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Float f = (Float)value;
            switch (primitive) {
                case Float:
                    return f;
                default: // Double
                    return new java.lang.Double(f); 
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code boolean} 
      * ({@code java.lang.Boolean.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Boolean}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Boolean}
      * <li> {@link #Byte}
      * <li> {@link #Short}
      * <li> {@link #Integer}
      * <li> {@link #Long}
      * <li> {@link #Float}
      * <li> {@link #Double}
      * </ul><p>
      * 
      * A value of {@code true} will be translated to one (1), and 
      * a value of {@code false} will be translated to zero (0). 
      */
    Boolean(java.lang.Boolean.TYPE, java.lang.Boolean.class) {
        public Boolean getDefaultValue() {
            return false;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            return (primitive != Character);
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Boolean b = (Boolean)value;
            switch (primitive) {
                case Boolean:
                    return b;
                case Byte:
                    return new java.lang.Byte((byte)(b ? 1 : 0));
                case Short:
                    return new java.lang.Short((short)(b ? 1 : 0)); 
                case Integer:
                    return new java.lang.Integer(b ? 1 : 0);
                case Long:
                    return new java.lang.Long(b ? 1 : 0);
                case Float:
                    return new java.lang.Float(b ? 1 : 0);
                default: // Double
                    return new java.lang.Double(b ? 1 : 0);
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code char} 
      * ({@code java.lang.Character.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Character}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Character}
      * <li> {@link #Short}
      * <li> {@link #Integer}
      * <li> {@link #Long}
      * <li> {@link #Float}
      * <li> {@link #Double}
      * </ul>
      */
    Character(java.lang.Character.TYPE, java.lang.Character.class) {
        public Character getDefaultValue() {
            return '\u0000';
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Boolean:
                case Byte:
                    return false;
                default:
                    return true;   
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Character c = (Character)value;
            switch (primitive) {
                case Character:
                    return c;
                case Short:
                    return new java.lang.Short((short)c.charValue());
                case Integer:
                    return new java.lang.Integer(c);
                case Long:
                    return new java.lang.Long(c);
                case Float:
                    return new java.lang.Float(c);
                default:
                    return new java.lang.Double(c);
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code byte} 
      * ({@code java.lang.Byte.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Byte}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Byte}
      * <li> {@link #Short}
      * <li> {@link #Integer}
      * <li> {@link #Long}
      * <li> {@link #Float}
      * <li> {@link #Double}
      * </ul>
      */
    Byte(java.lang.Byte.TYPE, java.lang.Byte.class) {
        public Byte getDefaultValue() {
            return (byte)0;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Boolean:
                case Character:
                    return false;
                default:
                    return true;
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Byte b = (Byte)value;
            switch (primitive) {
                case Byte:
                    return b;
                case Short:
                    return new java.lang.Short(b);
                case Integer:
                    return new java.lang.Integer(b);
                case Long:
                    return new java.lang.Long(b);
                case Float:
                    return new java.lang.Float(b);
                default: // Double
                    return new java.lang.Double(b); 
            }
        }
    },
    
    /**
      * Corresponding to the primitive type {@code short} 
      * ({@code java.lang.Short.TYPE}) and the non-primitive 
      * wrapper class {@code java.lang.Short}. <p>
      * 
      * This type is compatible with the following primitives: <ul><p>
      * 
      * <li> {@link #Short}
      * <li> {@link #Character}
      * <li> {@link #Integer}
      * <li> {@link #Long}
      * <li> {@link #Float}
      * <li> {@link #Double}
      * </ul>
      */
    Short(java.lang.Short.TYPE, java.lang.Short.class) {
        public Short getDefaultValue() {
            return (short)0;
        }
        public boolean isCompatibleWith(Primitive primitive) {
            switch (primitive) {
                case Byte:
                case Boolean:
                    return false;
                default:
                    return true;
            }
        }
        protected Object asPrimitive(Object value, Primitive primitive) {
            Short s = (Short)value;
            switch (primitive) {
                case Short:
                    return s;
                case Character:
                    return new java.lang.Character((char)s.shortValue());
                case Integer:
                    return new java.lang.Integer(s);
                case Long:
                    return new java.lang.Long(s);
                case Float:
                    return new java.lang.Float(s);
                default: // Double
                    return new java.lang.Double(s); 
            }
        }
    };
   
    /**
      * The primitive type associated with this enumeration type, e.g. 
      * {@code java.lang.Integer.TYPE} for the primitive type {@code int}. <p>
      *  
      * Never null.
      */
    public final Class<?> primitive;
    /**
      * The non-primitive wrapper class associated with this enumeration type. <p>
      *  
      * Never null.
      */
    public final Class<?> nonPrimitive;
    
    /**
      * Constructor, which creates this enumeration type associated with
      * the primitive and non-primitive classes {@code primitive}
      * and {@code nonPrimitive}, respectively.
      * 
      * @param  primitive    The primitive class; cannot be null.
      * @param  nonPrimitive The non-primitive wrapper class; cannot be null.
      * @throws NullPointerException If either parameter is null.
      * @throws IllegalArgumentException If {@code primitive} is a
      *         non-primitive type, or if {@code nonPrimitive} is
      *         a primitive type.
      */
    Primitive(Class<?> primitive, Class<?> nonPrimitive) {
        if (!(this.primitive = primitive).isPrimitive()) {
            throw new IllegalArgumentException("Primitive class cannot be non-primitive");
        } else if ((this.nonPrimitive = nonPrimitive).isPrimitive()) {
            throw new IllegalArgumentException("Non-primitive class cannot be primitive");
        }
    }
    
    /**
      * Returns the default primitive value of the primitive type 
      * associated with this enumeration type, wrapped in a wrapper class.
      *  
      * @return The default primitive value; never null.
      */
    public abstract Object getDefaultValue();
    
    /**
      * Returns true if the non-primitive type of {@code type}
      * represents the same non-primitive type as this enumeration type 
      * does, false if not.
      * 
      * @param  type The type to test; can be null in which case
      *              false is returned.
      * @return True if equal, false if not.
      */
    public boolean isEqualIgnorePrimitives(Class<?> type) {
        if (type == null) {
            return false;
        }
        return this.nonPrimitive.equals(getNonPrimitiveType(type));
    }
    
    /**
      * Returns true if the non-primitive type of this enumeration type is 
      * assignable from the non-primitive type of {@code type}. <p>
      * 
      * @param  type The type to test; cannot be null.
      * @return True if the non-primitive type of this type is assignable 
      *         from the non-primitive type of {@code type}, false if 
      *         not.
      * @throws NullPointerException If {@code type} is null.
      */
    public boolean isAssignableFromIgnorePrimitives(Class<?> type) {
        return this.nonPrimitive.isAssignableFrom(getNonPrimitiveType(type));
    }
    
    /**
      * Returns true if the type supplied as {@code primitive} is 
      * <i>compatible</i> with this enumeration type, false if not. <p>
      * 
      * {@code primitive} is compatible if an actual primitive value of this 
      * type can be (implicitly) represented by {@code primitive} without loss 
      * of precision. For example, the primitive {@code int} value stored 
      * in a {@code java.lang.Integer} wrapper can also be represented by
      * by a {@code java.lang.Long} wrapper without loss of precision. <p>
      * 
      * If {@code primitive} is compatible with this type, it is safe to invoke 
      * {@link #asPrimitive(Object, Class)} on this type the non-primitive type 
      * associated with {@code primitive} as the second argument.
      *  
      * @param  primitive The primitive to test; cannot be null.
      * @return True if compatible, false if not.
      * @throws NullPointerException If {@code primitive} is null.
      */
    public abstract boolean isCompatibleWith(Primitive primitive);
 
    /**
      * Returns true if the type corresponding to {@code type} is 
      * <i>compatible</i> with this enumeration type, false if not. <p>
      * 
      * The corresponding enumeration type is compatible if an actual 
      * primitive value of this type can be (implicitly) represented by the
      * corresponding enumeration type without loss of precision. For example, 
      * the primitive {@code int} value stored in a {@code java.lang.Integer} 
      * wrapper can also be represented by by a {@code java.lang.Long} wrapper 
      * without loss of precision. <p>
      * 
      * If the corresponding enumeration type is compatible with this type, it 
      * is safe to invoke {@link #asPrimitive(Object, Class)} on this type using 
      * {@code type} as the second argument.
      *  
      * @param  type The type to test; cannot be null.
      * @return True if compatible, false if not.
      * @throws NullPointerException If {@code type} is null.
      * @see    #getPrimitive(Class)
      */
    public boolean isCompatibleWith(Class<?> type) {
        Primitive p = getPrimitive(type);
        if (p != null) {
            return this.isCompatibleWith(p);
        }
        return false;
    }
     
    /**
      * Returns the value supplied as {@code value} wrapped in the non-primitive
      * type corresponding to the (primitive) type supplied as {@code type}. <p>
      * 
      * {@code value} must have the type stored as the non-primitive type for 
      * this enumeration type.
      * 
      * @param  <T>   The desired type.
      * @param  value The value; can be null.
      * @param  type  The type as a class literal; cannot be null.
      * @return The value as the compatible type supplied as {@code type}; can be null.
      * @throws NullPointerException If {@code type} is null.
      * @throws IllegalArgumentException If {@code type} is not compatible with
      *         this type, or if {@code value} is not null and not an instance of 
      *         the non-primitive type of this type.
      */
    public final <T> T asPrimitive(Object value, Class<T> type) {
        if (type == null) {
            throw new NullPointerException("Primitive cannot be null");
        }
        if (this.isCompatibleWith(type)) {
            throw new IllegalArgumentException("Primitive is not compatible: " + type.getSimpleName() + " <! " + this.nonPrimitive.getSimpleName());
        } else if (value == null) {
            return null;
        } else if (!this.nonPrimitive.isInstance(value)) {
            throw new IllegalArgumentException("Illegal value type: " + value.getClass().getSimpleName() + " <! " + this.nonPrimitive.getSimpleName());
        }
        return type.cast(this.asPrimitive(value, getPrimitive(type))); // getPrimitive known to succeed!
    }
    
    /**
      * Returns the value supplied as {@code value} wrapped in the non-primitive
      * type stored in in {@code primitive}. <p>
      * 
      * @param  value     The value; never null and known to be of correct type.
      * @param  primitive The type to represent {@code value} as; never null
      *                   and known to be compatible.
      * @return The value as the compatible type; never null.
      */
    protected abstract Object asPrimitive(Object value, Primitive primitive);
    
    /**
      * Returns the default primitive value of the primitive (wrapper) type 
      * supplied as {@code type}, wrapped in an appropriate wrapper class, 
      * or null if {@code type} is neither primitive nor a primitive wrapper 
      * class.
      *  
      * @param  type The type in question; cannot be null. 
      * @return The default primitive value; or null if neither primitive nor
      *         a primitive wrapper class.
      * @throws NullPointerException If {@code type} is null.
      * @see    #getDefaultValue()
      */
    public final static Object getDefaultValue(Class<?> type) {
        Primitive p = getPrimitive(type);
        return p == null ? null : p.getDefaultValue();
    }
    
    /**
      * Merges {@code primitive} with {@code primitives}. <p>
      * 
      * Null {@code Primitive} objects are ignored.
      * 
      * @param  primitive  The first primitive; can be null (ignored).
      * @param  primitives The remaining primitives; can contain null entries (ignored).
      * @return An array containing the (non-null) {@code Primitive} objects;
      *         never null.
      */
    private final static Primitive[] merge(Primitive primitive, Primitive... primitives) {
        List<Primitive> list = new ArrayList<Primitive>(1 + primitives.length);
        if (primitive != null) {
            list.add(primitive);
        }
        for (Primitive p : primitives) {
            if (p != null) {
                list.add(p);
            }
        }
        Primitive[] array = new Primitive[list.size()];
        return list.toArray(array);
    }
    
    /**
      * Returns an array containing the primitives corresponding to the 
      * primitives supplied in {@code primitives}. <p>
      * 
      * If {@code primitives} is null, or has zero length, all primitive
      * primitives will be returned. The order corresponds to the order of
      * the {@link Primitive} enumeration. <p>
      * 
      * Primitives having no associated primitive type are ignored, and
      * {@code primitives} may contain duplicates. Primitive primitives are 
      * included as is. <p>
      * 
      * The order is maintained in the returned array.
      *
      * @param  primitives An array of primitives; can be null.
      * @return An array containing the corresponding primitives, 
      *         if any, in the same order as {@code primitives}; never null.
      */
    public final static Class<?>[] getPrimitiveTypes(Class<?>... primitives) {
        if ((primitives == null) || (primitives.length == 0)) {
            Primitive[] ps = Primitive.values();
            Class<?>[] classes = new Class[ps.length];
            for (int i = 0; i < ps.length; i++) {
                classes[i++] = ps[i].primitive;
            }
            return classes;
            
        } else {
            List<Class<?>> classes = new ArrayList<Class<?>>(primitives.length);
            for (Class<?> clazz : primitives) {
                if ((clazz = getPrimitiveType(clazz)) != null) {
                    classes.add(clazz);
                }
            }
            Class<?>[] array = new Class[classes.size()];
            classes.toArray(array);
            return array;
        }
    }

    /**
      * Returns an array containing the primitives corresponding to the 
      * enumeration primitives supplied in {@code primitive} and {@code primitives}. <p>
      * 
      * If {@code primitive} is null and {@code primitives} is null or empty,
      * the returned array will contain all primitives. The order 
      * corresponds to the order of the {@link Primitive} enumeration. <p>
      * 
      * {@code primitive} and {@code primitives} may contain duplicates. <p> 
      * 
      * The order is maintained in the returned array.
      *
      * @param  primitive  The first primitive; can be null.
      * @param  primitives The last primitives; can contain null entries. 
      * @return An array containing the corresponding primitives, 
      *         in the same order as defined by {@code primitive} and 
      *         {@code primitives}; never null.
      */
    public final static Class<?>[] getPrimitiveTypes(Primitive primitive, Primitive... primitives) { // "type" parameter added to make the signature unique  
        if ((primitives = merge(primitive, primitives)).length == 0) {
            return getPrimitiveTypes();
            
        } else {
            Class<?>[] classes = new Class[primitives.length];
            for (int i = 0; i < primitives.length; i++) {
                classes[i] = primitives[i].primitive;
            }
            return classes;
        }
    }
    
    /**
      * Returns an array of the non-primitives corresponding to the 
      * primitives supplied in {@code primitives}. <p>
      * 
      * If {@code primitives} is null, or has zero length, all non-primitive
      * primitives having an associated primitive type will be returned. The 
      * order corresponds to the order of the {@link Primitive} enumeration. <p>
      * 
      * If {@code primitives} is not null, the length of the returned array
      * is always equal to the length of {@code primitives}. {@code primitives} 
      * may contain duplicates. Non-primitives are included as is. <p>
      * 
      * The order is maintained in the returned array.
      *
      * @param  primitives An array of primitives; can be null.
      * @return An array containing the corresponding non-primitives, 
      *         in the same order as {@code primitives}; never null.
      */
    public final static Class<?>[] getNonPrimitiveTypes(Class<?>... primitives) {
        if ((primitives == null) || (primitives.length == 0)) {
            Primitive[] ps = Primitive.values();
            Class<?>[] classes = new Class[ps.length];
            for (int i = 0; i < ps.length; i++) {
                classes[i++] = ps[i].nonPrimitive;
            }
            return classes;
            
        } else {
            Class<?>[] classes = new Class[primitives.length];
            for (int i = 0; i < primitives.length; i++) {
                classes[i] = getNonPrimitiveType(primitives[i]);
            }
            return classes;
        }
    }

    /**
      * Returns an array containing the non-primitives corresponding to the 
      * enumeration primitives supplied in {@code primitive} and {@code primitives}. <p>
      * 
      * If {@code primitive} is null and {@code primitives} is null or empty,
      * the returned array will contain all non-primitives having
      * an associated primitive type. The order corresponds to the declaration
      * order of this enumeration. <p>
      * 
      * {@code primitive} and {@code primitives} may contain duplicates. <p>
      * 
      * The order is maintained in the returned array.
      *
      * @param  primitive  The first primitive; can be null.
      * @param  primitives The last primitives; can contain null entries.
      * @return An array containing the corresponding non-primitives, 
      *         in the same order as defined by {@code primitive} and 
      *         {@code primitives}; never null.
      */
    public final static Class<?>[] getNonPrimitiveTypes(Primitive primitive, Primitive... primitives) { // "type" parameter added to make the signature unique
        if ((primitives = merge(primitive, primitives)).length == 0) {
            return getNonPrimitiveTypes();
            
        } else {
            Class<?>[] classes = new Class<?>[primitives.length];
            for (int i = 0; i < primitives.length; i++) {
                classes[i] = primitives[i].nonPrimitive;
            }
            return classes;
        }
    }
    
    /**
      * Returns a {@code Map} where the keys are the primitive
      * primitives, and the values the associated non-primitives.
      * 
      * @return A map; never null.
      */
    public final static Map<Class<?>,Class<?>> getPrimitivesMap() {
        Map<Class<?>,Class<?>> map = new HashMap<Class<?>,Class<?>>();
        for (Primitive p : Primitive.values()) {
            map.put(p.primitive, p.nonPrimitive);
        }
        return map;
    }
    
    /**
      * Returns a {@code Map} where the keys are the non-primitive
      * primitives having an associated primitive type, which is also
      * the value associated with a key.
      * 
      * @return A map; never null.
      */
    public final static Map<Class<?>,Class<?>> getNonPrimitivesMap() {
        Map<Class<?>,Class<?>> map = new HashMap<Class<?>,Class<?>>();
        for (Primitive p : Primitive.values()) {
            map.put(p.nonPrimitive, p.primitive);
        }
        return map;
    }
    
    /**
      * Returns the primitive type corresponding to the
      * type supplied as {@code type}, if any. If {@code type}
      * is primitive, it is simply returned. <p>
      * 
      * If no such primitive type can be found, null will
      * be returned.
      * 
      * @param  type The type to find the corresponding primitive
      *              type for; cannot be null.
      * @return The corresponding primitive type, or null.
      * @throws NullPointerException If {@code type} is null.
      */
    public final static Class<?> getPrimitiveType(Class<?> type) {
        if (type.isPrimitive()) {
            return type;
        } 
        for (Primitive p : Primitive.values()) {
            if (p.nonPrimitive.equals(type)) {
                return p.primitive;
            }
        }
        // No corresponding primitive type...
        return null;
    }

    /**
      * Returns the primitive type corresponding to the
      * enumeration type supplied as {@code primitive}; for 
      * example {@code int} ({@code java.lang.Integer.TYPE}) 
      * for {@link #Integer}.
      * 
      * @param  primitive The primitive; cannot be null.
      * @return The corresponding primitive type; never null.
      * @throws NullPointerException If {@code primitive} is null.
      */
    public final static Class<?> getPrimitiveType(Primitive primitive) {
        return primitive.primitive;
    }
    
    /**
      * Returns the non-primitive type corresponding to the
      * (primitive) type supplied as {@code type}, if any. <p>
      *  
      * If {@code type} is non-primitive, it is simply returned.
      * 
      * @param  type The type to find the corresponding non-primitive
      *              type for; cannot be null.
      * @return The corresponding non-primitive type; never null.
      * @throws NullPointerException If {@code type} is null.
      */
    public final static Class<?> getNonPrimitiveType(Class<?> type) {
        if (!type.isPrimitive()) {
            return type;
        }
        for (Primitive p : Primitive.values()) {
            if (p.primitive.equals(type)) {
                return p.nonPrimitive;
            }
        }
        // Will never be reached, since all primitives
        // are known and "type" is known to be one of them...
        return null;
    }

    /**
      * Returns the non-primitive type corresponding to the
      * enumeration type supplied as {@code primitive}; for example
      * {@code java.lang.Integer} for {@link #Integer}. 
      *  
      * @param  primitive The primitive; cannot be null.
      * @return The corresponding non-primitive type; never null.
      * @throws NullPointerException If {@code primitive} is null.
      */
    public final static Class<?> getNonPrimitiveType(Primitive primitive) {
        return primitive.nonPrimitive;
    }
    
    /**
      * Returns the enumeration type corresponding to {@code type},
      * if any; for example {@link #Integer} for either
      * {@code java.lang.Integer} or {@code int} 
      * ({@code java.lang.Integer.TYPE}). 
      * 
      * @param  type The type; cannot be null.
      * @return The corresponding enumeration type, or null.
      * @throws NullPointerException If {@code type} is null.
      */
    public final static Primitive getPrimitive(Class<?> type) {
        if (type.isPrimitive()) {
            for (Primitive p : Primitive.values()) {
                if (p.primitive.equals(type)) {
                    return p;
                }
            }
        } else {
            for (Primitive p : Primitive.values()) {
                if (p.nonPrimitive.equals(type)) {
                    return p;
                }
            }
        }
        return null;
    }

    /**
      * Returns true if and only if: <ol><p>
      * 
      * <li> {@code a} and {@code b} are both null; <b>or</b><p>
      * 
      * <li> {@code a} and {@code b} have the same non-primitive 
      *      type.
      *      
      * </ol>
      * 
      * @param  a The first type; can be null.
      * @param  b The second type; can be null.
      * @return True if equal, false if not.
      */
    public final static boolean isEqualIgnorePrimitives(Class<?> a, Class<?> b) {
        if (a == null) {
            return (b == null);
        }
        return getNonPrimitiveType(a).equals(getNonPrimitiveType(b));
    }

    /**
      * Returns true if and only if: <ol><p>
      * 
      * <li> {@code a} and {@code b} are both null or identical (==); <b>or</b><p>
      * 
      * <li> {@code a} and {@code b} contains the same number of elements and 
      *      primitives at the same index in {@code a} and {@code b} have the same
      *      non-primitive type (using {@code equals}).
      *      
      * </ol>
      * 
      * @param  a The first array; can be null.
      * @param  b The second array; can be null.
      * @return True if equal, false if not.
      * @throws NullPointerException If {@code a} or {@code b} contains
      *         a null entry.
      */
    public final static boolean isEqualIgnorePrimitives(Class<?>[] a, Class<?>[] b) {
        if (a == null) {
            return (b == null);
        } else if (b == null) {
            return false;
        } else if (a.length != b.length) {
            return false;
        } else if (a == b) {
            return true;
        }
        for (int i = 0; i < a.length; i++) {
            if (!getNonPrimitiveType(a[i]).equals(getNonPrimitiveType(b[i]))) {
                return false;
            }
        }
        return true;
    }
    
    /**
      * Returns true if {@code a} is assignable from {@code b} when primitive
      * primitives are ignored. <p>
      * 
      * For example, if {@code a} is a {@code Number} and {@code b} is 
      * {@code int} ({@code Integer.TYPE}), then the normal assignable test
      * would fail. Here, {@code int} is treated as an {@code Integer}, and
      * the assignable test succeeds.
      *    
      * @param  a The first type; cannot be null.
      * @param  b The second type; cannot be null.
      * @return True if {@code a} is assignable from {@code b} ignoring
      *         primitives, false if not.
      * @throws NullPointerException If either argument is null.         
      */
    public final static boolean isAssignableFromIgnorePrimitives(Class<?> a, Class<?> b) {
        return getNonPrimitiveType(a).isAssignableFrom(getNonPrimitiveType(b));
    }

    /**
      * Returns true if and only if: <ol><p>
      * 
      * <li> {@code a} and {@code b} are both null or identical (==); <b>or</b><p>
      * 
      * <li> {@code a} and {@code b} contains the same number of elements, 
      *      and each class in {@code a} is assignable from the class in 
      *      {@code b} at the corresponding index ignoring primitives, 
      *      respectively. <p> 
      *      
      * </ol>
      * 
      * For example, if {@code a[x]} is a {@code Number} and {@code b[x]} is 
      * {@code int} ({@code Integer.TYPE}), then the normal assignable test
      * would fail. Here, {@code int} is treated as an {@code Integer}, and
      * the assignable test succeeds.
      * 
      * @param  a The first array; can be null.
      * @param  b The second array; can be null.
      * @return True if each class in {@code a} is assignable from each class
      *         in {@code b} at the same index, respectively, false if not.
      * @throws NullPointerException If {@code a} or {@code b} contains
      *         a null entry.
      */
    public final static boolean isAssignableFromIgnorePrimitives(Class<?>[] a, Class<?>[] b) {
        if (a == null) {
            return (b == null);
        } else if (b == null) {
            return false;
        } else if (a.length != b.length) {
            return false;
        } else if (a == b) {
            return true;
        }
        for (int i = 0; i < a.length; i++) {
            if (!getNonPrimitiveType(a[i]).isAssignableFrom(getNonPrimitiveType(b[i]))) {
                return false;
            }
        }
        return true;
    }
    
    /**
      * Returns the string representing of {@code object} that ensures
      * that arrays of primitives are formatted correctly. <p>
      * 
      * If {@code object} is null, the string {@code null} is returned. 
      * If the class of {@code object} is not an array type, the 
      * {@code object.toString()} is returned. <p> 
      * 
      * If the class of {@code object} is an array and the component type
      * is primitive, the the proper {@code Arrays.toString(..)} method
      * is invoked, e.g. {@code Arrays.toString(boolean[])} for the 
      * boolean primitive type. If the component type is not primitive, the 
      * {@code Arrays.toString(Object[])} method is used. 
      * 
      * @param  object The object to format into a string; can be null.
      * @return The string representation; never null.
      */
    public final static String toString(Object object) {
        if (object == null) {
            return "null";
        }
        Class<?> clazz = object.getClass();
        if (clazz.isArray()) {
            Class<?> componentType = clazz.getComponentType();
            if (componentType.isPrimitive()) {
               switch (getPrimitive(componentType)) {
                   case Boolean:   return Arrays.toString((boolean[])object);
                   case Byte:      return Arrays.toString((byte[])object);
                   case Character: return Arrays.toString((char[])object);
                   case Double:    return Arrays.toString((double[])object);
                   case Float:     return Arrays.toString((float[])object);
                   case Integer:   return Arrays.toString((int[])object);
                   case Long:      return Arrays.toString((long[])object);
                   case Short:     return Arrays.toString((short[])object);
               }
            }
            return Arrays.toString((Object[])object);
        } 
        return object.toString();
    }
     
}
