/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.util.Collection;
import java.util.Set;

import dk.rode.thesis.meta.Participant;

/**
  * An <i>aspect observable object</i> is an object that has a 
  * collection of associated <i>observers</i> to be notified in
  * case the object changes its internal state, or <i>aspect</i>, 
  * and where the observers can subscribe certain aspects only. <p>
  *
  * This interface only supplies the logic for 
  * {@link #addObserver(Object, Object) adding}, 
  * {@link #removeObserver(Object, Object) 
  * removing}, and {@link #getObservers(Object) fetching} associated
  * observers based on specific aspects, but no notification method 
  * to notify the observers. Using the get, add, and remove methods
  * from {@link ObservableSequence} gets, subscribes, or removes
  * regardless of aspect value (all possible aspect values). <p>
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    Observable
  * 
  * @param  <O> The type of observers associated with this observable. 
  * @param  <A> The type of aspect associated with this observable. 
  */
@Participant("Subject")
public interface AspectObservable<O,A> extends Observable<O> {

    /**
      * Returns a read-only set of the aspects the observer supplied
      * as {@code observer} subscribes to in this observable, if any.
      * 
      * @param  observer The observer; cannot be null.
      * @return A read-only list containing the aspect subscriptions; never null,
      *         but can be null.
      * @throws NullPointerException If {@code observer} is null.         
      */
    public Set<A> getAspects(O observer);
    
    /**
      * Returns the observers subscribed to the aspect supplied as {@code aspect}
      * in this observable.
      * 
      * @param  aspect The aspect; null means all.
      * @return The observers subscribing to the aspect supplied as {@code aspect};
      *         never null, but can be empty.
      */
    public Collection<O> getObservers(A aspect);
    
    /**
      * Adds the observer supplied as {@code observer} to this observable for
      * the aspect supplied as {@code aspect}, if not already. <p>
      * 
      * If {@code aspect} is null, {@code observer} will be subscribed
      * to all possible aspect values.
      * 
      * @param  observer The observer to add; cannot be null.
      * @param  aspect   The aspect to subscribe to; null means all.
      * @return True if {@code observer} was added, false if not.
      * @throws NullPointerException If {@code observer} is null.
      * @throws IllegalArgumentException If {@code observer} is not 
      *         a valid observer to add to this observable (optional).
      * @see    #removeObserver(Object, Object)         
      */
    public boolean addObserver(@Participant("ConcreteObserver") O observer, A aspect);
    
    /**
      * Removes the observer supplied as {@code observer} from this 
      * observable with regards to the aspect supplied as {@code aspect}, 
      * if already added to it. <p>
      * 
      * Trying to remove an observer that is not associated with
      * this observable or to an unsubscribed aspect has no effect. 
      * 
      * @param  observer The observer to remove for {@code aspect}; cannot be null.
      * @param  aspect   The aspect in question; null means all.
      * @return True if {@code observer} was removed from the subscription
      *         to {@code aspect} (or all if null), false if not.
      * @throws NullPointerException If {@code observer} is null.
      * @see    #addObserver(Object, Object)         
      */
    public boolean removeObserver(Object observer, A aspect);

}
