/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.text.DateFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>birthday registry</i> is a simple <i>application</i> that stores
  * the name and date of birth of people registered to it. <p>
  *  
  * Once a given date is reached in a given 
  * {@link ObservableSequence ObservableSequence&lt;SequenceObserver,Date>}, 
  * the associated names will be listed according to the current date 
  * received from the sequence in question. <p>
  * 
  * Birthdays can be supplied as strings that will be parsed
  * into actual dates based on a date format supplied at
  * construction time, or supplied directly as {@code Date} 
  * objects. <p>
  * 
  * A birthday registry is <b>not</b> thread-safe! <p>
  * 
  * <b>Usage:</b>
  * <pre>
  *    // Create...
  *    DateFormat df = new SimpleDateFormat("dd-MM-yyyy");
  *    BirthdayRegistry br = new {@link #BirthdayRegistry(DateFormat) BirthdayRegistry}(df);
  *    
  *    // Add birthday and associated name - only date, month,
  *    // and year are used, regardless of date format used...
  *    br.add("08-01-1942", "Stephen Hawking");
  *    br.add("07-08-1885", "Niels Bohr");
  *    ..
  *    
  *    // The registry does not flaunt that it is an observer, but uses an
  *    // inner class to implement observer functionality that simply 
  *    // invokes the proper private method in the registry. Any observable 
  *    // sequence delivering dates and using the SequenceObserver
  *    // interface can be utilised...
  *    {@link DateSequence} ds1 = new {@link DateSequence#DateSequence() DateSequence}();
  *    {@link AspectObservableSequence}&lt;SequenceObserver,State,Date> ds2 = ..
  *    
  *    // Register only on NORMAL and RESET state...
  *    State next  = {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL},
  *          reset = {@link dk.rode.thesis.meta.model.Sequence.State#RESET};
  *    br.{@link #register(AspectObservableSequence, Object...) register}(ds1, next, reset); // varargs
  *    // Register for all states...
  *    br.{@link #register(AspectObservableSequence, Object...) register}(ds2);
  *    
  *    // Will not be added twice...
  *    br.{@link #register(AspectObservableSequence, Object...) register}(ds2);
  * </pre>
  *
  * <b>Implementation notes</b>:<br>
  * This registry uses an inner class to implement the observer functionality, 
  * which ensures that we do not expose the {@link SequenceObserver observer}
  * interface for misuse. This is common practice in Java. However, since the 
  * inner class is in effect an {@linkplain dk.rode.thesis.adapter adapter}, the 
  * exact type of notification mechanism used by the observable sequence must 
  * be known. Here, the {@link SequenceObserver} interface is used, and this 
  * limits the type of applicable observable sequences to 
  * {@code SequenceObserver<SequenceObserver,Date>}.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteObserver")
public class BirthdayRegistry {

    /**
      * A <i>registry observer</i> forwards call to the enclosing 
      * registry. 
      * 
      * {@code hashCode} and {@code equals} are overridden to
      * ensure that the same registry is not added twice to
      * a given observable sequence. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This is an application of the {@linkplain dk.rode.thesis.adapter Adapter}
      * pattern, in effect adapting a {@link BirthdayRegistry} to 
      * the {@link SequenceObserver} interface. <p>
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      * 
      * @param  <A> The type of aspect used in the notification. The actual value
      *             is ignored.
      */
    private final class RegistryObserver<A> implements SequenceObserver<A> {

        /**
          * No-arg constructor.
          */
        private RegistryObserver() {
            super();
        }

        /**
          * Returns the enclosing birthday registry for this observer.
          *  
          * @return The registry; never null.
          */
        private final BirthdayRegistry registry() {
            return BirthdayRegistry.this;
        }
        
        /**
          * 
          */
        public void sequenceEvent(Sequence<?> sequence, A aspect) {
            Object current = sequence.current();
            if (current instanceof Date) {
                if (BirthdayRegistry.this.size() > 0) {
                    // Remember to normalise birthday to only contain
                    // date, month, and year...
                    BirthdayRegistry.this.checkDate(getFullBirthday((Date)current));
                }
            }
        }
        
        /**
          * Returns the hash code of this observer.
          * 
          * @return The hash code.
          */
        public int hashCode() {
            return BirthdayRegistry.this.hashCode(); // used by equals(Object)
        }
        
        /**
          * Returns true if {@code object} is a registry
          * observer wrapping the same registry as this
          * observer, false if not.
          * 
          * @param  object The object to test; can be null.
          * @return True if equal, false if not.
          */
        public boolean equals(Object object) {
            if (!(object instanceof RegistryObserver)) {
                return false;
            }
            RegistryObserver<?> ro = (RegistryObserver<?>)object;
            return BirthdayRegistry.this.equals(ro.registry());
        }
        
        /**
          * Returns the string representation of this observer.
          * 
          * @return The string representation; never null.
          */
        public String toString() {
            return "RegistryObserver[" + BirthdayRegistry.this + "]"; 
        }
        
    }
    
    /**
      * The map storing the birthdays:
      * <pre>
      *   (key, value) = (name,birthday) = (String,Date) 
      * </pre>
      */
    private final Map<String,Date> birthdays;
    /**
      * The date format used to parse string date representations
      * into actual dates. <p>
      * 
      * Never null.
      */
    private final DateFormat dateFormat;
    
    /**
      * Constructor.
      * 
      * @param  dateFormat The format used to parse string representations
      *                    of dates; cannot be null.
      * @throws NullPointerException If {@code dateFormat} is null.                       
      */
    public BirthdayRegistry(DateFormat dateFormat) {
        if ((this.dateFormat = dateFormat) == null) {
            throw new NullPointerException("Date formatter cannot be null");
        }
        this.birthdays = new HashMap<String,Date>();
    }

    /**
      * Copy constructor.
      * 
      * @param  registry The registry to copy; cannot be null.
      * @throws NullPointerException If {@code registry} is null.                       
      */
    public BirthdayRegistry(BirthdayRegistry registry) {
        this.birthdays = new HashMap<String,Date>(registry.birthdays);
        // Thread-safe issues!
        this.dateFormat = registry.dateFormat;
    }
    
    /**
      * Registers this registry as an <i>observer</i> for the 
      * {@link ObservableSequence observable} sequence supplied
      * as {@code sequence}, if not already.
      *
      * @param  <A>      The type of aspect; ignored by this registry.
      * @param  sequence The sequence; cannot be null.
      * @param  aspects  The aspects to subscribe to; can be or contain null, i.e. 
      *                  all events.
      * @return True if this registry was added as an observer, false if not.
      * @throws NullPointerException If {@code sequence} is null.         
      */
    public <A> boolean register(AspectObservableSequence<SequenceObserver<A>,A,Date> sequence, A... aspects) {
        RegistryObserver<A> ro = new RegistryObserver<A>();
        if ((aspects == null) || (aspects.length == 0)) {
            return sequence.addObserver(ro, null); // all
        }
        boolean added = false;
        for (A aspect : aspects) {
            added = sequence.addObserver(ro, aspect) || added; // may contain null as well
        }
        return added;
    }

    /**
      * Registers this registry as an <i>observer</i> for the 
      * {@link ObservableSequence observable} sequence supplied
      * as {@code sequence}, if not already.
      *
      * @param  <A>      The type of aspect; ignored by this registry.
      * @param  sequence The sequence; cannot be null.
      * @return True if this registry was added as an observer, false if not.
      * @throws NullPointerException If {@code sequence} is null.         
      */
    public <A> boolean register(ObservableSequence<SequenceObserver<A>,A,Date> sequence) {
        return sequence.addObserver(new RegistryObserver<A>());
    }
    
    /**
      * Normalises the name supplied as {@code name} to
      * be trimmed, and each unique part of the name 
      * separated by a single space is upper-cased first. 
      *  
      * @param  name The name to normalise; cannot be null.
      * @return A new normalised name; never null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty.
      */
    private final static String getName(String name) {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        } else if ((name = name.trim()) == null) {
            throw new IllegalArgumentException("Name cannot be empty");
        }
        StringBuilder sb = new StringBuilder();
        String[] parts = name.toLowerCase().split("\\s+");
        for (int i = 0; i < parts.length; i++) {
            sb.append(Character.toUpperCase(parts[i].charAt(0)));
            sb.append(parts[i].substring(1));
            if (i < parts.length - 1) {
                sb.append(" ");
            }
        }
        return sb.toString();
    }
    
    /**
      * Normalises the date supplied as {@code birthday} to
      * only include day and month. 
      *  
      * @param  birthday The birthday to normalise; cannot be null.
      * @return A new normalised date; never null.
      * @throws NullPointerException If {@code birthday} is null.
      */
    private final static Date getBirthday(Date birthday) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(birthday);
        calendar.clear(Calendar.YEAR);
        calendar.clear(Calendar.HOUR);
        calendar.clear(Calendar.MINUTE);
        calendar.clear(Calendar.SECOND);
        calendar.clear(Calendar.MILLISECOND);
        return calendar.getTime();
    }

    /**
      * Normalises the date supplied as {@code birthday} to
      * only include day, month, and year. 
      *  
      * @param  birthday The birthday to normalise; cannot be null.
      * @return A new normalised date; never null.
      * @throws NullPointerException If {@code birthday} is null.
      */
    private final static Date getFullBirthday(Date birthday) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(birthday);
        calendar.clear(Calendar.HOUR);
        calendar.clear(Calendar.MINUTE);
        calendar.clear(Calendar.SECOND);
        calendar.clear(Calendar.MILLISECOND);
        return calendar.getTime();
    }
    
    /**
      * Associates the person having the name and birthday
      * supplied as {@code name} and {@code birthday},
      * respectively, with this registry. <p>
      *  
      * An existing registration on {@code name} will be 
      * overwritten.
      * 
      * @param  birthday The birthday; cannot be null.
      * @param  name     The name; cannot be null, or empty.
      * @return True if this registration overwrites an existing one,
      *         false if not.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code name} is empty. 
      * @throws ParseException If {@code birthday} has an illegal
      *         format.
      */
    public boolean add(String birthday, String name) throws ParseException {
        return this.add(this.dateFormat.parse(birthday), name);
    }

    /**
      * Associates the person having the name and birthday
      * supplied as {@code name} and {@code birthday},
      * respectively, with this registry. <p>
      * 
      * An existing registration on {@code name} will be 
      * overwritten.
      *  
      * @param  birthday The birthday; cannot be null.
      * @param  name     The name; cannot be null, or empty.
      * @return True if this registration overwrites an existing one,
      *         false if not.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code name} is empty. 
      */
    public boolean add(Date birthday, String name) {
        return (this.birthdays.put(getName(name), getFullBirthday(birthday)) != null);
    }
    
    /**
      * Returns a  collection of names associated
      * with the birthday supplied as {@code birthday}. <p>
      * 
      * Modifying the returned collection will not affect
      * this registry.
      * 
      * @param  birthday The birthday; cannot be null.
      * @param  full     If true, the birthday must match on 
      *                  day, month, and year; if false, only
      *                  on day and month.
      * @return A list of names; never null, but can be empty.
      * @throws NullPointerException If {@code birthday} is null.
      * @throws ParseException If {@code birthday} has an illegal
      *         format.
      */
    public List<String> getNames(String birthday, boolean full) throws ParseException {
        return getNames(this.dateFormat.parse(birthday), full);
    }
    
    /**
      * Returns a  collection of names associated
      * with the birthday supplied as {@code birthday}. <p>
      * 
      * Modifying the returned collection will not affect
      * this registry.
      * 
      * @param  birthday The birthday; cannot be null.
      * @param  full     If true, the birthday must match on 
      *                  day, month, and year; if false, only
      *                  on day and month.
      * @return A list of names; never null, but can be empty.
      * @throws NullPointerException If {@code birthday} is null.
      */
    public List<String> getNames(Date birthday, boolean full) {
        birthday = full ? getFullBirthday(birthday) : getBirthday(birthday);
        List<String> names = new ArrayList<String>();        
        for (Entry<String,Date> e : this.birthdays.entrySet()) {
            // Invariant: dates stored are full dates...
            Date d = full ? e.getValue() : getBirthday(e.getValue()); 
            if (d.equals(birthday)) {
                names.add(e.getKey());
            }
        }
        return names;
    }
    
    /**
      * Returns the full birthday of the person having 
      * the name supplied as {@code name}, if registered. <p>
      * 
      * @param  name The name; cannot be null, or empty.
      * @return The birthday, or null if not registered.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty. 
      */
    public Date getBirthday(String name) {
        // Normalise name...
        return this.birthdays.get(getName(name));
    }

    /**
      * Returns the number of birthday dates, not people, registered
      * in this registry.
      *  
      * @return The number of birthday dates.
      */
    public int size() {
        return this.birthdays.size();
    }
    
    /**
      * Returns a map containing all the registered names and 
      * associated birthdays. <p>
      * 
      * Modifying the returned map will not affect this registry.
      * 
      * @return The map; never null, but may be empty.
      */
    public Map<String,Date> getAll() {
        // Cannot simply use Collections.unmodifiableMap, because
        // that will still expose the dates...
        Map<String,Date> map = new HashMap<String,Date>();
        for (String name : this.birthdays.keySet()) {
            map.put(name, new Date(this.birthdays.get(name).getTime()));
        }
        return map;
    }
    
    /**
      * Actual observer method that lists the names of persons
      * registered in this registry that were born on {@code date}, 
      * if any. <p>
      * 
      * Invariant: {@code date} is a {@link #getFullBirthday(Date) full} 
      * birthday date.
      * 
      * @param date The date; never null.
      * @see   #register(AspectObservableSequence, Object...)
      */
    private void checkDate(Date date) {
        // Get all on day and month...
        List<String> names = this.getNames(date, false);
        if (names.size() == 0) {
            return;
        }
        Log out = LogFactory.getLog(this);
        out.println("********************************");
        out.println("TADAA...!");
        Calendar c = Calendar.getInstance();
        c.setTime(date);
        
        out.println(this.dateFormat.format(date) + " matches the birthday of: ");
        // Sort names alphabetically...
        Collections.sort(names);
        for (String name : names) {
            out.print("- " + name + ", who ");
            Date full = this.birthdays.get(name);
            if (full.equals(date)) {
                out.println("was born on this date!");
            } else {
                Calendar cc = Calendar.getInstance();
                cc.setTime(full);
                int age = c.get(Calendar.YEAR) - cc.get(Calendar.YEAR);
                if (age > 0) {
                    out.println("was born on " + this.dateFormat.format(full) + " and is thus " + age + " year(s) old on this date!");
                } else {
                    out.println("in " + (-1 * age) + " years(s) will be born on the " + this.dateFormat.format(full) + "!");
                }
            }
        }
        out.println("********************************");
        out.println();
    }

    /**
      * Returns the string representation of this registry. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; registered: ").append(this.birthdays.size()).append(']').toString();
    }
    
}
