/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.observer;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.state.ReversiblePrimeSequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Observer</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Observer</b> tests.
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the test fails.             
      */
    public static void main(String[] args) throws Exception {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) throws Exception {
        boolean ok = true;
        
        out.heading("Observer tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        // True for verbose class logging...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(ObserverManager.class).setActive(log);
        LogFactory.getLog(AnnotatedObserversSequence.class).setActive(log);
        LogFactory.getLog(AnnotatedObserversSequenceDecorator.class).setActive(log);
        LogFactory.getLog(SequenceObserversSequence.class).setActive(log);
        LogFactory.getLog(SequenceObserversSequenceDecorator.class).setActive(log);
        LogFactory.getLog(PrintSequenceObserver.class).setActive(log);
        LogFactory.getLog(ProbeSequenceObserver.class).setActive(log);

        // Very verbose if active!
        LogFactory.getLog(DateSequence.class).setActive(false);
        LogFactory.getLog(BirthdayRegistry.class).setActive(false);
        
        //
        // SequenceObserver observers and sequences
        //
        
        Calendar calendar = Calendar.getInstance();
        
        // "now"
        Date now = calendar.getTime();
        // 6 days into the future...
        calendar.add(Calendar.DATE, 6);
        Date future = calendar.getTime(); 
        
        // "now" -> infinite...
        DateSequence unbounded = new DateSequence(now);
        // "now" -> "now + 6"...
        DateSequence bounded = new DateSequence(now, future);
        // Random...
        calendar.add(Calendar.YEAR, -1 * MathFacade.getInstance().getRandomNumber(50));
        DateSequence random = new DateSequence(calendar.getTime(), 360*80);
        // Sequence to correlate...
        LongSequence correlated = new LongSequence();
        
        // Create and add observers...
        CorrelatedSequenceObserver observer = 
            new CorrelatedSequenceObserver(correlated);
        unbounded.addObserver(observer);
        bounded.addObserver(observer);

        bounded.addObserver(new PrintSequenceObserver(SequencePolicy.VERBOSE));
        unbounded.addObserver(new PrintSequenceObserver(SequencePolicy.IDENTITY));
        
        // Nested observers: unbounded correlated with bounded -> print unbounded
        bounded.addObserver(new CorrelatedSequenceObserver(unbounded));
        
        // Special feature because of *this* implementation: 
        // Observable sequences are also observers themselves...
        bounded.addObserver(unbounded);
        unbounded.addObserver(bounded);

        //
        // Birthday registry (SequenceObserver)
        //
        
        // Date format used to parse and format dates...
        DateFormat dateFormat = new SimpleDateFormat("dd-MM-yyyy");
        // Birthday sequences...
        DateSequence birthdays1 = new DateSequence(dateFormat.parse("01-01-1875"), dateFormat.parse("31-12-1900"));
        DateSequence birthdays2 = new DateSequence(dateFormat.parse("01-01-1960"), dateFormat.parse("31-12-2010"));

        BirthdayRegistry br1 = new BirthdayRegistry(dateFormat);
        br1.register(birthdays1);               // all state events
        br1.register(birthdays2, State.NORMAL); // only on next
        // Also applicable... 
        br1.register(bounded);

        // Register people...
        br1.add("08-01-1942", "Stephen Hawking");
        br1.add("14-03-1879", "Albert Einstein");
        br1.add("07-08-1885", "Niels Bohr");
        br1.add("19-02-1471", "Nicolas Copernicus");

        // Copy registry...
        BirthdayRegistry br2 = new BirthdayRegistry(br1);
        
        br2.register(birthdays1);
        // Also applicable... 
        br1.register(random);
        br2.register(unbounded, State.RESET, State.RESTART);

        // Register people
        br2.add("23-01-1944", "Rutger Hauer");
        br2.add("13-07-1942", "Harrison Ford");
        br2.add("13-07-1946", "cheech marin");
        br2.add("08-01-1935", "ELVIS PRESLEY");
        br2.add("09-06-1963", "Johnny Depp");
        br2.add("09-06-1981", "Natalie Portman");
        br2.add("09-06-1961", "Michael J. \"Back to the future\" Fox");
        
        //
        // Annotated observers and sequences
        //
        
        // Make the reversible prime sequence observable using
        // observers with notification methods identified by
        // annotations...
        AnnotatedObserversSequence<Integer> prime =
            new AnnotatedObserversSequenceDecorator<Integer>(new ReversiblePrimeSequence(50));
        
        // Implements SequenceObserver, but is also annotated... Note that
        // the observer added below has not other external references and
        // can thus be garbage collected at any time!
        prime.addObserver(new PrintSequenceObserver(SequencePolicy.VERBOSE));
        // Keep a live reference to prevent observer from being garbage
        // collected...
        Object liveReference1 = null;
        prime.addObserver(liveReference1 = new PrintSequenceObserver(SequencePolicy.ID));
        // Must not be added twice...
        if (prime.addObserver(liveReference1)) {
            out.error("Could add same observer instance twice: ", liveReference1);
            ok = false;
        }
        // Implements SequenceObserver, but is not annotated. It will be
        // ignored...
        prime.addObserver(new CorrelatedSequenceObserver(new RangeSequence(1,10)));
        // Ignore...
        prime.addObserver(new Object());
        // To prevent garbage collection...
        Object liveReference2 = null;
        // Annotations only - two notification methods...
        prime.addObserver(liveReference2 = new ProbeSequenceObserver());
        
        // Suppress warnings for unused local variables...
        if (liveReference1 == liveReference2) {/*..*/}
        
        // All sequences...
        Sequence<?>[] sequences = {
            unbounded, 
            bounded, 
            correlated, 
            random, 
            birthdays1, 
            birthdays2,
            prime
        };
        
        out.heading("Created sequences");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println("Sequence ", i, ": ", sequence);
        }
        
        out.heading("Created observable sequences and observers");
        
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            if (sequence instanceof ObservableSequence) {
                ObservableSequence<?,?,?> os = (ObservableSequence<?,?,?>)sequence;
                out.println("Sequence ", i, " is observable: ", os.toString(SequencePolicy.ID));
                for (Object o : os.getObservers()) {
                    out.println("-> ", o);
                }
                out.println();
            } else {
                out.println("Sequence ", i, " is not observable: ", sequence.toString(SequencePolicy.ID));
                out.println();
            }
        }

        //
        // Test illegal operations
        //
        
        out.heading("Observer tests");
        
        // Trouble if performed: undetectable self reference!
        // bounded.addObserver(new CorrelatedSequenceObserver<Date>(bounded));

        // Trouble if allowed, but detectable: self reference! 
        try {
            out.println();
            out.print("Trying to add one self as an observer: ");
            bounded.addObserver(bounded);
            out.error("ALLOWED!");
            ok = false;
            
        } catch (RuntimeException e) {
            out.println("NOT allowed - OK: ", e);
        }
        
        //
        // Testing observers
        //
        
        out.heading("Advancing sequences");
        
        // Now advance sequences...
        for (int i = 0; i < 55; i++) {
            out.println();
            out.println("Iteration ", i, ":");
            
            for (int j = 0; j < sequences.length; j++) {
                Sequence<?> sequence = sequences[j];
                out.println();
                out.println("Advancing sequence ", j, ": ", sequence, "...");
                Object value = sequence.next();
                out.println("Advancing done - value is: ", value);
                
                // Reset?
                if (i % 8 == 7) {
                    out.println();
                    out.println("Reseting sequence: ", sequence, "...");
                    sequence.reset();
                    out.println("Reseting done - value is: ", sequence.current());
                }
                // Remove an observer...
                if (i % 12 == 11) {
                    if (sequence instanceof ObservableSequence) {
                        ObservableSequence<?,?,?> os = (ObservableSequence<?,?,?>)sequence;
                        // Remove first observer, if any...
                        for (Object o : os.getObservers()) {
                            if (os.removeObserver(o)) {
                                out.println();
                                out.println("Removed observer: ", os.toString(SequencePolicy.ID), " -> ", o);
                                break;
                            }
                        }
                    }
                }
            }            
        }

        //
        // Testing birthday registry 
        //
        
        out.heading("Using birthday registry");

        if (!LogFactory.getLog(BirthdayRegistry.class).isActive()) {
            out.warn("Logging disabled for birthday registry!");
        }
        // Disable logging of notifications, or it will flood the
        // test log! 
        LogFactory.getLog(birthdays2).setActive(false);
        
        // Re-register because of the above tests...
        br2.register(birthdays2);
        
        out.println();
        out.println("Registry: ", birthdays2);
        while (birthdays2.state() != State.RESTART) {
            birthdays2.next();
        }
        
        out.heading("Observer tests ", ok ? "succeeded" : "failed", "!");
        
        return ok;
    }
    
}