/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.prototype;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>non-copyable count-down sequence</i> is a singleton sequence used 
  * for count-downs: 
  * <pre>
  *   Five, Four, Three, Two, One, FIRE! 
  * </pre>
  * 
  * Invoking {@link #clone()} will always throw a 
  * {@code CloneNotSupportedException}, and the {@link #copy()}
  * method returns the same instance. <p>
  * 
  * A count-down sequence is {@link Sequence#bounded() bounded}, 
  * {@link Sequence#consistent() consistent}, and 
  * {@link Sequence#unique() unique}.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcretePrototype")
public class CountdownSequence extends ArraySequence<String> {

    /**
      * The various elements in a count-down sequence. 
      */
    private final static String[] elements = {  
        "Five", "Four", "Three", "Two", "One", "FIRE!" 
    };
    
    /**
      * Single instance of the sequence, initialised thread-safe at class 
      * loading time. <p>
      * 
      * Never null.
      */
    private final static CountdownSequence instance = new CountdownSequence();
    
    /**
      * No-arg constructor.
      */
    private CountdownSequence() {
        super(elements);
    }
    
    /**
      * Returns the {@link CountdownSequence} instance to use. 
      *  
      * @return The sequence; never null.
      */
    public final static CountdownSequence getInstance() {
        return instance;
    }
    
    /**
      * 
      */
    public CountdownSequence copy() {
        // Dangerous!
        return this;
    }

    /**
      * Always throws a {@code CloneNotSupportedException}. 
      *
      * @return Never returns.
      * @throws CloneNotSupportedException Always thrown.
      */
    public CountdownSequence clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }
    
}
