/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.proxy;

import java.util.List;

import dk.rode.thesis.decorator.DecoratedPolicy;
import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * Additional policies for formatting proxied {@link Sequence sequences} 
  * into char sequences (<b>not</b> part of the core Proxy implementation). <p>
  *
  * The policies handle both specific proxy classes based on a 
  * {@link SequenceDecorator decorated} sub-class, or anonymous
  * sequence proxies created via a {@link SequenceProxyFactory}. <p>
  *  
  * <b>Implementation notes:</b><br>
  * <b>This is NOT part of the core Proxy functionality, merely 
  * to provide better log descriptions!</b><p>
  * 
  * See implementation notes for the {@link ObjectPolicy} enumeration.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequencePolicy
  * @see    DecoratedPolicy
  */
public enum ProxiedPolicy implements StringablePolicy<Sequence<?>> {

    /**
      * Policy used to format string representations of
      * proxies that indicates the nested structure. <p>
      * 
      * Both specific proxy classed created based on a decorated
      * sequence and proxies based on the {@link java.lang.reflect.Proxy 
      * java.lang.reflect.Proxy} class are respected. <p>
      * 
      * Notice that a given sequence will be listed with 
      * <b>all</b> proxies registered to it even if the
      * sequence is accessed outside the scope of the
      * proxy for the current context. <p>
      * 
      * Proxies using {@link #HIDDEN} as their default
      * policy are respected. 
      */
    VISIBLE {
       public CharSequence toString(Sequence<?> sequence) {
           StringBuilder sb = new StringBuilder();
           
           // For specific proxy classes based on a decorator...
           if (sequence instanceof SequenceDecorator) {
               sb.append(sequence.toString(DecoratedPolicy.VISIBLE));
               
           // For java.lang.reflect.Proxy classes...
           } else if (SequenceProxyFactory.factory.isProxied(sequence)) {
               // Can be more then one proxy per sequence!
               List<?> proxies = SequenceProxyFactory.factory.getProxies(sequence);
               if (proxies.size() > 1) {
                   sb.append('[');
               }
               for (int i = 0; i < proxies.size(); i++) {
                   if (i > 0) {
                       sb.append(';');
                   }
                   sb.append(ObjectPolicy.ID.toString(proxies.get(i)));
               }
               if (proxies.size() > 1) {
                   sb.append(']');
               }
               sb.append('.').append(sequence.toString(SequencePolicy.DEFAULT));
               
           // For ordinary sequence classes...
           } else {
               sb.append(sequence.toString(SequencePolicy.VERBOSE));
           }
           return sb.toString();
       }
    }, 
    
    /**
      * Policy used to format string representations of
      * proxies that hides the nested structure "from
      * the sequence on". 
      */
    HIDDEN {
        public CharSequence toString(Sequence<?> sequence) {
            // Proxy class uses the standard sequence string
            // representation, so we only have to test for decorators...
            if (sequence instanceof SequenceDecorator) {
                return sequence.toString(DecoratedPolicy.HIDDEN);
            }
            return sequence.toString(SequencePolicy.DEFAULT);
        }
    };
    
    /**
      * @return {@link dk.rode.thesis.strategy.StringablePolicy.Type#APPENDABLE}. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
    
}
