/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.proxy;

import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>synchronised sequence</i> ensures thread-safe access to a 
  * proxied {@link Sequence sequence} <b>if and only if</b> the proxied
  * sequence is accessed through the proxy only. The lock used in the
  * proxied sequence instance. <p>
  * 
  * A copy of a synchronised sequence will proxy the same original
  * proxied object and use the same lock. <p>
  * 
  * A synchronised sequence exhibit the same {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link Sequence#unique() unique}
  * properties as the proxied sequence. <p> 
  *
  * <b>Implementation notes</b>:<br>
  * A synchronised sequence corresponds to a <i>smart reference</i> as described
  * by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.209].
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceProxyFactory#getSynchronisedSequence(Sequence)
  * 
  * @param  <E> The type of generator delivered by this sequence. 
  */
@ParticipantUsage(value = "Subject", type = Sequence.class)
@Participant("Proxy")
public class SynchronisedSequence<E> extends SequenceDecorator<E> implements Sequence<E> {

    /**
      * The lock to synchronise on. 
      */
    private final Object lock;
    
    /**
      * Constructor. <p>
      * 
      * The lock used is {@code sequence}.
      * 
      * @param  sequence The sequence to proxy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SynchronisedSequence(@Participant("RealSubject") Sequence<E> sequence) {
        this(sequence, sequence);
    }

    /**
      * Constructor.
      * 
      * @param  sequence The sequence to proxy; cannot be null.
      * @param  lock     The lock object to synchronise on; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public SynchronisedSequence(@Participant("RealSubject") Sequence<E> sequence, Object lock) {
        super(sequence, ProxiedPolicy.HIDDEN);
        if ((this.lock = lock) == null) {
            throw new NullPointerException("Lock cannot be null");
        }
    }
    
    /**
      *  
      */
    public SynchronisedSequence<E> copy() {
        return new SynchronisedSequence<E>(this.getSequence(true), this.lock);
    }
    
    /**
      * 
      */
    public void reset() {
        synchronized (this.lock) {
            LogFactory.getLog(this).println("Synchronised access to reset(): ", this.getSequence(false).toString(SequencePolicy.ID));
            super.reset();
        }
    }

    /**
      * 
      */
    public E current() {
        synchronized (this.lock) {
            LogFactory.getLog(this).println("Synchronised access to current(): ", this.getSequence(false).toString(SequencePolicy.ID));
            return super.current();
        }
    }

    /**
      * 
      */
    public E next() {
        synchronized (this.lock) {
            LogFactory.getLog(this).println("Synchronised access to next(): ", this.getSequence(false).toString(SequencePolicy.ID));
            return super.next();
        }
    }
    
}
