/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Proxy</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.207]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Provide a surrogate placeholder for another object to 
  *    control access to it.
  *    
  * </i></ul>
  *  
  * Here, the <b>Subject</b> participant is represented by the
  * {@link dk.rode.thesis.meta.model.Sequence} interface, and
  * the <b>RealSubject</b> participant by any {@code Sequence} 
  * implementation, for example 
  * {@link dk.rode.thesis.meta.model.PrimeSequence},
  * {@link dk.rode.thesis.meta.model.EnglishAlphabetSequence},
  * {@link dk.rode.thesis.meta.model.LongSequence}, etc. <p>
  * 
  * The <b>Proxy</b> participant is represented by the specific 
  * proxy sequence classes defined in this package that control 
  * the access to the real subjects, i.e. sequences, namely 
  * {@link dk.rode.thesis.proxy.SynchronisedSequence}, and
  * {@link dk.rode.thesis.proxy.NonResettableSequence}, and    
  * {@link dk.rode.thesis.proxy.ImmutableSequence}. They all work via the
  * {@linkplain dk.rode.thesis.decorator Decorator} pattern, decorating
  * the actual {@code Sequence} in order to control access to it. <p>
  * 
  * However, the <b>Proxy</b> participant is also represented by any
  * sequence proxy returned by the {@link dk.rode.thesis.proxy.SequenceProxyFactory}
  * class. It utilises Java's {@link java.lang.reflect.Proxy java.lang.reflect.Proxy} 
  * class that allows for runtime adaptation of interface behaviour using
  * reflection. The factory defines three specific factory methods
  * to create proxies based on any {@code Sequence}, mimicking the
  * {@code SynchronisedSequence}, {@code NonResettableSequence}, and
  * {@code ImmutableSequence} classes, respectively 
  * {@link dk.rode.thesis.proxy.SequenceProxyFactory#getSynchronisedSequence(dk.rode.thesis.meta.model.Sequence)},
  * {@link dk.rode.thesis.proxy.SequenceProxyFactory#getNonResettableSequence(dk.rode.thesis.meta.model.Sequence)},
  * and {@link dk.rode.thesis.proxy.SequenceProxyFactory#getImmutableSequence(dk.rode.thesis.meta.model.Sequence)}. <p>
  * 
  * It also supplies a factory method to create a proxy that will log all 
  * access to the proxied object in form of the  
  * {@link dk.rode.thesis.proxy.SequenceProxyFactory#getLoggableSequence(dk.rode.thesis.meta.model.Sequence)}
  * factory method, as well as the
  * {@link dk.rode.thesis.proxy.SequenceProxyFactory#getVirtualSequence(dk.rode.thesis.meta.reflect.InstantiableTypeLiteral, Object...)}
  * factory method that will create a sequence proxy that will delay creation of 
  * the actual {@code Sequence} instance until the first {@code Sequence} method is
  * invoked. Finally, the {@link dk.rode.thesis.proxy.SequenceProxyFactory#getSharedSequence(dk.rode.thesis.meta.model.Sequence)}
  * method allows a given sequence to be shared among copies of it until a copy
  * invokes either {@code next()} or {@code reset()}, which will detach the
  * copy from the shared reference. <p> 
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/Proxy.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link java.lang.reflect.Proxy java.lang.reflect.Proxy} class is 
  * quite flexible. It allows casting to any (runtime) implemented interface, 
  * so type information need not be lost for the proxied sequences. A proxy that 
  * works as a (dynamic) {@linkplain dk.rode.thesis.decorator Decorator} will 
  * effectively change the type (class), so casts and {@code instanceof} tests 
  * will fail for interfaces implemented by the proxied object. This is the case 
  * with the fixed proxy classes created in this package. For example, a 
  * {@link dk.rode.thesis.meta.model.ReversibleSequence} proxied by the 
  * {@code SynchronisedSequence} class will to the context no longer be
  * reversible, because {@code SynchronisedSequence} do not implement the
  * {@code ReversibleSequence} interface. A {@code java.lang.reflect.Proxy} 
  * instance representing a {@code Sequence} that implements 
  * {@code ReversibleSequence} still allows casts and {@code instanceof}
  * tests. <p>
  * 
  * The price of using {@code java.lang.reflect.Proxy} instances is naturally speed 
  * as the implemented interfaces are accessed using reflection in form of a 
  * {@link java.lang.reflect.InvocationHandler java.lang.reflect.InvocationHandler} 
  * instance. However, reflection in modern JVM's is quite fast, so it is probably 
  * acceptable in most cases. Also, though a {@code java.lang.reflect.Proxy} object 
  * can be cast into any runtime implemented interface, the original type (class) 
  * information of the proxied object is lost. Class cast exceptions can also occur.
  * Once a proxy instance is created based on a specific set of interfaces, the 
  * instance represents a unique proxy class that implements these exact interfaces. 
  * Creating another proxy instance with the same ordered set of interfaces will reuse 
  * the proxy class. The returned proxy is thus an instance of that proxy class and 
  * not the original type, for example a {@link dk.rode.thesis.meta.model.LongSequence}. 
  * There are also other problems with such proxies, for example marshaling and 
  * serialization. <p>
  * 
  * For an in-depth introduction to the {@code java.lang.reflect.Proxy} class,
  * please refer the its JavaDoc {@linkplain java.lang.reflect.Proxy documentation}. <p>
  * 
  * Gamma et al. define four overall types of proxies: <i>remote proxy</i>, 
  * <i>virtual proxy</i>, <i>protection proxy</i>, and <i>smart reference</i>
  * [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.208-209]. The {@code getVirtualSequence} method returns a 
  * <i>virtual proxy</i>. The {@code NonResettableSequence} and 
  * {@code ImmutableSequence} classes represent <i>protection proxies</i> 
  * and so do the proxies returned by the {@code getNonResettableSequence} 
  * and {@code getImmutableSequence} methods. The {@code SynchornisedSequence} 
  * class as well as the proxy returned by the {@code getSynchronisedSequence} 
  * method represent a <i>smart reference</i>. The {@code getSharedSequence}
  * method also returns a <i>smart reference</i>, more precisely a 
  * <i>smart pointer</i>. This package defines no <i>remote proxies</i>, but 
  * an example could be a sequence fetching its values using a socket channel 
  * to a sequence on a different machine or JVM delivering the values. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Proxy", 
         scope        = Scope.Object,
         purpose      = Purpose.Structural,
         participants = {"Proxy", "Subject", "RealSubject"})
package dk.rode.thesis.proxy;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;

