/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.reflect.CallerClass;

/**
  * A <i>singleton alphabet sequence</i> for letters in the Norwegian alphabet:
  * {@code a} though {@code z} as well as {@code }, {@code }, and
  * {@code }, all lower-cased. <p>
  * 
  * Only a single instance of this class is created, and is available
  * via the {@link #getInstance()} method. <p>
  * 
  * This class cannot be sub-classed. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A single instance of {@code NorwegianAlphabetSequence} is always created at
  * class loading time in a thread-safe manner by using standard static
  * initialisation. For example, accessing the static field {@link #SIZE} 
  * <b>will</b> cause the singleton instance to be created. Hence, each call to
  * {@code getInstance()} does not have to be synchronised or check to see if 
  * the single instance has been created. <p>
  * 
  * A {@link CallerClass} is used to ensure that only this class itself can 
  * create the singleton instance. <p>
  * 
  * Any singleton instance in Java is only a singleton per <i>class loader</i>
  * if created based on and utilising intrinsic class features. <p>
  * 
  * Note that the standard {@link ArraySequence} {@code hashCode} and
  * {@code equals(Object)} still works for singleton sub-classes of it. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DanishAlphabetSequence
  * @see    SimpsonsFamilySequence
  * @see    SimpsonsAndBouvierFamilySequence
  * @see    SmileySequence
  */
@Participant("Singleton")
@Singleton
public final class NorwegianAlphabetSequence extends ArraySequence<String> {

    /**
      * The letters in the Norwegian alphabet.
      */
    private final static String[] alphabet = {
        "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", 
        "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
        "", "", ""
    };

    /**
      * The size of the Norwegian alphabet. <p>
      *  
      * <b>Implementation notes</b>:<br>
      * Accessing this static field <b>before</b> {@link #getInstance()} 
      * <b>will</b> cause the singleton instance to be created!
      * 
      * @see DanishAlphabetSequence#SIZE
      * @see SimpsonsFamilySequence#SIZE
      * @see SimpsonsAndBouvierFamilySequence#SIZE
      * @see SmileySequence#SIZE
      */
    public final static int SIZE = alphabet.length; // access would also force load even if it was a constant...
    
    /**
      * Single instance of the sequence, initialised thread-safe at class 
      * loading time. <p>
      * 
      * Never null.
      */
    private final static NorwegianAlphabetSequence instance = new NorwegianAlphabetSequence();

    /**
      * Returns the {@link NorwegianAlphabetSequence} singleton
      * instance to use.
      *
      * @return The sequence; never null.
      * @see    Singleton
      */
    public final static NorwegianAlphabetSequence getInstance() {
        return instance;
    }
    
    /**
      * Private no-arg constructor.
      * 
      * @throws SingletonError If another class tries to invoke
      *         this constructor (via reflection).
      */
    private NorwegianAlphabetSequence() {
        super(alphabet);
        Log out = LogFactory.getLog(this);
        // Guard against reflection...
        CallerClass caller = new CallerClass();
        // Sufficient to test on the caller class...
        if (!caller.isThisClass()) {
            out.error("Illegal access to singleton constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Created singleton: ", this);
    }
    
    /**
      * Returns this instance.
      * 
      * @return This instance; never null.
      */
    public final NorwegianAlphabetSequence copy() {
        // Disallow copy because this class is a singleton...
        return this;
    }
        
}
