/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.Caller;
import dk.rode.thesis.meta.util.Util;

/**
  * A <i>Simpsons and Bouvier family sequence</i> is a singleton 
  * sequence representing the primary family members in the  
  * <i>The Simpsons</i> and <i>Bouvier</i> families. <p>
  * 
  * Only a single instance of this class is created, on demand only, and 
  * is available via the {@link #getFamily()} method. <p>
  * 
  * It is <b>not</b> possible to sub-class this class! <p>
  *  
  * <b>Implementation notes</b>:<br>
  * Any singleton instance in Java is only a singleton per <i>class loader</i>
  * if created based on and utilising intrinsic class features. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DanishAlphabetSequence
  * @see    NorwegianAlphabetSequence
  * @see    SimpsonsFamilySequence
  * @see    SmileySequence
  */
@Participant("Singleton")
@Singleton("getFamily")
public final class SimpsonsAndBouvierFamilySequence extends SimpsonsFamilySequence {

    /**
      * The primary members in <i>The Simpsons</i> <b>and</b> <i>Bouvier</i> 
      * families. 
      */
    private final static String[] members;
    static {
        // This is *not* a nice way to inherit functionality, but it suffices
        // for this example. Notice that accessing the SIZE attribute will
        // not cause the SimpsonsFamilySequence singleton instance to be created...
        int size = SimpsonsFamilySequence.SIZE;
        members = Util.copy(SimpsonsFamilySequence.members, new String[size + 5]);
        members[size++] = "Patricia Bouvier";
        members[size++] = "Selma Bouvier Terwilliger Hutz McClure Stu Simpson";
        members[size++] = "Ling Bouvier";
        members[size++] = "Jacqueline Ingrid Bouvier";
        members[size++] = "Clancy Bouvier";
    }

    /**
      * The number of primary members in <i>The Simpsons</i> 
      * and <i>Bouvier</i> families. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Accessing this static field will <b>before</b> {@link #getFamily()}
      * will <b>not</b> cause the singleton instance to be created!
      * 
      * @see DanishAlphabetSequence#SIZE
      * @see NorwegianAlphabetSequence#SIZE
      * @see SimpsonsFamilySequence#SIZE
      * @see SmileySequence#SIZE
      */
    public final static int SIZE = members.length; 
        
    /**
      * An <i>instance container</i> class that when loaded will create 
      * the {@link SimpsonsAndBouvierFamilySequence} instance to use and return
      * from the {@link SimpsonsAndBouvierFamilySequence#getFamily()} method. <p>
      *
      * The initialisation of the sequence is thread-safe as it
      * utilises standard static initialisation. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This corresponds to the <i>initialise-on-demand holder class</i> idiom
      * described in [<a href="http://www.rode.dk/thesis/Bibliography.htm#Bloch01" target="thesis-bibliography">Bloch01</a>, p.194].
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private final static class Instance {
        
        /**
          * The {@link SimpsonsAndBouvierFamilySequence singleton} instance. <p>
          * 
          * Not created until the {@link SimpsonsAndBouvierFamilySequence#getFamily()}
          * method is invoked.
          */
        private final static SimpsonsAndBouvierFamilySequence instance = new SimpsonsAndBouvierFamilySequence();

        /**
          * Private off-limit constructor. 
          */
        private Instance() {
            throw new AssertionError(); // guard against reflection...
        }
    }
    
    /**
      * Returns the {@link SimpsonsAndBouvierFamilySequence} singleton 
      * instance to use. <p>
      * 
      * The singleton instance is not created before the first time
      * this method is invoked! <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This method hides the static {@link SimpsonsFamilySequence#getFamily()}
      * method.
      *  
      * @return The sequence; never null.
      * @see    Singleton
      */
    public static SimpsonsAndBouvierFamilySequence getFamily() {
        // Forces static initialisation of the *inner* class, if
        // not already, in a thread-safe manner!
        return Instance.instance;
    }
    
    /**
      * Private no-arg constructor.
      * 
      * @throws SingletonError If another class tries to invoke
      *         this constructor (via reflection).
      */
    private SimpsonsAndBouvierFamilySequence() {
        super(members);
        Log out = LogFactory.getLog(this);
        // Guard against reflection...
        Caller caller = new Caller().getNonClassCaller(); // never null
        // Can only be invoked during static initialisation of the 
        // nested Instance class...
        if ((caller.callerClass != Instance.class) || (!caller.isStatic())) {
            out.error("Illegal access to singleton constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Created singleton: ", this);
    }
    
    /**
      * Returns this instance.
      * 
      * @return This instance; never null.
      */
    public SimpsonsAndBouvierFamilySequence copy() {
        // Disallow copy because more than one Simpsons and Bouvier 
        // family is more than the world can manage... :)
        return this;
    }
    
}
