/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import static dk.rode.thesis.meta.util.Util.newIdentityHashSet;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>statefull singleton registry</i> stores a reference to each
  * acquired singleton instance using a <i>live daemon thread</i>: as 
  * long as the registry is not garbage collected, the singleton instances
  * acquired through it will not be garbage collected either. <p>  
  *
  * By using composition, this registry defers acquisition of the singleton types
  * to another registry. The registry to use is specified at 
  * {@link #StatefullSingletonRegistry(SingletonRegistry) construction} time. <p>
  * 
  * References to created singletons can also be 
  * {@link #discard(Class) discarded} explicitly. <p>
  * 
  * <b>Implementation notes</b>:</br>
  * This registry do <b>not</b> utilise an internal cache to store the
  * acquired singletons because that would require synchronisation. 
  * Furthermore, we do not know how the actual singleton classes are 
  * implemented, i.e. static initialisation, lazily initialised, synchronised, 
  * or not synchronised!? Hence, each time the {@link #getInstance(Class)} 
  * method is invoked the call is forwarded to the registry actually
  * used to acquire the singleton types. <p>
  *
  * By deferring the responsibility to another registry, the other registry
  * may also be able to acquire broader singleton types, i.e. any singleton
  * type {@code <? super T>}. It also ensures that the internal state
  * kept by this registry has nothing to do with the actual creation of
  * singleton types, only with references to the unique instances of the 
  * singleton types once created. <p>
  * 
  * The anonymous {@link java.lang.Runnable java.lang.Runnable} target 
  * object used by the live daemon thread is an application of the 
  * {@linkplain dk.rode.thesis.adapter Adapter} pattern. 
  *   
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The super type of all singleton types handled by
  *             this registry. Use {@code Object} if any type
  *             of singleton type can be handled.
  */
public class StatefullSingletonRegistry<T> 
  implements SingletonRegistry<T>, Iterable<T> {

    /**
      * The actual registry used to acquire the singleton 
      * types supplied at construction time. <p>
      * 
      * Never null.
      */
    private final SingletonRegistry<? super T> registry;
    /**
      * Live daemon thread to ensure that the references to 
      * acquired singleton instances never goes out of scope. <p>
      * 
      * Never null.
      */
    private final Thread thread;
    /**
      * References to all acquired singleton types based
      * on their <b>identity</b> so they will not be garbage
      * collected. <p>
      * 
      * Never null, but can be empty.
      */
    private final Set<T> singletons;
    
    /**
      * Constructor. <p>
      * 
      * An instance of the {@link StatelessSingletonRegistry} type
      * will be used to acquire the singleton instances.
      */
    public StatefullSingletonRegistry() {
        this(new StatelessSingletonRegistry<T>());
    }
        
    /**
      * Constructor. <p>
      * 
      * @param  registry The registry that will be used to acquire the
      *                  singleton instances; cannot be null.              
      * @throws NullPointerException If {@code registry} is null.              
      */
    public StatefullSingletonRegistry(SingletonRegistry<? super T> registry) {
        if ((this.registry = registry) == null) {
            throw new NullPointerException("Registry cannot be null");
        }
        this.singletons = newIdentityHashSet();
        this.thread = new Thread(new Runnable() { // Adapter!
            public void run() {
                while (true) {
                    synchronized (StatefullSingletonRegistry.this.singletons) {
                        try {
                            // Wait forever and keep a reference to this instance...
                            StatefullSingletonRegistry.this.singletons.wait();
                        } catch (InterruptedException ie) {
                            LogFactory.getLog(this).warn("Statefull singleton registry thread interrupted").warn(ie);
                            // Do nothing, i.e. continue...
                        }
                    }
                }
            }
        });
        this.thread.setDaemon(true); // Do not block normal exit
        this.thread.setName(ObjectPolicy.ID.asString(this));
        this.thread.start();
    }
        
    /**
      * @throws NullPointerException {@inheritDoc} 
      * @throws SingletonException   {@inheritDoc}
      */
    public <S extends T> S getInstance(Class<S> type) {
        // Delegate to the (stateless) registry...
        S instance = this.registry.getInstance(type);
        // ...but keep a reference to the singleton instance... 
        this.singletons.add(instance);
        LogFactory.getLog(this).println("Acquired singleton instance: ", type.getName(), " -> ", instance);
        return instance;
    }

    /**
      * Discards the reference to the singleton instance of the 
      * type supplied as {@code type}, if such a reference is
      * kept by this registry. <p>
      * 
      * If this registry has no such reference, false is returned. 
      * 
      * @param  <S>  The actual singleton type.
      * @param  type The class token identifying the type; cannot be null.
      * @return True if {@code type} represents a singleton type this registry
      *         had a reference to, false if not. 
      * @throws NullPointerException If {@code type} is null.
      * @see    #isSingleton(Class)
      */
    public <S extends T> boolean discard(Class<S> type) {
        return this.singletons.remove(type);
    }
    
    /**
      * @throws NullPointerException {@inheritDoc} 
      * @see    #discard(Class)
      */
    public boolean isSingleton(Class<?> type) {
        return this.registry.isSingleton(type);
    }

    /**
      * Returns an iterator to iterate over all singleton
      * instances referenced by this registry. <p>
      * 
      * Modifying the returned iterator will not affect this
      * registry.
      * 
      * @return An iterator; never null, but can be empty.
      */
    public Iterator<T> iterator() {
        // A list is used, since it will allow duplicates based on equality...
        return new ArrayList<T>(this.singletons).iterator();
    }
    
    /**
      * Returns the string representation of this registry.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; registry: ").append(this.registry)
                 .append("; references: ").append(this.singletons)
                 .append(']').toString();
    }
    
}
