/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.state;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>State</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>State</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("State tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        boolean ok = true;

        // True to class logging...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(PrimeSequence.class).setActive(log);
        LogFactory.getLog(ReversiblePrimeSequence.class).setActive(log);
        
        // Verbose...
        log = arguments.getArgumentValue("log.verbose", false, Boolean.class);
        LogFactory.getLog(StepSequence.class).setActive(log);
        
        out.heading("Created sequences");

        // Max prime...
        int max = 200;
        // n = iterations, m = reset, l = reverse
        int i = 0, n = 555, m = n / 5, l = n / 7;
        
        PrimeSequence a = new PrimeSequence(max);
        ReversiblePrimeSequence b = new ReversiblePrimeSequence(max);
        // True to enable logging from reversible prime sequence sequence as well...
        LogFactory.getLog(b).setActive(log);
        // Copy which will use its own sieve...
        ReversiblePrimeSequence c = b.copy();
        
        // State sequence...
        MathFacade mf = MathFacade.getInstance();
        int step  = mf.getRandomNumber(1, 19);
            step *= mf.getRandomBoolean() ? 1 : -1;
        StepSequence ss = StepSequence.Factory.create(step, mf.getRandomNumber(-100, 100));
                
        out.println("PrimeSequence (a):           ", a.toString(SequencePolicy.VERBOSE));
        out.println("ReversiblePrimeSequence (b): ", b.toString(SequencePolicy.VERBOSE));
        out.println("ReversiblePrimeSequence (c): ", c.toString(SequencePolicy.VERBOSE)).println();
        out.println("StepSequence (step: ", step, "): ", ss.toString(SequencePolicy.VERBOSE));
        out.println();
        out.println("Maximum prime considered: ", max);
        out.println("Iterations:               ", n);
        out.println("Reset every:              ", m, " iterations");
        out.println("Reversed:                 0-", n, " % ", l , " < 7");
        out.println();
        
        out.heading("Comparing PrimeSequence (a) and ReversiblePrimeSequence (b)");
        
        do {
            
            State stateA = a.state(), stateB = b.state();
            Integer valueA = a.current(), valueB = b.current();
            
            out.println(i, " ", a.toString(SequencePolicy.VERBOSE));
            out.println(i, " ", b.toString(SequencePolicy.VERBOSE));

            if (stateA != stateB) {
                out.error("Different states at iteration ", i, ": ", stateA, " != ", stateB);
                ok = false;
            }
            if (!valueA.equals(valueB)) {
                out.error("Different values at iteration ", i, ": ", valueA, " != ", valueB);
                ok = false;
            }
            
            // Advance...
            a.next();
            b.next();
            
            // Reset?...
            if (i % m == 3) {
                a.reset();
                b.reset();
            }    
            out.println();
            
        } while (++i < n);

        out.heading("Testing illegal state of ReversiblePrimeSequence (c)");
        
        // "c" was copied before the primes were calculated, so trying to
        // reverse it now will cause an exception...
        try {
            c.reverse();
            out.error("Could reverse ", c);
            ok = false;
            
        } catch (Exception e) {
            out.println("OK - could not reverse (as expected) ", c, ": ", e);
        }
        
        out.heading("Listing states of ReversiblePrimeSequence (c)");

        out.println("Sequence: ", c.toString(SequencePolicy.VERBOSE));
        out.println();
        
        i = 0;
        do {
            
            // Fetch functional state...
            FunctionalState<Integer> fs = c.getFunctionalState();
            // Fetch internal state...
            State s = c.state();
            
            // Print...
            out.println(i, " ", c.toString(SequencePolicy.IDENTITY), " -> [", fs, ", ", s, "]");

            // Reset?...
            if (i % m == 4) {
                c.reset();
            }    
            // Reverse?...
            if (i % l < 7) {
                // Cannot reverse if all primes have not been calculated
                if (c.reversible()) {
                    try {
                        c.reverse();
                    } catch (Exception e) {
                        out.error("Could not reverse ", c, ": ", e);
                        ok = false;
                    }
                } else {
                    out.warn(c, " is not reversible at this state: ", c.getFunctionalState());
                }
            }
            
            // Advance...
            c.next();
            
            out.println();
            
        } while (++i < n);

        out.heading("Testing StepSequence");

        out.println("StepSequence (step: ", step, "): ", ss.toString(SequencePolicy.IDENTITY)).println();
        
        for (int j = 0; j < 100; j++) {
            out.println(j, ": ", ss, " -> ", ss.current());
            ss.next();
            if (j % 4 == 3) {
                ss.reset();
            }
        }
        
        out.heading("State tests ", ok ? "succeeded" : "failed", "!");
        
        return ok;
    }
    
}
