/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.strategy;

import java.lang.reflect.Method;

import dk.rode.thesis.meta.Participant;

/**
  * Returns a textual representation in form of a char sequence 
  * of a given {@link Stringable stringable} object.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Stringable
  * 
  * @param  <T> The type of the applicable stringable objects. 
  */
@Participant("Strategy")
public interface StringablePolicy<T> {

    /**
      * A <i>policy type</i> is a <i>hint</i> a {@link Stringable stringable} 
      * context can use to decide how to use the textual representation a
      * given {@link StringablePolicy policy} generates in conjunction with 
      * additional textual information required to generate an overall 
      * representation. <p>
      *  
      * A {@link Stringable} object that has been formatted with a 
      * policy having the {@link #CLOSED} type hints to the stringable context 
      * that additional text should not be appended to its representation, 
      * where as the {@link #OPEN} type on the contrary expects the 
      * context to do so. <p>
      * 
      * The {@link #APPENDABLE} type represents a complete stand-alone
      * textual representation (as do {@code CLOSED}), but where 
      * stringable contexts may append additional text if they desire, 
      * or simply use as is. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      * 
      * @see    StringablePolicy#type()
      */
    public static enum Type {
        /**
          * The context is expected not to append text to the
          * generated representation.
          */
        CLOSED,
        /**
          * The context is expected to append text to the generated
          * representation. 
          */
        OPEN,
        /**
          * The context is free to append text to the generated 
          * representation if it so desires.
          */
        APPENDABLE;
    }
    
    /**
      * An idiom that allows the retrieval of a given {@link StringablePolicy}
      * defined as an <b>enumeration constant</b> based on an exact name of
      * another, possibly unrelated, policy. <p>
      * 
      * The name idiom can be used in case a given stringable type contains
      * other stringable types that should be formatted using <i>equivalent</i>
      * policies (by name) in order to supply an overall textual representation. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class NameIdiom {
      
        /**
          * Returns the policy of type {@code S} having the <b>exact</b> name
          * as the {@code toString()} value of the policy supplied as 
          * {@code policy}, or null if no such constant can be found. <p>
          * 
          * For enumeration constants, the {@code toString()} method
          * always returns its name. 
          *  
          * @param  <S>    The type of the policy to return.
          * @param  <T>    The type of the policy that supplies the name to find.
          * @param  type   The enumeration type where the desired policy is to
          *                be looked up; can be null, in which case null is returned.
          * @param  policy The policy to supply the name to look up based on its
          *                {@code toString} representation; can be null, in which
          *                case null is returned.
          * @return A policy defined as an enumeration constant of type {@code S} 
          *         having the same name as {@code policy.toString()}, or null in 
          *         case no such constant can be found.
          * @see    #getSimilarByName(Class, StringablePolicy, Enum)
          */
        public final static <S extends Enum<S> & StringablePolicy<?>,T extends StringablePolicy<?>> S getSimilarByName(Class<S> type, T policy) {
            return getSimilarByName(type, policy, null);
        }
        
        /**
          * Returns the policy of type {@code S} having the <b>exact</b> name
          * as the {@code toString()} value of the policy supplied as 
          * {@code policy}, or {@code defaultPolicy} if no such constant
          * can be found. <p>
          * 
          * For enumeration constants, the {@code toString()} method
          * always returns its name. 
          * 
          * @param  <S>           The type of the policy to return.
          * @param  <T>           The type of the policy that supplies the name to find.
          * @param  type          The enumeration type where the desired policy is to
          *                       be looked up; can be null, in which case {@code defaultPolicy}
          *                       is returned.
          * @param  policy        The policy to supply the name to look up based on its
          *                       {@code toString} representation; can be null, in which
          *                       case {@code defaultPolicy} is returned.
          * @param  defaultPolicy The default policy; can be null.               
          * @return A policy defined as an enumeration constant of type {@code S} 
          *         having the same name as {@code policy.toString()}, or 
          *         {@code defaultPolicy} in case no such constant can be found.
          * @see    #getSimilarByName(Class, StringablePolicy)
          */
        public final static <S extends Enum<S> & StringablePolicy<?>,T extends StringablePolicy<?>> S getSimilarByName(Class<S> type, T policy, S defaultPolicy) {
            try {
                // See java.lang.Enum doc...
                Method method = type.getDeclaredMethod("valueOf", String.class);
                // If "policy" is an enumeration context, the "toString()" method
                // returns its name...
                return type.cast(method.invoke(null/*static*/, policy.toString()));
              
            // No such constant for one reason or another, so return default (which may be null)...
            } catch (Exception e) {
                return defaultPolicy;
            }
        }
        
        /**
          * Private off-limit constructor.
          */
        private NameIdiom() {
            throw new AssertionError(); // guard against reflection
        }
    }
    
    /**
      * Returns a textual representation of the object supplied 
      * as {@code object} in form of a char sequence.
      *  
      * @param  object The object; cannot be null.
      * @return The char sequence representation; never null.
      * @throws NullPointerException If {@code object} is null.
      * @see    Stringable#toString(StringablePolicy)
      */
    public CharSequence toString(T object);
    
    /**
      * Returns the {@link Type type} of this policy. <p>
      * 
      * The {@link Stringable stringable} context can use the
      * type to decide how to use the textual representation of 
      * this policy in conjunction with additional text. 
      * 
      * @return The type; never null.
      */
    public Type type();
        
}
