/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.templatemethod;

import static dk.rode.thesis.meta.util.Util.readBytes;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;

/**
  * A <i>file sequence</i> delivers byte arrays as sequence
  * values, where each array corresponds to the contents of 
  * the current file or directory in a given directory. <p>
  *
  * <b>Implementation notes</b>:<br>
  * Returning a byte array instead of an input stream is done 
  * to illustrate the use of {@link #nextValueClose(File, byte[])}. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteClass")
public class FileSequence extends SequenceTemplate<File,byte[]> {
    
    /**
      * The directory where the files are read from. <p>
      * 
      * Never null.
      */
    private File directory;
    /**
      * A filter to filter the files in {@link #directory}. <p>
      * 
      * Can be null, in which case all files are fetched.
      */
    private FileFilter filter;
    /**
      * The current file input stream. <p>
      */
    private FileInputStream stream;
    /**
      * The matching files. <p>
      * 
      * Never null, and never empty as this sequence must always
      * have a current value. 
      */
    private File[] files;

    /**
      * Constructor. <p>
      * 
      * @param  directory The directory to deliver the files and/or 
      *                   directories; cannot be null.
      * @param  filter    A filter to filter the files in {@code directory};
      *                   can be null, in which case all files and
      *                   directories are returned.                   
      * @throws NullPointerException If {@code directory} is null.
      * @throws SequenceValueException If {@code directory} does not exist, or if it 
      *         represents a file, or if the initialisation fails otherwise. 
      */
    public FileSequence(File directory, FileFilter filter) throws SequenceValueException {
        // File and filter are passed to the initialisation method... 
        super(directory, filter);
    }

    /**
      * Copy constructor. <p>
      * 
      * If {@code sequence} is {@link #isClosed() closed}, this 
      * sequence will be closed as well.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws SequenceValueException If the initialisation fails, or if the 
      *         current value of this sequence cannot be fetched.
      */
    public FileSequence(FileSequence sequence) throws SequenceValueException {
        // File and filter are passed to the initialisation method... 
        super(sequence, sequence.directory.getAbsoluteFile(), sequence.filter);
    }

    /**
      * 
      */
    public FileSequence copy() {
        return new FileSequence(this);
    }
    
    /**
      * Opens a new file input stream based on the file
      * supplied as {@code file}. <p>
      * 
      * @param  file The file to open the stream for; never null.
      * @return A <b>new</b> stream; never null.
      * @throws IOException If the stream cannot be opened.
      */
    private final FileInputStream openStream(File file) throws IOException {
        this.stream = new FileInputStream(file);
        LogFactory.getLog(this).println("Opened stream: ", this.stream);
        return this.stream;
    }

    /**
      * @throws IOException If the closing fails. 
      */
    private final void closeStream() throws IOException {
        if (this.stream != null) {
            this.stream.close();
            LogFactory.getLog(this).println("Closed stream: ", this.stream);
            this.stream = null;
        }
    }
    
    /**
      * @throws NullPointerException If no directory is supplied at 
      *         construction time.
      * @throws IllegalArgumentException If the buffer size supplied at 
      *         construction time is illegal, or the file supplied does
      *         not exist, or if not a directory, or if the directly
      *         is empty.
      * @throws Exception If the super sequence open method throws an
      *         exception.
      */
    @Override
    protected boolean sequenceOpen(Object... arguments) throws Exception {
        // Varargs will never be null here, because the applicable 
        // constructor declares File and FileFilter parameters. The casts 
        // below are known to succeed for the same reason. We could have 
        // used an additional generic type parameter on SequenceTemplate 
        // to narrow the argument type for the constructors and this method 
        // even further, but if more than one argument is needed with 
        // different type hierarchies, the actual type parameter would still 
        // need to be Object. Here, for example, we use both File and FileFilter 
        // with only Object in common...
        File directory = (File)arguments[0];
        // The actual argument can still be null in case a single null 
        // argument was supplied to the constructor...
        if ((this.directory = directory) == null) {
            throw new NullPointerException("Directory cannot be null");
            
        } else if (!this.directory.exists()) {
            throw new IllegalArgumentException("Directory does not exist");
            
        } else if (!this.directory.isDirectory()) {
            throw new IllegalArgumentException("File does not denote a directory");
        }
        
        // Filter, if any...
        this.filter = (FileFilter)arguments[1];

        // Fetch the actual files...
        this.files = this.directory.listFiles(this.filter);
        // Files cannot be empty because this sequence must always have a 
        // current value...
        if (this.files.length == 0) {
            throw new IllegalArgumentException("Directory contains no files");
        }
        
        // Ok...
        return true;
    }
    
    /**
      * @throws Exception If the super sequence close method throws an
      *         exception.
      */
    @Override
    protected void sequenceClose() throws Exception {
        // Simply for logging purposes, not required!
        super.sequenceClose(); // (abstract) operation
        // Clear files...
        this.files = null;
    }
    
    /**
      * @return The next {@code File} in the file stream; can be null in
      *         case the stream has no more files (restart). 
      */
    @Override
    protected File nextValueOpen(byte[] current) throws Exception {
        // Next file...
        int index = this.index();
        if (index < this.files.length) {
            return this.files[index];
        }
        // No more entries...
        return null;
    }
    
    /**
      * Creates a new input stream based on {@code file} and reads 
      * and returns the data. 
      */
    @Override
    protected byte[] nextValue(File file, byte[] current) throws Exception {
        // If entry is null, we are about to restart...
        if (file == null) {
            return null;
        }
        // Open new file stream and read the bytes. This will not
        // close the stream...
        return readBytes(this.openStream(file));
    }

    /**
      * Closes the current stream.  
      */
    @Override
    protected void nextValueClose(File key, byte[] value) throws Exception {
        // Close the current value...
        this.closeStream();
    }
    
}
