/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.templatemethod;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>negative sequence</i> represents a perhaps 
  * {@link #bounded() bounded} integer value that is decreased 
  * with each call to {@link #next() next()}, starting at minus one. <p>
  * 
  * A negative sequence is {@link #bounded() unbounded} unless a specific 
  * minimum number is supplied at construction time, but is {@link #unique() 
  * unique} and {@link Sequence#consistent() consistent}. <p>
  *
  * <b>Implementation notes</b>:<br>
  * A negative sequence quite possibly represents the simplest implementation 
  * of a sequence {@link SequenceTemplate template}, and it could easily 
  * have been implemented in a fashion like the {@link LongSequence} class. 
  * <b>Note that no explicit class attribute is required to store the sequence 
  * value</b>! The use of the sequence template here is only to illustrate its use. <p>
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteClass")
public class NegativeSequence extends SequenceTemplate<Integer,Integer> {

    /**
      * The minimum value of this sequence. <p>
      * 
      * A value of zero means no minimum, i.e. unbounded.
      */
    private int minimum;
    
    /**
      * Constructor, which creates this negative sequence as unbounded. <p>
      */
    public NegativeSequence() {
        this(0);
    }

    /**
      * Constructor, which creates this negative sequence as bounded. <p>
      * 
      * @param  minimum The minimum value of this sequence; cannot be positive.
      * @throws IllegalArgumentException If {@code minimum} is positive. 
      */
    public NegativeSequence(int minimum) {
        super(minimum);
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NegativeSequence(NegativeSequence sequence) {
        super(sequence, sequence.minimum);
    }

    /**
      * 
      */
    public NegativeSequence copy() {
        return new NegativeSequence(this);
    }

    /**
      * Returns true if a minimum sequence value were supplied
      * at construction time, false if not.
      *
      * @return True if a minimum value was supplied at construction time,
      *         false if not.
      */
    public boolean bounded() {
        return (this.minimum != 0);
    }
    
    /**
      * @throws IllegalArgumentException If the minimum sequence value is
      *         positive.
      * @throws Exception If the super sequence open method throws an
      *         exception.
      */
    @Override
    protected boolean sequenceOpen(Object... arguments) throws Exception {
        // Varargs will never be null here, because the applicable 
        // constructors declares an int parameter. The cast below is known 
        // to succeed for the same reason and can never be null. We could 
        // have used an additional generic type parameter on SequenceTemplate 
        // to narrow the argument type for the constructors and this method 
        // even further, but if more than one argument is needed with 
        // different type hierarchies, the actual type parameter would still 
        // need to be Object. Here, we could have used File to avoid the 
        // explicit cast...
        int minimum = (Integer)arguments[0];
        if (minimum > 0) {
            throw new IllegalArgumentException("Illegal minimum value: " + minimum);
        }
        this.minimum = minimum;

        return true;
    }
    
    /**
      * 
      */
    @Override
    protected Integer nextValueOpen(Integer current) {
        // key and value are identical!
        if (current == null) {
            current = 1; // start;
        }
        return --current;
    }

    /**
      * 
      */
    @Override
    protected Integer nextValue(Integer key, Integer current) {
        // key and value are identical!
        current = key;
        if ((this.minimum != 0) && (this.minimum >= current)) {
            current = null; // end
        }
        return current;
    }

}
