/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.util.Date;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>counting visitor</i> counts the different types of
  * {@link ValueVisitableSequence value} visitable sequences 
  * encountered during traversal, and is thus independent 
  * of any given {@link SequenceValueScanner scanner} implementation. <p>
  *   
  * A counting visitor does accumulate information about the visited 
  * sequences, but does not require any arguments at execution time.
  * The result of the visitation can be fetched using the appropriate
  * getter methods, for example {@link #getIntegerValued()} for
  * the number of sequences visited that delivered {@code Integer}
  * values. Each time a given sequence instance is visited, the count
  * will be increased, even if the sequence has already been visited. 
  * Before reuse, a counting visitor must be {@link #reset() reset} 
  * for the count to be accurate.   
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    LoggingVisitor
  * @see    TypeVisitor
  * @see    SimpleScanner
  */
@Participant("ConcreteVisitor")
public class CountingVisitor implements SequenceValueVisitor<Void> { // Void = no arguments required
    
    /**
      * Number of visitable sequences that delivered 
      * {@code Date} values. 
      */
    private int dateValued;
    /**
      * Number of visitable sequences that delivered 
      * {@code Integer} values. 
      */
    private int integerValued;
    /**
      * Number of visitable sequences that delivered 
      * {@code Long} values. 
      */
    private int longValued;
    /**
      * Number of visitable sequences that delivered 
      * any type of {@code CharSequence} values. 
      */
    private int stringValued;
    /**
      * Number of visitable sequences that were 
      * {@link CompositeSequence composite} sequences
      * with a specific value type. 
      */
    private int composites;
    /**
      * Number of visitable sequences that delivered 
      * unknown values (excluding composites). 
      */
    private int unknowns;
    
    /**
      * No-arg constructor.
      */
    public CountingVisitor() {
        super();
        this.reset();
    }
    
    /**
      * Resets this visitor before re-scanning. 
      */
    public void reset() {
        this.dateValued = 0;
        this.integerValued = 0;
        this.longValued = 0;
        this.stringValued = 0;
        this.composites = 0;
        this.unknowns = 0;
    }
    
    /**
      * 
      */
    public void visitDateValued(Sequence<Date> sequence, Void unused) {
        this.dateValued++;
    }
    
    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences that delivered {@code Date} values. 
      * 
      * @return The number of sequences.
      */
    public int getDateValued() {
        return this.dateValued;
    }
    
    /**
      * 
      */
    public void visitIntegerValued(Sequence<Integer> sequence, Void unused) {
        this.integerValued++;
    }

    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences that delivered {@code Integer} values. 
      * 
      * @return The number of sequences.
      */
    public int getIntegerValued() {
        return this.integerValued;
    }
    
    /**
      * 
      */
    public void visitLongValued(Sequence<Long> sequence, Void unused) {
        this.longValued++;
    }

    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences that delivered {@code Long} values. 
      * 
      * @return The number of sequences.
      */
    public int getLongValued() {
        return this.longValued;
    }
    
    /**
      * 
      */
    public void visitStringValued(Sequence<? extends CharSequence> sequence, Void unused) {
        this.stringValued++;
    }

    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences that delivered any type of {@code CharSequence} values. 
      * 
      * @return The number of sequences.
      */
    public int getStringValued() {
        return this.stringValued;
    }
    
    /**
      * 
      */
    public void visitUnknown(Sequence<?> sequence, Void unused) {
        if (sequence instanceof CompositeSequence) {
            this.composites++;
        } else {
            this.unknowns++;
        }
    }
    
    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences that were composites and where the type of value was 
      * unknown. 
      * 
      * @return The number of sequences.
      */
    public int getComposites() {
        return this.composites;
    }

    /**
      * Returns the number of {@link ValueVisitableSequence value} visitable
      * sequences where the type of value was unknown, excluding composites. 
      * 
      * @return The number of sequences.
      */
    public int getUnknowns() {
        return this.unknowns;
    }
    
    /**
      * Returns the total number of {@link ValueVisitableSequence value} 
      * visitable sequences that were visited.
      *  
      * @return The number of sequences.
      */
    public int getTotal() {
        return this.dateValued    + 
               this.integerValued + 
               this.longValued    +
               this.stringValued  +
               this.composites    +
               this.unknowns;
    }
    
    /**
      * Returns the string representation of this visitor.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.NAME.asString(this);
    }
    
}
