/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.util.Date;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.state.StateableSequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>logging visitor</i> logs each visited {@link Sequence sequence}
  * to a {@link Log log} supplied at <i>execution time</i> using a given 
  * sequence {@link SequencePolicy policy} supplied at <i>construction 
  * time</i>. <p>
  * 
  * It handles both {@link ValueVisitableSequence value} and 
  * {@link TypeVisitableSequence type} sequences. <p>
  * 
  * A logging visitor does not accumulate information about the visited 
  * sequences.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CountingVisitor
  * @see    TypeVisitor
  * @see    SimpleScanner
  */
@Participant("ConcreteVisitor")
public class LoggingVisitor 
  implements SequenceValueVisitor<Log>, SequenceTypeVisitor<Log> {

    /**
      * The policy used to format visited sequences. <p>
      * 
      * Can be null, i.e. default.
      */
    private SequencePolicy policy;
    
    /**
      * No-arg constructor. <p>
      * 
      * The default {@link SequencePolicy policy} is used to format 
      * visited sequences.
      */
    public LoggingVisitor() {
        this(null); // default
    }
    
    /**
      * Constructor. <p>
      * 
      * @param  policy The {@link SequencePolicy policy} used to format 
      *                visited sequences; can be null, i.e default.
      */
    public LoggingVisitor(SequencePolicy policy) {
        this.setSequencePolicy(policy);
    }
  
    /**
      * Return the currently used {@link SequencePolicy}
      * by this log visitor to log each visited sequence.
      *  
      * @return The used policy; null means default.
      * @see    #setSequencePolicy(SequencePolicy)
      */
    public SequencePolicy getSequencePolicy() {
        return this.policy;
    }
    
    /**
      * Sets the {@link SequencePolicy} to use by this log visitor 
      * to log each visited sequence to {@code policy}.
      *  
      * @param policy The policy to use; can be null, which means 
      *               default policy.
      * @see   #getSequencePolicy()
      */
    public void setSequencePolicy(SequencePolicy policy) {
        this.policy = policy;
    }
    
    //
    // Value visitation methods
    //
    
    /**
      * 
      */
    public void visitDateValued(Sequence<Date> sequence, Log out) {
        out.println("Visiting sequence delivering Date values: ", sequence.toString(this.policy));
    }

    /**
      * 
      */
    public void visitIntegerValued(Sequence<Integer> sequence, Log out) {
        out.println("Visiting sequence delivering Integer values: ", sequence.toString(this.policy));
    }

    /**
      * 
      */
    public void visitLongValued(Sequence<Long> sequence, Log out) {
        out.println("Visiting sequence delivering Long values: ", sequence.toString(this.policy));
    }

    /**
      * 
      */
    public void visitStringValued(Sequence<? extends CharSequence> sequence, Log out) {
        out.println("Visiting sequence delivering CharSequence values: ", sequence.toString(this.policy));
    }

    //
    // Type visitation methods
    //
    
    /**
      * 
      */
    public void visitReversible(ReversibleSequence<?> sequence, Log out) {
        out.println("Visiting reversible sequence: ", sequence.toString(this.policy));
    }
    
    /**
      * 
      */
    public void visitMemorizable(MemorizableSequence<?> sequence, Log out) {
        out.println("Visiting memorizable sequence: ", sequence.toString(this.policy));
    }
    
    /**
      * 
      */
    public void visitObservable(ObservableSequence<?,?,?> sequence, Log out) {
        out.println("Visiting observable sequence: ", sequence.toString(this.policy));
    }
    
    /**
      * 
      */
    public void visitStateable(StateableSequence<?> sequence, Log out) {
        out.println("Visiting stateable sequence: ", sequence.toString(this.policy));
    }
    
    /**
      * 
      */
    public void visitComposite(CompositeSequence<?> sequence, Log out) {
        out.println("Visiting composite sequence: ", sequence.toString(this.policy));
    }

    //
    // Common visitation method
    //
    
    /**
      * 
      */
    public void visitUnknown(Sequence<?> sequence, Log out) {
        out.println("Visiting unhandleable sequence: ", sequence.toString(this.policy));
    }
    
    /**
      * Returns the string representation of this visitor.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        sb.append("policy: ");
        if (this.policy == null) {
            sb.append("default");
        } else {
            sb.append(this.policy);
        }
        return sb.append(']').toString();
    }
    
}
