/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import static dk.rode.thesis.meta.util.Const.NULL;

import dk.rode.thesis.composite.AbstractCompositeSequence;
import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.composite.CompositeStrategy;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Const;

/**
  * An <i>visitable composite sequence</i> is {@link ValueVisitableSequence value} 
  * <b>and</b> {@link TypeVisitableSequence type} visitable, storing
  * {@link Sequence sequences} regardless of the type of value they
  * deliver. <p>

  * <b>The children of a composite sequence are NOT visited explicitly
  * during visits to it!</b> Their visitation must be handled by 
  * {@link SequenceValueScanner value} and {@link SequenceTypeScanner type}
  * scanners! <p>
  *
  * The value returned by the {@link #current()} and {@link #next()} 
  * methods is the value of the <b>first</b> contained sequence. In 
  * case a visitable composite sequence has no children, 
  * {@link Const#NULL NULL} is returned. <p>
  * 
  * This implementation is not thread-safe.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteElement")
public class VisitableCompositeSequence 
     extends AbstractCompositeSequence<Object> 
  implements ValueVisitableSequence<Object>, TypeVisitableSequence<Object> {

    /**
      * No-arg constructor. <p>
      * 
      * The {@link CompositeStrategy#DEFAULT} strategy is used
      * for traversal.
      */
    public VisitableCompositeSequence() {
        super();
    }
    
    /**
      * Constructor.
      * 
      * @param strategy The traversal strategy; can be null, in which
      *                 case {@link CompositeStrategy#DEFAULT} is used. 
      */
    public VisitableCompositeSequence(CompositeStrategy strategy) {
        super(strategy);
    }

    /**
      * Copy constructor, which also copies associated sequences.
      * 
      * @param  sequence The composite sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null. 
      */
    public VisitableCompositeSequence(VisitableCompositeSequence sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public VisitableCompositeSequence copy() {
        return new VisitableCompositeSequence(this);
    }

    /**
      * 
      *
      */
    public Object current() {
        if (this.size() == 0) {
            return NULL; // to comply with the sequence contract
        }
        return this.sequences.get(0).current();
    }

    /**
      * 
      *
      */
    public Object next() {
        Object value = NULL; // to comply with the sequence contract
        for (Sequence<? extends Object> sequence : this) {
            Object v = sequence.next();
            if (value == null) {
                value = v;
            }
        }
        return value;
    }
    
    /**
      * @see SequenceValueVisitor#visitUnknown(Sequence, Object)
      */
    public <P> void accept(SequenceValueVisitor<P> visitor, P argument) {
        visitor.visitUnknown(this, argument);
    }
    
    /**
      * @see SequenceTypeVisitor#visitComposite(CompositeSequence, Object)
      */
    public <P> void accept(SequenceTypeVisitor<P> visitor, P argument) {
        visitor.visitComposite(this, argument);
    }
        
}
