/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */

/**
  * Implementations and examples of the <b>Singleton</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.127]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  *    
  *    Ensure a class only has one instance, and provide
  *    a global point of access to it.
  *    
  * </i></ul>
  * 
  * Here, the <b>Singleton</b> participant is represented by the 
  * {@link dk.rode.thesis.singleton.DanishAlphabetSequence},
  * {@link dk.rode.thesis.singleton.NorwegianAlphabetSequence},
  * {@link dk.rode.thesis.singleton.SimpsonsFamilySequence}, and 
  * {@link dk.rode.thesis.singleton.SimpsonsAndBouvierFamilySequence} 
  * classes, all but {@code NorwegianAlphanetSequence} performing lazy 
  * initialisation when the singleton instance is requested. However,
  * synchronisation is not required in either case because Java's standard 
  * static initialisation procedures are exploited. The
  * {@link dk.rode.thesis.singleton.SmileySequence} class, which also
  * represents the <b>Singleton</b> participant, therefore illustrates 
  * lazy initialisation using synchronisation (see implementation notes
  * in the class JavaDoc for each singleton type for further information). <p>
  * 
  * <i>Singleton Registry</i> functionality in a manner similar to
  * that described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.130-133] is also included 
  * in this package. The {@link dk.rode.thesis.singleton.SingletonRegistry} 
  * interface represents such a registry, using <i>class tokens</i> as arguments
  * to identify a given singleton instance in the registry using
  * the {@link dk.rode.thesis.singleton.SingletonRegistry#getInstance(Class)}
  * method. Here, however, registry functionality is independent of
  * inheritance as the registry is completely separated from actual
  * singleton types/instances, and registries have no influence over
  * when and how actual singleton instances are created. Furthermore, 
  * by declaring an interface to implement for required registry functionality, 
  * as opposed to using static methods directly, we get all the standard 
  * benefits of object behaviour and polymorphism at the cost of having 
  * to create (possibly singleton) registries as well. For example, a 
  * registry implementation may be synchronised or otherwise proxied or 
  * decorated, or used as a private member variable in a factory. <p> 
  * 
  * Three {@code SingletonRegistry} implementations are supplied, namely 
  * {@link dk.rode.thesis.singleton.StatelessSingletonRegistry},
  * {@link dk.rode.thesis.singleton.StatefullSingletonRegistry}, and
  * {@link dk.rode.thesis.singleton.LoadableSingletonRegistry}.
  * All defers the responsibility of creating the singleton objects 
  * elsewhere, either directly to the singleton type or to another registry 
  * (for the last two). <p>
  * 
  * The {@code StatelessSingletonRegistry} is stateless, and can return
  * singleton types <i>annotated</i> with the 
  * {@link dk.rode.thesis.singleton.Singleton} annotation. The annotation 
  * identifies the static <i>singleton method</i> used to acquire the 
  * singleton instance, e.g. {@code getInstance()}, and the actual creation
  * of the singleton instance is thus left up to the actual singleton type. 
  * Different means of singleton initialisation can therefore be employed 
  * "within" the same registry, for example static or lazy initialisation. 
  * Here, an annotation is preferred over an interface, because different 
  * singleton objects offer no common functionality except for conceptual 
  * behaviour in form of only a single instance available. The singleton 
  * method must in any case be static, and interfaces cannot declare static 
  * methods. An annotation also allows for different method names for 
  * different types, which is exemplified by the 
  * {@link dk.rode.thesis.singleton.NorwegianAlphabetSequence#getInstance()}
  * and {@link dk.rode.thesis.singleton.SimpsonsFamilySequence#getFamily()}
  * singleton methods. <p> 
  * 
  * The {@code StatefullSingletonRegistry} is a singleton registry that 
  * defers the actual acquiring of singleton instances to another 
  * registry, for example {@code StatelessSingletonRegistry}, but keeps 
  * a reference to each singleton instance that is <i>reachable</i> from 
  * a live daemon thread. Hence, as long as the statefull registry is 
  * within scope, all acquired singleton instances will be in scope and 
  * never be garbage collected. <p> 
  * 
  * The {@code LoadableSingletonRegistry} extends {@code StatefullSingletonRegistry}
  * to provide even more specialised functionality: singleton classes can be
  * dynamically loaded (if not already) based on their class name using the
  * <b>same</b> class loader. The registry only allows singleton types of the
  * proper generic bounds to loaded as well. <p>
  *  
  * <a name="uml"></a><b>UML Class Diagram</b>:<br>
  * <img src="doc-files/Singleton.png" border="0"><p>
  * 
  * <b>Implementation notes:</b><br>
  * Any singleton instance in Java is only a singleton per <i>class loader</i>
  * if created based on and utilising intrinsic class features. <p>
  * 
  * Java does not support multiple functional inheritance. Inheriting private
  * functionality is not possible. Hence, to allow a singleton type to be 
  * sub-classed as discussed by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.128,130-132], an
  * access modifier of either public or protected must be used for the class, 
  * behaviour, and data to be inherited. Any context can access public elements,
  * while members in the same package or sub-classes can access protected
  * elements. This poses potential problems because if a constructor in 
  * a singleton class is not declared private, it may be accessed by a given
  * context. Even worse, using reflection, even private members can be 
  * accessed if the security manager allows it. Hence, declaring a constructor
  * private is no guarantee that it cannot be invoked. <p>    
  * 
  * To remedy this while still allowing singleton types to be sub-classed, 
  * the singleton types declared in this package all use the 
  * {@link dk.rode.thesis.meta.reflect.CallerClass} or 
  * {@link dk.rode.thesis.meta.reflect.Caller} class to establish 
  * <i>who</i> the <i>caller</i> of the constructor is; that is, which 
  * class and/or member is invoking the constructor. The classes utilises 
  * Java's reflection capabilities. The idea is that if the constructor 
  * is the private singleton constructor, only the <i>exact same class</i>
  * is allowed to call it. If a sub-class constructor is called, for 
  * example a sub-class copy constructor, the constructor checks to see 
  * if the caller is a <i>constructor in a sub-class</i>. If so, the call
  * is allowed, otherwise a {@link dk.rode.thesis.singleton.SingletonError} 
  * exception is thrown. As an example, the 
  * {@link dk.rode.thesis.singleton.SimpsonsFamilySequence} is a singleton 
  * type that is allowed to be sub-classed. The  
  * {@link dk.rode.thesis.singleton.SimpsonsAndBouvierFamilySequence} class
  * inherits it and is also a singleton type itself, while the 
  * {@link dk.rode.thesis.singleton.MutatedSimpsonsFamilySequence} class
  * also inherits it, but is not a singleton type itself. We
  * call types enforcing such runtime access checks for 
  * <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.2.1" target="thesis"><i>Guarded Types</i></a>. <p>
  * 
  * However, it would in fact be possible to create a super-singleton
  * instance from a sub-class constructor! A {@code Caller} cannot 
  * determine <i>how</i> the super-class constructor is invoked from a 
  * sub-class, only from <i>where</i>. In the sub-class constructor, 
  * the super-constructor could be invoked as {@code super(..)} as 
  * would be expected, but also by explicitly creating a new instance 
  * using {@code new}. It does not help to declare the super-singleton
  * type as abstract. Hence, any singleton type that allows sub-classing 
  * has to trust the sub-classes to some extent, but must common misuse 
  * can be caught with the {@code CallerClass} and {@code Caller} 
  * classes. <p>
  * 
  * An alternative is to use enumerations with a single constant to represent
  * a singleton type. This is illustrated with the 
  * {@link dk.rode.thesis.singleton.DanishAlphabetSequence} implementation. 
  * We call this the 
  * <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.1.4" target="thesis"><i>Singleton-as-Single-Constant</i></a> idiom. <p>
  * 
  * No synchronisation is used in singleton types <b>or</b> in registries! <p> 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Singleton", 
         scope        = Scope.Object,
         purpose      = Purpose.Creational,
         participants = "Singleton")
package dk.rode.thesis.singleton;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;
